#include "Utils.h"
#include "Likelihood.h"
#include "PriorIndependence.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "VariableSet.h"

#include "PosteriorSampleHandlerFileDumper.h"

// Data generated from Poisson(0.5) using R: rpois(100, 0.5)
static const unsigned long N = 100;
static int data[N] = 
{
	1,0,0,2,2,1,0,0,0,0,0,1,0,1,1,0,2,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,1,0,0,1,0,0,0,1,0,0,0,0,0,0,1,0,0,0,1,1,0,1,0,1,0,0,0,1,1,0,0,0,1,0,2,1,1,1,0,0,0,0,0,0,0,0,1,1,1,0,2,2,1,1,0,0,0,0,1,1,1,0,0,
};

// Model 1: Poisson
class PoissonLikelihood : public bcm::Likelihood
{
public:
	bool EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
	{
		Real lambda = values[0];
		Real log_lambda = log(lambda);

		logp = 0;
		for (int i = 0; i < N; i++) {
			logp += log_lambda * data[i] - log((Real)bcm::factorial(data[i])) - lambda;
		}

		return true;
	}
};

// Model 2: Geometric
class GeometricLikelihood : public bcm::Likelihood
{
public:
	bool EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
	{
		Real mu = values[1];
		if (mu == 1.0) {
			mu = 1 - 1e-10;
		} else if (mu == 0.0) {
			mu = 1e-10;
		}
		Real logmu = log(mu);
		Real log1mmu = log(1 - mu);
		
		logp = 0;
		for (int i = 0; i < N; i++) {
			logp += log1mmu * data[i] + logmu;
		}

		return true;
	}
};

int main(int argc, char* argv[])
{
	bcm::logger->SetLogToConsole(bcm::Logger::Info);
	bcm::logger->LogTime();

	bcm::VariableSet varset;
	varset.AddVariable("lambda");
	varset.AddVariable("mu");

	bcm::PriorIndependence prior(&varset);
	prior.SetMarginal("lambda", bcm::UnivariateMarginal::CreateExponential(1.0));
	prior.SetMarginal("mu", bcm::UnivariateMarginal::CreateUniform(0.0, 1.0));

	PoissonLikelihood poisson_likelihood;
	GeometricLikelihood geometric_likelihood;
	
	//bcm::Sampler* sampler = bcm::SamplerFactory::Create("nested", 1);
	//bcm::Sampler* sampler = bcm::SamplerFactory::Create("mh", 1);
	bcm::Sampler* sampler = bcm::SamplerFactory::Create("smc", 1);
	//bcm::Sampler* sampler = bcm::SamplerFactory::Create("pt", 1);
	sampler->SetLatentVariableSet(&varset);
	sampler->SetReentrantPrior(&prior);

	Real evidence_poisson = std::numeric_limits<Real>::quiet_NaN();
	Real evidence_geometric = std::numeric_limits<Real>::quiet_NaN();
	Real evidence_poisson_sd = std::numeric_limits<Real>::quiet_NaN();
	Real evidence_geometric_sd = std::numeric_limits<Real>::quiet_NaN();

	bcm::PosteriorSampleHandlerFileDumper dumper;
	dumper.SetFile("samples.tsv");
	dumper.SetVariableSet(&varset);
	dumper.Initialize();
	sampler->SetPosteriorSampleHandler(dumper.GetFunctionPointer(), NULL);

	// Calculate evidence for model 1
	sampler->SetReentrantLikelihood(&poisson_likelihood);
	if (!sampler->Initialize()) {
		return -1;
	}
	if (!sampler->Run(&evidence_poisson, &evidence_poisson_sd)) {
		return -2;
	}
	sampler->Uninitialize();
	
	// Calculate evidence for model 2
	sampler->SetReentrantLikelihood(&geometric_likelihood);
	if (!sampler->Initialize()) {
		return -3;
	}
	if (!sampler->Run(&evidence_geometric, &evidence_geometric_sd)) {
		return -4;
	}
	sampler->Uninitialize();

	Real bf = evidence_poisson - evidence_geometric;
	Real bf_sd = sqrt(evidence_poisson_sd*evidence_poisson_sd + evidence_geometric_sd*evidence_geometric_sd);

	// Analytic calculation of evidence, as comparison
	int s1 = 0;
	Real t1 = 0.0;
	for (int i = 0; i < N; i++) {
		s1 += data[i];
		t1 += log((Real)bcm::factorial(data[i]));
	}
	Real analytic_evidence_poisson = bcm::logfactorial_approx(s1) - t1 - (s1 + 1) * log(N + 1.0);
	Real analytic_evidence_geometric = bcm::logfactorial_approx(N) + bcm::logfactorial_approx(s1) - bcm::logfactorial_approx(N + s1 + 1);
	Real analytic_bf = analytic_evidence_poisson - analytic_evidence_geometric;

	LOG("");
	LOG("Model 1: Poisson");
	LOG("  Analytic log evidence: %g", analytic_evidence_poisson);
	LOG("  Log evidence from sampler: %g +/- %g", evidence_poisson, evidence_poisson_sd);
	LOG("");
	LOG("Model 2: Geometric");
	LOG("  Analytic log evidence: %g", analytic_evidence_geometric);
	LOG("  Log evidence from sampler: %g +/- %g", evidence_geometric, evidence_geometric_sd);
	LOG("");
	LOG("Analytic log Bayes factor: %g", analytic_bf);
	LOG("Log Bayes factor from sampler: %g +/- %g", bf, bf_sd);

	delete sampler;

	return 0;
}
