library(XML)
library(h5)

load_model <- function(model_name, output_name) {
  model <- list()
  model$name = model_name
  model$output_folder <- paste(model_name, "/", output_name, sep="")
  
  # Prior
  model$prior <- list()
  varxml <- xmlTreeParse(paste(model_name, "/prior.xml", sep=""))
  variables <- xmlRoot(varxml)["variable", all=T]
  model$prior$variable_attrs <- lapply(variables, xmlAttrs)
  model$variables <- rep("", length(variables))
  for (i in 1:length(variables)) {
    model$variables[i] <- model$prior$variable_attrs[[i]]["name"]
  }
  model$nvar <- length(model$variables)
  
  # Posterior
  model$posterior <- list()
  samples <- load_binary(paste(model$output_folder, "/samples.bin", sep=""), numvar=model$nvar+4)
  model$posterior$samples <- samples[,1:model$nvar]
  colnames(model$posterior$samples) <- model$variables
  model$posterior$lprior <- samples[,1:model$nvar+1]
  model$posterior$llikelihood <- samples[,1:model$nvar+2]
  model$posterior$lposterior <- samples[,1:model$nvar+3]
  model$posterior$weights <- samples[,1:model$nvar+4] / sum(model$samples[,model$nvar+4])
  
  # Output values
  model$noutvar <- as.numeric(xmlValue(xmlRoot(varxml)[["output_variables"]]))
  if (model$noutvar > 0) {
    model$output <- load_binary(paste(model$output_folder, "/output_values.bin", sep=""), numvar=model$noutvar)
  }
  
  return(model);
}

load_ode_model <- function(model_name, output_folder, output_filename="output.nc") {
  model <- list()
  model$name = model_name
  model$output_folder <- paste(model_name, "/", output_folder, sep="")
  
  # Prior
  model$prior <- list()
  varxml <- xmlTreeParse(paste(model_name, "/prior.xml", sep=""))
  variables <- xmlRoot(varxml)["variable", all=T]
  model$prior$variable_attrs <- lapply(variables, xmlAttrs)
  model$variables <- rep("", length(variables))
  for (i in 1:length(variables)) {
    model$variables[i] <- model$prior$variable_attrs[[i]]["name"]
  }
  model$nvar <- length(model$variables)
  
  # Posterior
  output_file <- h5file(paste(model$output_folder, "/", output_filename, sep=""), 'r')
  
  model$posterior <- list()
  model$posterior$samples <- output_file["samples/variable_values"][]
  model$posterior$lprior <- output_file["samples/log_prior"][]
  model$posterior$llikelihood <- output_file["samples/log_likelihood"][]
  model$posterior$lposterior <- output_file["samples/log_posterior"][]
  model$posterior$weight <- output_file["samples/weight"][]
  colnames(model$posterior$samples) <- output_file["samples/variables"][]
  
  return(model);
}

load_sbmlpd_model <- function(model_name, output_folder, prior_file="prior.xml", output_filename="output.nc") {
  model <- list()
  model$name = model_name
  model$output_folder <- paste(model_name, "/", output_folder, sep="")
  
  # Load prior information
  model$prior <- list()
  varxml <- xmlTreeParse(paste(model_name, "/", prior_file, sep=""))
  variables <- xmlRoot(varxml)["variable", all=T]
  model$prior$variable_attrs <- lapply(variables, xmlAttrs)
  model$variables <- rep("", length(variables))
  for (i in 1:length(variables)) {
    model$variables[i] <- model$prior$variable_attrs[[i]]["name"]
  }
  model$nvar <- length(model$variables)
  
  # Load output
  output_file <- h5file(paste(model$output_folder, "/", output_filename, sep=""), 'r')
  
  model$posterior <- list()
  model$posterior$samples <- output_file["samples/variable_values"][]
  model$posterior$lprior <- output_file["samples/log_prior"][]
  model$posterior$llikelihood <- output_file["samples/log_likelihood"][]
  model$posterior$lposterior <- output_file["samples/log_posterior"][]
  colnames(model$posterior$samples) <- output_file["samples/variables"][]
  
  model$output <- list()
  model$output$signaling_molecules <- output_file["output_values/signaling_molecules"][]
  experiment_names <- list.groups(output_file["output_values"], full.names=F, recursive=F)
  for (ei in 1:length(experiment_names)) {
    en <- experiment_names[ei]
    model$output[[en]] <- list()
    model$output[[en]]$cell_lines <- output_file[paste("output_values", en, "cell_lines", sep="/")][]
    model$output[[en]]$signaling_activities <- output_file[paste("output_values", en, "signaling_activities", sep="/")][]
    model$output[[en]]$data_names <- output_file[paste("output_values", en, "data_names", sep="/")][]
    
    if (length(model$output[[en]]$data_names) > 0) {
      model$output[[en]]$data_model_names <- output_file[paste("output_values", en, "data_model_names", sep="/")][]
      model$output[[en]]$modeled_data_values <- output_file[paste("output_values", en, "modeled_data_values", sep="/")][]
      model$output[[en]]$observed_data_values <- list()
      for (di in 1:length(model$output[[en]]$data_names)) {
        model$output[[en]]$observed_data_values[[di]] <- output_file[paste("output_values/", en, "/observed_data_values_", di-1, sep="")][]
      }
    }
    
    if (length(dim(model$output[[en]]$signaling_activities)) == 4) {
      model$output[[en]]$drug_concentrations <- output_file[paste("output_values", en, "drug_concentrations", sep="/")][]
    }
  }
  h5close(output_file)
  return(model)
}

as.coda.mcmc <- function(model) {
  require(coda)
  return(mcmc(model$samples[,1:model$nvar]))
}

as.ggmcmc <- function(model) {
  require(coda)
  require(ggmcmc)
  return(ggs(mcmc(model$samples[,1:model$nvar])))
}

# Retrieve the index of a given variable
model_get_var_ix <- function(model, var_name) {
  match(var_name, model$variables)
}

# Plot prior and posterior distribution for a given variable
# You can specify either a variable name or a variable index
plot_variable_distribution <- function(model, var_ix=NULL, var_name=NULL, xlab="", ylim=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    plot_variable_distribution_impl(model$posterior$samples[,var_ix], varattrs, model$posterior$weights, xlab, ylim)
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

plot_trace <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    plot(model$posterior$samples[,var_ix], main=varattrs["name"])
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_mean <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="mean")
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_sd <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="sd")
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_median <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="median")
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_quantile <- function(model, var_ix=NULL, var_name=NULL, q)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="quantile", q=q)
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_autocorrelation <- function(model, var_ix=NULL, var_name=NULL, lag)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="autocorrelation", lag=lag)
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_decorrelation_lag <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="decorr_lag")
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_effective_sample_size <- function(model, var_ix=NULL, var_name=NULL)
{
  if (!is.null(var_name)) {
    var_ix <- model_get_var_ix(model, var_name)
  }
  if (!is.null(var_ix)) {
    varattrs <- model$prior$variable_attrs[[var_ix]]
    variable_statistic(model$posterior$samples[,var_ix], varattrs, statistic="ess")
  } else {
    stop("Either a variable index or a variable name has to be specified")
  }
}

variable_summary <- function(model)
{
  df <- data.frame(row.names=model$variables)
  for (j in 1:model$nvar) {
    df$mean[j] <- variable_mean(model, j)
    df$sd[j] <- variable_sd(model, j)
    df$median[j] <- variable_median(model, j)
    df$q025[j] <- variable_quantile(model, j, q=0.025)
    df$q975[j] <- variable_quantile(model, j, q=0.975)
    df$autocorrelation_lag1[j] <- variable_autocorrelation(model, j, lag=1)
    df$decorrelation_lag[j] <- variable_decorrelation_lag(model, j)
    df$ess[j] <- variable_effective_sample_size(model, j)
  }
  return(df)
}

prior_bounds <- function(model, var_ix, q=c(0.01,0.99))
{
  p <- rep(NA, length(q))
  varattrs <- model$prior$variable_attrs[[var_ix]]
  priortype <- varattrs["type"]
  if (is.na(priortype) || priortype == "regular") {
    distribution <- varattrs["distribution"]
    if (distribution == "normal") {
      mu = as.numeric(varattrs["mu"]);
      sigma = as.numeric(varattrs["sigma"]);
      p <- qnorm(q, mean=mu, sd=sigma);
    } else if (distribution == "gamma") {
      k = as.numeric(varattrs["k"]);
      theta = as.numeric(varattrs["theta"]);
      p <- qgamma(q, shape=k, scale=theta);
    } else if (distribution == "uniform") {
      a = as.numeric(varattrs["lower"]);
      b = as.numeric(varattrs["upper"]);
      p <- c(a,b)
      p <- qunif(q, min=a, max=b);
    } else if (distribution == "half_cauchy") {
      scale = as.numeric(varattrs["scale"]);
      p <- qcauchy(q, scale=scale);
    } else if (distribution == "beta") {
      a = as.numeric(varattrs["a"]);
      b = as.numeric(varattrs["b"]);
      p <- qbeta(q, a, b)
    } else if (distribution == "exponential") {
      rate = as.numeric(varattrs["lambda"]);
      p <- qexp(q, rate=rate)
    }
  }
  return(p)
}

# Posterior predictive distribution as a barplot
ppd_barplot <- function(variable_samples, data, labels, bounds=c(0.05, 0.95), ylim=NULL, ...) {
  bound_color <- rgb(43,131,186, max=255)
  bound_color_alpha <- rgb(43,131,186, 0.4*255, max=255)
  
  if (is.null(ncol(data))) {
    if (ncol(variable_samples) != length(data)) {
      stop("Number of columns of data and posterior samples should be the same")
    }
  } else {
    if (ncol(variable_samples) != ncol(data)) {
      stop("Number of columns of data and posterior samples should be the same")
    }
  }
  
  if (bounds[2] < bounds[1]) {
    bounds <- rev(bounds)
  }
  
  if (is.null(ylim)) {
    sample_min <- min(quantile(variable_samples, bounds[1], na.rm=T))
    sample_max <- max(quantile(variable_samples, bounds[2], na.rm=T))
    ylim <- c(min(data, sample_min, na.rm=T), max(data, sample_max, na.rm=T))
    ylim <- c(ylim[1] - 0.05 * (ylim[2]-ylim[1]), ylim[2] + 0.05 * (ylim[2]-ylim[1]))
  }
  
  xpos <- barplot(rep(0,ncol(variable_samples)), names.arg=labels, las=2, border=NA, ylim=ylim, ...)
  for (i in 1:ncol(variable_samples)) {
    barwidth <- 0.6
    lx <- xpos[i] - barwidth / 2
    ux <- xpos[i] + barwidth / 2
    
    # Posterior
    ly <- quantile(variable_samples[,i], probs=bounds[1], na.rm=T)
    uy <- quantile(variable_samples[,i], probs=bounds[2], na.rm=T)
    polygon(c(lx,lx,ux,ux), c(ly,uy,uy,ly), col=bound_color_alpha, border=F)
    lines(c(lx,ux), c(ly,ly), col=bound_color)
    lines(c(lx,ux), c(uy,uy), col=bound_color)
    
    # Data
    if (is.null(nrow(data))) {
      points(xpos[i], data[i], pch=19, cex=1.5)
    } else {
      points(rep(xpos[i],nrow(data)), data[,i], pch=19, cex=1.5)
    }
  }
  return(xpos)
}

# Posterior predictive distribution as a lineplot
ppd_lineplot <- function(x.data, y.data, x.samples, y.samples, bounds=c(0.05, 0.95), ylim=NULL, ...) {
  bound_color <- rgb(43,131,186, max=255)
  bound_color_alpha <- rgb(43,131,186, 0.4*255, max=255)
  
  if (bounds[2] < bounds[1]) {
    bounds <- rev(bounds)
  }
  
  ly <- apply(y.samples, 2, quantile, bounds[1], na.rm=T)
  my <- apply(y.samples, 2, quantile, 0.5, na.rm=T)
  uy <- apply(y.samples, 2, quantile, bounds[2], na.rm=T)
  
  if (is.null(ylim)) {
    sample_min <- min(ly)
    sample_max <- max(uy)
    ylim <- c(min(y.data, sample_min), max(y.data, sample_max))
    ylim <- c(ylim[1] - 0.05 * (ylim[2]-ylim[1]), ylim[2] + 0.05 * (ylim[2]-ylim[1]))
  }
  
  plot(x.samples, my, typ="l", lwd=2, ylim=ylim, col=bound_color, ...)
  polygon(c(x.samples, rev(x.samples)), c(ly,rev(uy)), col=bound_color_alpha, border=F)
  lines(x.samples, ly, col=bound_color)
  lines(x.samples, uy, col=bound_color)
  if (is.null(nrow(y.data))) {
    points(x.data, y.data, pch=19)
  } else {
    for (i in 1:nrow(y.data)) {
      points(x.data, y.data[i,], pch=19)
    }
  }
}



#
# Internal functions
#

load_binary <- function(filename, numvar) {
  file_size <- file.info(filename)$size
  numsamples <- file_size / (4 * numvar)
  
  infile <- file(filename, "rb");
  data <- matrix(NA, numsamples, numvar);
  for (sample in 1:numsamples) {
    d <- readBin(infile, double(), n=numvar, size=4);
    data[sample,] <- d;
  }
  close(infile);
  return(data);
}

plot_variable_distribution_impl <- function(samples, varattrs, weights=NULL, xlab="", ylim=NULL)
{
  name = varattrs["name"];
  
  minx <- min(samples);
  maxx <- max(samples);
  
  lbound <- NA
  ubound <- NA
  
  priortype <- varattrs["type"]
  if (is.na(priortype) || priortype == "regular") {
    distribution <- varattrs["distribution"]
    if (distribution == "normal") {
      mu = as.numeric(varattrs["mu"]);
      sigma = as.numeric(varattrs["sigma"]);
      minx <- min(minx, qnorm(0.01, mean=mu, sd=sigma));
      maxx <- max(maxx, qnorm(0.99, mean=mu, sd=sigma));
      priorx <- seq(minx, maxx, length.out=200);
      priory <- dnorm(priorx, mean=mu, sd=sigma);
    } else if (distribution == "gamma") {
      k = as.numeric(varattrs["k"]);
      theta = as.numeric(varattrs["theta"]);
      minx <- 0;#min(minx, qgamma(0.01, shape=k, scale=theta));
      maxx <- max(maxx, qgamma(0.99, shape=k, scale=theta));
      priorx <- seq(minx, maxx, length.out=200);
      priory <- dgamma(priorx, shape=k, scale=theta);
      lbound <- 0;
    } else if (distribution == "uniform") {
      a = as.numeric(varattrs["lower"]);
      b = as.numeric(varattrs["upper"]);
      minx <- a;
      maxx <- b;
      priorx <- seq(minx, maxx, length.out=200);
      priory <- dunif(priorx, min=a, max=b);
      lbound <- a;
      ubound <- b;
    } else if (distribution == "half_cauchy") {
      scale = as.numeric(varattrs["scale"]);
      minx <- 0;
      maxx <- max(maxx, qcauchy(0.95, scale=scale));
      priorx <- seq(minx, maxx, length.out=200);
      priory <- 2.0 * dcauchy(priorx, scale=scale);
      lbound <- 0;
    } else if (distribution == "beta") {
      a = as.numeric(varattrs["a"]);
      b = as.numeric(varattrs["b"]);
      minx <- 1e-6;
      maxx <- 1-1e-6;
      priorx <- seq(minx, maxx, length.out=200);
      priory <- dbeta(priorx, a, b);
      lbound <- 0;
      ubound <- 1;
    } else if (distribution == "exponential") {
      rate = as.numeric(varattrs["lambda"]);
      minx <- 0;
      maxx <- max(maxx, qexp(0.99, rate=rate));
      priorx <- seq(minx, maxx, length.out=200);
      priory <- dexp(priorx, rate=rate);
      ubound <- 0;
    }
  }
  
  hmax <- 10 * sd(samples)
  
  # Calculate KDE bandwidth
  bw <- bw.SJ(samples, lower=1e-20*hmax, upper=hmax)
  
  # If necessary, reflect samples around bounds and add the reflection to the sample, to get a more reasonable KDE near the bounds
  factor <- 1
  density_samples <- samples
  density_weights <- weights
  if (!is.na(lbound)) {
    reflected <- lbound - (samples - lbound)
    density_samples <- c(density_samples, reflected)
    density_weights <- c(density_weights, weights)
    factor <- factor + 1
  }
  if (!is.na(ubound)) {
    reflected <- ubound + (ubound - samples)
    density_samples <- c(density_samples, reflected)
    density_weights <- c(density_weights, weights)
    factor <- factor + 1
  }
  if (!is.null(density_weights)) {
    density_weights <- density_weights / factor 
  }
  
  # Calculate density
  d <- density(density_samples, bw=bw, weights=density_weights, from=minx, to=maxx);
  d$y <- d$y * factor
  
  maxy <- max(d$y, priory);
  if (is.null(ylim)) {
    ylim <- c(0, maxy)
  }
  plot(priorx, priory, col="blue", main=name, xlab=xlab, ylab="Probability density", type="l", xlim=c(minx, maxx), ylim=ylim, lwd=2);
  lines(d$x, d$y, col="red", lwd=2)
}

variable_statistic <- function(samples, weights=NULL, xlab="", ylim=NULL, statistic, ...)
{
  if (statistic == "mean") {
    return(mean(samples))
  }
  if (statistic == "median") {
    return(median(samples))
  }
  if (statistic == "sd") {
    return(sd(samples))
  }
  if (statistic == "quantile") {
    arguments <- list(...)
    return(quantile(samples, probs=arguments$q))
  }
  if (statistic == "autocorrelation") {
    lag <- list(...)$lag
    ac <- acf(samples, plot=F, lag.max=lag)$acf[lag+1]
    return(ac)
  }
  if (statistic == "decorr_lag") {
    ac <- acf(samples, plot=F, lag.max=length(samples)/2)
    threshold <- 2.0 / sqrt(length(samples))
    sign <- ac$acf < threshold
    return(match(T, sign))
  }
  if (statistic == "ess") {
    ac <- acf(samples, plot=F, lag.max=length(samples)/2)
    first_neg <- Position(function(x) x < 0, ac$acf[,1,1])
    if (first_neg > 2) {
      ess <- length(samples) / (1 + 2*sum(ac$acf[2:(first_neg-1),1,1]))
    } else {
      ess <- length(samples)
    }
    return(ess)
  }
}

png_tile <- function(filename, width, height, nplots)
{
  png(filename, width=width, height=height)
  nx <- ceiling(sqrt(nplots))
  ny <- ceiling(nplots / nx)
  par(mfrow=c(ny, nx))
}