#include "Utils.h"
#include "Likelihood.h"
#include "Model.h"
#include "ParseUtilFunctions.h"

#include <fstream>
#include <boost/regex.hpp>

namespace ModelParser {

Likelihood::Likelihood(Model* model)
	: model(model)
	, sum_ix(0)
	, indent_ix(0)
{
}

Likelihood::~Likelihood()
{
}

bool Likelihood::Parse(const std::string& line, unsigned int linenr)
{
	boost::smatch results;

	// begin function
	if (boost::regex_match(line, results, boost::regex("function[\\s\\t]{0,}(\\w+)"))) {
		if (!in_function.empty()) {
			LOGERROR("Function definition inside another function");
			return false;
		}

		in_function = results[1].str();
		return true;
	}

	// end function
	if (boost::regex_match(line, results, boost::regex("end[\\s\\t]{0,}function"))) {
		if (in_function.empty()) {
			LOGERROR("end function outside function definition");
			return false;
		}

		in_function.clear();
		return true;
	}

	std::string code;

#if 0
	// set_variable
	if (boost::regex_match(line, results, boost::regex("set_variable[\\s\\t]{0,}\\((.+?)\\)"))) {
		std::string set_variable_code = results[1].str();

		return true;
	}
#endif

	// begin for
	if (boost::regex_match(line, results, boost::regex("for[\\s\\t]{0,}(\\w+)[\\s\\t]{0,}\\=[\\s\\t]{0,}(\\w+)[\\s\\t]{0,}to[\\s\\t]{0,}(.+)"))) {
		std::string iter_var = results[1].str();
		std::string start = results[2].str();
		std::string end = results[3].str();

		code = "for (size_t " + iter_var + " = " + start + "; " + iter_var + " < " + end + "; " + iter_var + "++) {";
		AddCodeLine(code);
		indent_ix++;

		return true;
	}

	// end for
	if (boost::regex_match(line, results, boost::regex("end[\\s\\t]{0,}for"))) {
		indent_ix--;
		code = "}";
		AddCodeLine(code);
		return true;
	}

	// begin if
	if (boost::regex_match(line, results, boost::regex("if[\\s\\t]{0,}(.+?)"))) {
		std::string condition = results[1].str();

		code = "if (" + condition + ") {";
		AddCodeLine(code);
		indent_ix++;

		return true;
	}

	// else
	if (boost::regex_match(line, results, boost::regex("^[\\s\\t]{0,}else[\\s\\t]{0,}$"))) {
		indent_ix--;
		AddCodeLine("} else {");
		indent_ix++;
		return true;
	}

	// end if
	if (boost::regex_match(line, results, boost::regex("end[\\s\\t]{0,}if"))) {
		indent_ix--;
		AddCodeLine("}");
		return true;
	}

	return ParseDirectCode(line, linenr);
}

void Likelihood::WriteFunctionsHeader(std::ofstream& ofs) const
{
	for (std::map<std::string, std::vector<std::string> >::const_iterator fi = function_code.begin(); fi != function_code.end(); ++fi) {
		const std::string& function_name = fi->first;
		ofs << "\tvoid " << function_name << "();\n";
	}
}

void Likelihood::WriteFunctionsSource(std::ofstream& ofs) const
{
	std::string class_name = model->GetClassName();

	for (std::map<std::string, std::vector<std::string> >::const_iterator fi = function_code.begin(); fi != function_code.end(); ++fi) {
		ofs << "\n";
		const std::string& function_name = fi->first;
		ofs << "void " << class_name << "::" << function_name << "()\n";
		ofs << "{\n";

		const std::vector<std::string>& lines = fi->second;
		for (std::vector<std::string>::const_iterator li = lines.begin(); li != lines.end(); ++li) {
			const std::string& line = *li;
			ofs << "\t" << line << "\n";
		}

		ofs << "}\n";
	}
}

void Likelihood::WriteSource(std::ofstream& ofs) const
{
	for (std::vector<std::string>::const_iterator li = code.begin(); li != code.end(); ++li) {
		const std::string& line = *li;
		ofs << "\t" << line << "\n";
	}
}

void Likelihood::AddCodeLine(const std::string& line)
{
	std::string indented_line;
	for (size_t i = 0; i < indent_ix; i++) indented_line += "\t";
	indented_line += line;

	if (!in_function.empty()) {
		std::vector<std::string>& fn_code = function_code[in_function];
		fn_code.push_back(indented_line);
	} else {
		code.push_back(indented_line);
	}
}

bool Likelihood::ParseDirectCode(const std::string& line, unsigned int linenr)
{
	std::string code_line = line;

	// Built-in functions, so far: sum, product, min, max
	{
		boost::regex sum_re("\\b(sum|product|min|max)\\((.+?)\\)");
		boost::smatch results;
		std::string::const_iterator text_iter = code_line.cbegin();
		while (boost::regex_search(text_iter, code_line.cend(), results, sum_re))
		{
			std::string sum_var_name = "_" + std::to_string((uint64)sum_ix);
			if (results.size() == 3) {
				std::string function = results[1];
				std::string contents = results[2];

				typedef std::pair<const Variable*, size_t> VariableWithIndex;
				std::vector<VariableWithIndex> vars;

				boost::regex var_re("(\\w+)(\\[[\\w:]+\\])");
				boost::smatch results;
				boost::regex_match(contents, results, var_re);
				if (results.size() < 3) {
					LOGERROR("Syntax error in line %u: \"%s\", related to sum()", linenr, line.c_str());
					return false;
				}

				// Find variable
				const std::string varname = results[1].str();
				const Variable* variable = model->FindVariable(varname);
				if (variable == NULL) {
					LOGERROR("Unknown variable \"%s\" in line %u: \"%s\"", results[1].str().c_str(), linenr, line.c_str());
					return false;
				}

				std::string indexing = results[2];
				if (indexing == "[:]") {
					for (size_t i = 0; i < variable->GetNumReplicates(); i++) {
						VariableWithIndex varix;
						varix.first = variable;
						varix.second = i;
						vars.push_back(varix);
					}
				} else {
					boost::regex ix_re("\\[(\\d+):(\\d+)\\]");
					boost::smatch results;
					boost::regex_match(indexing, results, ix_re);
					if (results.size() < 3) {
						LOGERROR("Syntax error in line %u: \"%s\", variable indexing", linenr, line.c_str());
						return false;
					}

					size_t start = (size_t)atoi(results[1].str().c_str());
					size_t end = (size_t)atoi(results[2].str().c_str());
					for (size_t i = start; i <= end; i++) {
						if (i == 0 || i > variable->GetNumReplicates()) {
							LOGERROR("Error in line %u: \"%s\", variable indexing out of bounds (note indexing starts at 1)", linenr, line.c_str());
							return false;
						}
						VariableWithIndex varix;
						varix.first = variable;
						varix.second = i - 1;
						vars.push_back(varix);
					}
				}

				std::string function_code;
				if (function == "sum") {
					function_code = "double " + sum_var_name + " = 0.0;";
					for (std::vector<VariableWithIndex>::const_iterator vi = vars.begin(); vi != vars.end(); vi++) {
						function_code += " " + sum_var_name + " += " + vi->first->GetName() + "[" + std::to_string((uint64)vi->second) + "];";
					}
				} else if (function == "product") {
					function_code = "double " + sum_var_name + " = 1.0;";
					for (std::vector<VariableWithIndex>::const_iterator vi = vars.begin(); vi != vars.end(); vi++) {
						function_code += " " + sum_var_name + " *= " + vi->first->GetName() + "[" + std::to_string((uint64)vi->second) + "];";
					}
				} else if (function == "min") {
					function_code = "double " + sum_var_name + " = std::numeric_limits<double>::infinity();";
					for (std::vector<VariableWithIndex>::const_iterator vi = vars.begin(); vi != vars.end(); vi++) {
						function_code += " " + sum_var_name + " = std::min(" + sum_var_name + ", " + vi->first->GetName() + "[" + std::to_string((uint64)vi->second) + "]);";
					}
				} else if (function == "max") {
					function_code = "double " + sum_var_name + " = -std::numeric_limits<double>::infinity();";
					for (std::vector<VariableWithIndex>::const_iterator vi = vars.begin(); vi != vars.end(); vi++) {
						function_code += " " + sum_var_name + " = std::max(" + sum_var_name + ", " + vi->first->GetName() + "[" + std::to_string((uint64)vi->second) + "]);";
					}
				} else {
					LOGERROR("Unknown built-in function \"%s\"", function.c_str());
					return false;
				}
				AddCodeLine(function_code);
			} else {
				LOGERROR("Syntax error in line %u: \"%s\", related to sum/product/min/max()", linenr, line.c_str());
				return false;
			}
			
			std::string newline;
			newline = code_line.substr(0, results.position());
			newline += sum_var_name;
			newline += code_line.substr(results.position() + results[0].length());
			code_line = newline;

			text_iter = code_line.cbegin();
			sum_ix++;
		}
	}
	
	// Parse probability distribution functions
	{
		boost::smatch results;
		if (boost::regex_search(code_line, results, boost::regex("\\Wnormal[\\t\\s]{0,}\\((.+?)\\)"))) {
			std::string normal_contents = results[1].str();
			boost::trim(normal_contents);

			std::string replace_code;
			if (!ParseNormal(normal_contents, replace_code, line, linenr)) {
				return false;
			}
			
			std::string newline;
			newline = code_line.substr(0, results.position());
			newline += replace_code;
			newline += code_line.substr(results.position() + results[0].length());
			code_line = newline;
		}
	}
	{
		boost::smatch results;
		if (boost::regex_search(code_line, results, boost::regex("\\Wstudentt[\\t\\s]{0,}\\((.+?)\\)"))) {
			std::string normal_contents = results[1].str();
			boost::trim(normal_contents);

			std::string replace_code;
			if (!ParseTDist(normal_contents, replace_code, line, linenr)) {
				return false;
			}
			
			std::string newline;
			newline = code_line.substr(0, results.position());
			newline += replace_code;
			newline += code_line.substr(results.position() + results[0].length());
			code_line = newline;
		}
	}

	// Expand-to-array?
	bool expand_to_array = false;
	size_t array_size = 1;
	const Variable* assign_to_variable = NULL;
	{
		boost::regex array_re("(\\w+)(\\[:\\]).+?=");
		boost::smatch results;
		if (boost::regex_search(code_line, results, array_re)) {
			expand_to_array = true;
			assign_to_variable = model->FindVariable(results[1].str());
			if (assign_to_variable == NULL) {
				LOGERROR("Unknown variable \"%s\" in line %u: \"%s\"", results[1].str().c_str(), linenr, line.c_str());
				return false;
			}
			array_size = assign_to_variable->GetNumReplicates();

			std::string array_code;
			array_code += "for (size_t array_i = 0; array_i < ";
			array_code += std::to_string((uint64)array_size);
			array_code += "; array_i++) {";
			AddCodeLine(array_code);
		}
	}

	// Data variable?
	{
		boost::smatch results;
		while (boost::regex_search(code_line, results, boost::regex("data->(\\w+)\\[(.+?)\\]"))) {
			std::string dataset = results[1].str();
			std::string indexing = results[2].str();

			boost::trim(indexing);
			std::vector<std::string> indices;
			bcm::tokenize(indexing, indices, ",");

			const Data* data = model->FindData(dataset);
			if (!data) {
				LOGERROR("Unknown data \"%s\" in line %u", dataset.c_str(), linenr);
				return false;
			}

			if (indices.size() != 2) {
				LOGERROR("Data \"%s\" must have 2 indices in line %u: %s", dataset.c_str(), linenr, line.c_str());
			}

			boost::smatch ix_results;
			if (boost::regex_match(indices[1], ix_results, boost::regex("\"(.+?)\""))) {
				std::string varname = ix_results[1].str();
				size_t var_ix;
				if (!data->GetVariableIndex(varname, var_ix)) {
					LOGERROR("No variable names \"%s\" in data \"%s\" in line %u: %s", varname.c_str(), dataset.c_str(), linenr, line.c_str());
					return false;
				}
				indices[1] = std::to_string((uint64)var_ix);
			}

			if (indices[1] == ":") {
				if (expand_to_array) {
					std::string mapping_code = "\tstatic const size_t var_to_data_remapping[" + std::to_string((uint64)array_size) + "] = { ";
					for (size_t i = 0; i < array_size; i++) {
						const std::string& name = assign_to_variable->GetReplicateName(i);
						size_t ix;
						if (!data->GetVariableIndex(name, ix)) {
							LOGERROR("Data does not have a variable \"%s\" for assignment to variable replicate in line %u: %s", name.c_str(), linenr, line.c_str());
							return false;
						}

						mapping_code += std::to_string((uint64)ix) + ",";
					}
					mapping_code += "};";
					AddCodeLine(mapping_code);

					indices[1] = "var_to_data_remapping[array_i]";
				} else {
					LOGERROR("Can only index data array with \":\" in an assignment to \":\" in line %u: \"%s\"", linenr, line.c_str());
					return false;
				}
			}

			std::string replace_code = "data_" + dataset + "[" + indices[0] + "][" + indices[1] + "]";
			
			std::string newline;
			newline = code_line.substr(0, results.position());
			newline += replace_code;
			newline += code_line.substr(results.position() + results[0].length());
			code_line = newline;
		}
	}
	
	std::string work_code_line = code_line;

	// Replace variables
	boost::regex var_re("(\\w+)(\\[.+?\\])");
	std::string::const_iterator text_iter = work_code_line.cbegin();
	boost::smatch results;

	while (boost::regex_search(text_iter, work_code_line.cend(), results, var_re))
	{
		if (results.size() < 3) {
			LOGERROR("Syntax error in line %u: \"%s\"", linenr, line.c_str());
			return false;
		}

		if (results[1].str() == "var_to_data_remapping") {
			text_iter += results.position() + results.length();
			continue;
		}
		if (strncmp(results[1].str().c_str(), "data_", 5) == 0) {
			text_iter += results.position() + results.length();
			continue;
		}

		// Find variable
		const Variable* variable = model->FindVariable(results[1].str());
		if (variable == NULL) {
			LOGERROR("Unknown variable \"%s\" in line %u: \"%s\"", results[1].str().c_str(), linenr, line.c_str());
			return false;
		}
		
		std::string replacement_code;
		std::string indexing = results[2];
		if (indexing == "[:]") {
			if (expand_to_array) {
				if (variable->GetNumReplicates() != array_size) {
					LOGERROR("Syntax error in line %u: \"%s\", inconsistent array size in expand-to-array line", linenr, line.c_str());
					return false;
				}
					
				replacement_code = variable->GetName() + "[array_i]";
			} else {
				LOGERROR("Syntax error in line %u: \"%s\", whole-array indexing outside of built in function or expand-to-array line", linenr, line.c_str());
				return false;
			}
		} else {
			boost::smatch results;
			if (boost::regex_match(indexing, results, boost::regex("\\[\"(.+?)\"\\]"))) {
				size_t ix;
				if (variable->GetReplicateIndex(results[1].str(), ix)) {
					replacement_code = variable->GetName() + "[" + std::to_string((uint64)ix) + "]";
				} else {
					LOGERROR("Error in line %u: \"%s\", unknown name \"%s\"", linenr, line.c_str(), results[1].str().c_str());
					return false;
				}
			} else {
				replacement_code = variable->GetName() + indexing;
			}
		}

		std::string newline;
		newline = work_code_line.substr(0, text_iter - work_code_line.cbegin() + results.position());
		newline += replacement_code;
		size_t next_start = newline.size();
		newline += work_code_line.substr(text_iter - work_code_line.cbegin() + results.position() + results[0].length());
		work_code_line = newline;
		text_iter = work_code_line.cbegin() + next_start;
	}

	work_code_line += ";";

	if (expand_to_array) {
		work_code_line = std::string("\t") + work_code_line;
	}
	
	AddCodeLine(work_code_line);

	if (expand_to_array) {
		AddCodeLine("}");
	}

	return true;
}

bool Likelihood::ParseNormal(const std::string& contents, std::string& code, const std::string line, unsigned int linenr)
{
	boost::smatch results;
	if (boost::regex_search(contents, results, boost::regex("(.+?)\\|(.+?),(.+?)$"))) {
		std::string x = results[1].str();

		std::string mu = results[2].str();
		boost::smatch mu_results;
		if (boost::regex_search(mu, mu_results, boost::regex("[\\s\\t]{0,}mu[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			mu = mu_results[1].str();
		} else {
			LOGERROR("Error in normal likelihood specification at line %u: \"%s\", for mu", linenr, line.c_str());
			return false;
		}

		std::string sigma = results[3].str();
		boost::smatch sigma_results;
		if (boost::regex_search(sigma, sigma_results, boost::regex("[\\s\\t]{0,}sigma[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			sigma = sigma_results[1].str();
		} else {
			LOGERROR("Error in normal likelihood specification at line %u: \"%s\", for sigma", linenr, line.c_str());
			return false;
		}

		code = "bcm::LogPdfNormal(" + x + ", " + mu + ", " + sigma + ")";
		return true;
	} else {
		LOGERROR("Error in normal likelihood specification at line %u: \"%s\"", linenr, line.c_str());
		return false;
	}
}

bool Likelihood::ParseTDist(const std::string& contents, std::string& code, const std::string line, unsigned int linenr)
{
	boost::smatch results;
	if (boost::regex_search(contents, results, boost::regex("(.+?)\\|(.+?),(.+?),(.+?),(.+?)$"))) {
	} else if (boost::regex_search(contents, results, boost::regex("(.+?)\\|(.+?),(.+?),(.+?)$"))) {
	} else {
		LOGERROR("Error in studentt likelihood specification at line %u: \"%s\"", linenr, line.c_str());
		return false;
	}
	
	std::string x = results[1].str();
	
	std::string nu, mu, sigma, logc;
	for (boost::smatch::size_type i = 2; i < results.size(); i++) {
		std::string element = results[i].str();
		
		boost::smatch el_results;
		if (boost::regex_search(element, el_results, boost::regex("[\\s\\t]{0,}nu[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			nu = el_results[1].str();
		} else if (boost::regex_search(element, el_results, boost::regex("[\\s\\t]{0,}mu[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			mu = el_results[1].str();
		} else if (boost::regex_search(element, el_results, boost::regex("[\\s\\t]{0,}sigma[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			sigma = el_results[1].str();
		} else if (boost::regex_search(element, el_results, boost::regex("[\\s\\t]{0,}logc[\\s\\t]{0,}=[\\s\\t]{0,}(.+?)[\\s\\t]{0,}$"))) {
			logc = el_results[1].str();
		}
	}

	if (nu.empty() || mu.empty() || sigma.empty()) {
		LOGERROR("Error in studentt likelihood specification at line %u: \"%s\" - need at least nu, mu and sigma", linenr, line.c_str());
		return false;
	}

	if (logc.empty()) {
		code = "bcm::LogPdfT(" + x + ", " + mu + ", " + sigma + ", " + nu + ", true)";
	} else {
		code = "bcm::LogPdfT(" + x + ", " + mu + ", " + sigma + ", " + nu + ", " + logc + ", true)";
	}
	return true;
}

}
