#include "Utils.h"
#include "Model.h"

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>

namespace po = boost::program_options;

int main(int argc, char* argv[])
{
	printf("BCM mdlparse version %u.%u.%u\n", BCM_VERSION_MAJOR, BCM_VERSION_MINOR, BCM_VERSION_FIX);

	bcm::logger->SetLogToConsole(bcm::Logger::Info);

	std::string inputfn;
	std::string bcm_path;

	try {
		po::options_description desc("Allowed options");
		desc.add_options()
			("filename,f",	po::value<std::string>()->default_value("model.txt"),	"input filename, can include a path and should end in .txt")
			("bcm",			po::value<std::string>()->default_value(""),			"path to the root of the bcm folder")
			("optimize-native",														"generate CMake configuration with settings to optimize for the native CPU architecture")
		;

		po::positional_options_description pod;
		pod.add("filename", 1);

		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(desc).positional(pod).run(), vm);
		po::notify(vm);

		inputfn = vm["filename"].as<std::string>();
		bcm_path = vm["bcm"].as<std::string>();
	} catch (po::error &e) {
		LOGERROR("Error processing program options: %s", e.what());
		return -1;
	}

	if (bcm_path == "") {
		char* bcm_env = getenv("BCM_ROOT");
		if (!bcm_env) {
			LOGERROR("Cannot determine the path containing the bcm installation.");
			LOGERROR("Either set the BCM_ROOT environment variable, or call this program with --bcm=/path/to/bcm");
			return -1;
		} else {
			bcm_path = bcm_env;
		}
	}
	bcm::fix_path(bcm_path);
	
	// Verify we have a good filename as input
	// 1. must have .txt extension
	std::string extension;
	std::string path;
	std::string model_name;
	std::string inputfn_without_extension;
	size_t extension_pos = inputfn.find_last_of(".");
	if (extension_pos != std::string::npos) {
		extension = inputfn.substr(extension_pos + 1);
		inputfn_without_extension = inputfn.substr(0, extension_pos);
		std::transform(extension.begin(), extension.end(), extension.begin(), ::tolower);
		if (extension != "txt") {
			extension.clear();
		}
	}
	
	// 2. Extract path
	if (!extension.empty()) {
		size_t path_end = inputfn_without_extension.find_last_of("\\/");
		if (path_end != std::string::npos) {
			path = inputfn_without_extension.substr(0, path_end);
			model_name = inputfn_without_extension.substr(path_end);
		} else {
			path = "";
			model_name = inputfn_without_extension;
		}
	}
	bcm::fix_path(path);

	if (extension.empty() || model_name.empty()) {
		LOGERROR("Input filename \"%s\" is not in the right format.", inputfn.c_str());
		LOGERROR("Filename should be: [path/]model_name.txt");
		return -1;
	}

	ModelParser::Model model;
	bool result;

	LOG("Parsing model file \"%s\"...", inputfn.c_str());
	result = model.Parse(inputfn);
	if (!result) {
		return -2;
	}

	std::string output_path = path + model_name + "/";
	if (!boost::filesystem::is_directory(output_path)) {
		std::string create_path = output_path.substr(0, output_path.size()-1);
		if (!boost::filesystem::create_directories(create_path)) {
			LOGERROR("Could not create output directory.");
			return -3;
		}
	}

	LOG("Writing output to path \"%s\"...", output_path.c_str());
	result = model.WriteHeader(output_path);
	if (!result) {
		return -4;
	}
	result = model.WriteSource(output_path);
	if (!result) {
		return -5;
	}
	result = model.WriteVariableFile(output_path);
	if (!result) {
		return -6;
	}
	result = model.WriteCMakeFile(output_path, bcm_path);
	if (!result) {
		return -7;
	}

	return 0;
}
