#include "Utils.h"
#include "Prior.h"
#include "NestedSamplingStrategyEllipsoid.h"
#include "NestedSamplingUtils.h"
#include "VariableSet.h"

#include <Eigen/Eigenvalues>

namespace bcm {

NestedSamplingStrategyEllipsoid::NestedSamplingStrategyEllipsoid(SamplerNested* sampler)
	: NestedSamplingStrategy(sampler)
	, enlargement(1.1)
	, full_update_steps(population_size/4)
	, steps_since_update(0)
{
	mean.resize(num_variables);
	covariance.resize(num_variables, num_variables);
	transformation.resize(num_variables, num_variables);
}

NestedSamplingStrategyEllipsoid::~NestedSamplingStrategyEllipsoid()
{
}

bool NestedSamplingStrategyEllipsoid::Initialize(Real prior_volume_estimate, RNG* rng)
{
	// Convert samples from original space to unit hypercube
	bool result = true;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		result &= bcm::NestedSamplingUtils::ConvertToUnitSpace((*si)->variable_values, (*si)->unit_values, prior);
	}
	if (!result) {
		return false;
	}

	// Calculate initial bounding ellipsoid
	result = CalculateBounds();
	if (!result) {
		return false;
	}

	steps_since_update = 0;

	return true;
}

bool NestedSamplingStrategyEllipsoid::UpdateRemoveSample(const Sample* removed_sample, Real prior_volume_estimate, bool& reset, RNG* rng)
{
	if (steps_since_update == full_update_steps) {
		// Calculate new ellipsoid
		bool result = CalculateBounds();
		reset = true;
		if (!result) {
			return false;
		}
		steps_since_update = 0;
	} else {
		steps_since_update++;
	}

	return true;
}

bool NestedSamplingStrategyEllipsoid::UpdateAddSample(const Sample* added_sample, RNG* rng)
{
	// Convert the new sample from original space to unit hypercube
	bool result = bcm::NestedSamplingUtils::ConvertToUnitSpace(added_sample->variable_values, added_sample->unit_values, prior);
	if (!result) {
		return false;
	}

	// Scale existing ellipsoid to make sure the new point is inside.
	// It might be outside because the sampling ellipsoid is "enlargement" bigger than the bounding ellipsoid
	VectorReal centered = added_sample->unit_values - mean;
	VectorReal optimize(num_variables);
	for (size_t j = 0; j < num_variables; j++) {
		optimize(j) = 0;
		for (size_t k = 0; k < num_variables; k++) {
			optimize(j) += inv_covariance(j,k) * centered(k);
		}
	}
	Real k = sqrt(optimize.dot(centered));
	scaling = std::max(scaling, k);

	return true;
}

bool NestedSamplingStrategyEllipsoid::GetNewSample(Real min_llh, Sample* sample, RNG* rng, size_t threadix) const
{
	VectorReal newsample(num_variables);

	// Generate a new sample from the bounding ellipsoid
	while (1) {
		// Sample from unit sphere
		bcm::NestedSamplingUtils::GeneratePointInUnitSphere(newsample, rng);

		// Transform to the bounding ellipsoid
		newsample = scaling * enlargement * transformation * newsample + mean;
		
		// Make sure it's inside the unit hypercube
#if 1
		bool in_unit_hypercube = true;
		for (size_t i = 0; i < num_variables; i++) {
			if (newsample(i) < 0.0 || newsample(i) > 1.0) {
				in_unit_hypercube = false;
				break;
			}
		}
		if (in_unit_hypercube) {
			break;
		}
#else
		for (size_t i = 0; i < num_variables; i++) {
			Real nx = newsample(i);
			while (1) {
				if (nx < 0.0) {
					nx = -nx;
				} else if (nx > 1.0) {
					nx = 1.0 - (nx - 1.0);
				} else {
					break;
				}
			}
			newsample(i) = nx;
		}
		break;
#endif
	}

	// Convert from unit hypercube back to the original space
	bcm::NestedSamplingUtils::ConvertToOriginalSpace(newsample, sample->variable_values, prior);

	if (!EvaluatePriorLikelihood(threadix, sample->variable_values.data(), sample->lprior, sample->llh)) {
		return false;
	}
	if (sample->lprior == -std::numeric_limits<Real>::infinity()) {
		sample->lposterior = sample->lprior;
	} else {
		if (!GetOutputValues(threadix, sample->output_values.data())) {
			return false;
		}
		sample->lposterior = sample->lprior + sample->llh;
	}
	return true;
}

bool NestedSamplingStrategyEllipsoid::CalculateBounds()
{
	// Calculate mean
	mean.setZero();
	size_t count = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		const VectorReal& s = (*si)->unit_values;
		Real one_over_kp1 = 1.0 / (count + 1.0);
		for (size_t i = 0; i < num_variables; i++) {
			mean(i) += (s(i) - mean(i)) * one_over_kp1;
		}
		count++;
	}
	
	// Calculate covariance matrix
	covariance.setZero();
	count = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		const VectorReal& s = (*si)->unit_values;
		Real one_over_kp1 = 1.0 / (count + 1.0);
		for (size_t i = 0; i < num_variables; i++) {
			for (size_t j = 0; j < num_variables; j++) {
				Real d1 = s(i) - mean(i);
				Real d2 = s(j) - mean(j);
				covariance(i,j) += (d1 * d2 - covariance(i,j)) * one_over_kp1;
			}
		}
		count++;
	}
	covariance *= count / ((Real)count - 1.0);

	// Calculate ellipsoid orientation
	MatrixReal eigvec, eigval;
	bcm::NestedSamplingUtils::EigenDecomposition(covariance, eigval, eigvec, inv_covariance);

	// Calculate ellipsoid scaling so that it encloses all points
	VectorReal centered(num_variables);
	VectorReal optimize(num_variables);
	scaling = 0.0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		const VectorReal& s = (*si)->unit_values;
		centered = s - mean;
		
#if 1
		// Optimized to prevent memory allocation for the transpose.
		for (size_t i = 0; i < num_variables; i++) {
			optimize(i) = 0;
			for (size_t j = 0; j < num_variables; j++) {
				optimize(i) += inv_covariance(i,j) * centered(j);
			}
		}
		Real d = optimize.dot(centered);
#else
		Real d = s.transpose() * inv_cov * s->values;
#endif

		if (d < 9.0) {
			scaling = std::max(scaling, d);
		} else {
			// More than 3 standard deviations away - ignore as outlier
		}
	}
	//scaling = sqrt(scaling);

	// Finally the transformation matrix
	for (size_t i = 0; i < num_variables; i++) {
		transformation.col(i) = eigvec.col(i) * sqrt(eigval(i,i));
	}

	return true;
}

}
