#include "Utils.h"
#include "Prior.h"
#include "NestedSamplingStrategyMCMC.h"

namespace bcm {

static const Real target_acceptance_rate = (Real)0.234;

NestedSamplingStrategyMCMC::NestedSamplingStrategyMCMC(SamplerNested* sampler, size_t numsteps)
	: NestedSamplingStrategy(sampler)
	, steps(numsteps)
	, empirical_mean(VectorReal::Zero(num_variables))
	, empirical_variance(VectorReal::Zero(num_variables))
	, proposal_sigma(VectorReal::Zero(num_variables))
	, removed_ix(0)
	, proposal_scale(1.0)
	, current_acceptance_rate_ema(target_acceptance_rate)
{
}

NestedSamplingStrategyMCMC::~NestedSamplingStrategyMCMC()
{
}

bool NestedSamplingStrategyMCMC::Initialize(Real prior_volume_estimate, RNG* rng)
{
	bool reset = false;
	return UpdateRemoveSample(NULL, prior_volume_estimate, reset, rng);
}

bool NestedSamplingStrategyMCMC::UpdateRemoveSample(const Sample* removed_sample, Real prior_volume_estimate, bool& reset, RNG* rng)
{
	// Calculate empirical mean
	empirical_mean.setZero();
	size_t k = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si, ++k) {
		const Sample* s = *si;
		Real one_over_kp1 = (Real)1.0 / (k + (Real)1.0);
		for (size_t i = 0; i < num_variables; i++) {
			empirical_mean(i) += (s->variable_values(i) - empirical_mean(i)) * one_over_kp1;
		}
	}

	// Calculate empirical variance
	empirical_variance.setZero();
	k = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si, ++k) {
		const Sample* s = *si;
		Real one_over_kp1 = (Real)1.0 / (k + (Real)1.0);
		for (size_t i = 0; i < num_variables; i++) {
			Real d = s->variable_values(i) - empirical_mean(i);
			empirical_variance(i) += (d * d - empirical_variance(i)) * one_over_kp1;
		}
	}
	Real n = (Real)sampling_population.size();
	empirical_variance *= n / (n - 1.0);
	
	// Set proposal variance for the next iteration
	for (size_t i = 0; i < num_variables; i++) {
		Real prior_var;
		if (!prior->EvaluateMarginalVariance(i, prior_var)) {
			// TODO - somehow come up with something reasonable?
			prior_var = 1.0;
		}
		Real proposal_var = std::max(empirical_variance(i), (Real)1e-6 * prior_var);
		proposal_sigma(i) = sqrt(proposal_var) * (Real)2.4 * rsqrt((Real)num_variables);
	}

	return true;
}

bool NestedSamplingStrategyMCMC::GetNewSample(Real min_llh, Sample* sample, RNG* rng, size_t threadix)  const
{
	std::list<Sample*>::const_iterator si = sampling_population.GetRandomSample(rng);
	*sample = **si;

	VectorReal new_sample = sample->variable_values;

	for (size_t j = 0; j < steps; j++) {
		// Update scale
		Real learn_rate = 1.0 + rng->GetReal() * 0.005;
		if (current_acceptance_rate_ema < 0.952381 * target_acceptance_rate) {
			proposal_scale /= learn_rate;
			proposal_scale = std::max(proposal_scale, (Real)1e-10);
		} else if (current_acceptance_rate_ema > 1.05 * target_acceptance_rate) {
			proposal_scale *= learn_rate;
			proposal_scale = std::min(proposal_scale, (Real)1e3);
		}

		// Generate new sample
		for (size_t i = 0; i < num_variables; i++) {
			Real x = rng->GetNormal(0.0, proposal_sigma(i));
			Real nx = sample->variable_values(i) + x;

			// Reflect for variables with bounded support, this is still a symmetrical proposal
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}
			new_sample(i) = nx;
		}

		Real new_lprior;
		Real new_llh;
		if (!EvaluatePriorLikelihood(threadix, new_sample.data(), new_lprior, new_llh)) {
			return false;
		}

		bool accept = true;
		
		// Make sure the prior density is non-zero; in case the random sampling from the prior is inaccurate.
		if (new_lprior == -std::numeric_limits<Real>::infinity()) {
			accept = false;
		} else if (new_llh <= min_llh) {
			accept = false;
		} else {
			Real transition_probability = new_lprior - sample->lprior;
			transition_probability = exp(transition_probability);
			transition_probability = std::min((Real)1.0, transition_probability);
	
			Real alpha = rng->GetReal();
			if (alpha >= transition_probability) {
				accept = false;
			}
		}
		
		if (accept) {
			// Keep new state
			sample->variable_values = new_sample;
			GetOutputValues(threadix, sample->output_values.data());
			sample->llh = new_llh;
			sample->lprior = new_lprior;
			sample->lposterior = sample->llh + sample->lprior;

			// Update acceptance rate
			current_acceptance_rate_ema += (1.0 - current_acceptance_rate_ema) * 2.0 / (1001.0);
		} else {
			// Revert to previous state

			// Update acceptance rate
			current_acceptance_rate_ema += (0.0 - current_acceptance_rate_ema) * 2.0 / (1001.0);
		}
	}
	
	return true;
}

}
