#include "Utils.h"
#include "BivariateGMM.h"
#include "PriorBivariateMarginalGMM.h"
#include "VariableSet.h"

#include <boost/property_tree/xml_parser.hpp>

namespace bcm {

PriorBivariateMarginalGMM::PriorBivariateMarginalGMM(VariableSet* varset)
	: PriorMarginalGMM(varset)
{
}

PriorBivariateMarginalGMM::~PriorBivariateMarginalGMM()
{
}

bool PriorBivariateMarginalGMM::LoadFromXML(const std::string& filename)
{
	if (!PriorMarginalGMM::LoadFromXML(filename)) {
		return false;
	}

	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(filename, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading variable file: %s", e.what());
		return false;
	}
	
	try {
		boost::property_tree::ptree model_node = pt.get_child("variableset");
		
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, model_node.get_child("")) {
			if (var.first == "bivariate") {
				std::string var1 = var.second.get<std::string>("<xmlattr>.var1");
				std::string var2 = var.second.get<std::string>("<xmlattr>.var2");
				
				size_t ix1 = varset->GetVariableIndex(var1);
				if (ix1 == std::numeric_limits<size_t>::max()) {
					LOGERROR("Unknown variable \"%s\"", var1.c_str());
					return false;
				}
				
				size_t ix2 = varset->GetVariableIndex(var2);
				if (ix2 == std::numeric_limits<size_t>::max()) {
					LOGERROR("Unknown variable \"%s\"", var2.c_str());
					return false;
				}

				boost::optional<std::string> transform1 = var.second.get_optional<std::string>("<xmlattr>.transform1");
				ETransform t1 = Transform_None;
				if (transform1) {
					if (*transform1 == "log") {
						t1 = Transform_Log;
					} else if (*transform1 == "logit") {
						t1 = Transform_Logit;
					} else if (*transform1 == "none") {
						t1 = Transform_None;
					} else {
						LOGERROR("Invalid transform \"%s\"", transform1->c_str());
						return false;
					}
				}

				boost::optional<std::string> transform2 = var.second.get_optional<std::string>("<xmlattr>.transform2");
				ETransform t2 = Transform_None;
				if (transform2) {
					if (*transform2 == "log") {
						t2 = Transform_Log;
					} else if (*transform2 == "logit") {
						t2 = Transform_Logit;
					} else if (*transform2 == "none") {
						t2 = Transform_None;
					} else {
						LOGERROR("Invalid transform \"%s\"", transform2->c_str());
						return false;
					}
				}

				BivariateGMM* gmm = new BivariateGMM;
				if (!gmm->Initialize(var.second)) {
					return false;
				}
				BivariateGMMs.push_back(gmm);
				BivariateGMM_Indices.push_back(std::pair<size_t, size_t>(ix1, ix2));
				BivariateGMM_Transforms.push_back(std::pair<ETransform, ETransform>(t1, t2));
			}
		}

		return true;
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing variable file: %s", e.what());
		return false;
	}
}

bool PriorBivariateMarginalGMM::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	if (!PriorMarginalGMM::EvaluateLogPDF(threadix, values, logp)) {
		return false;
	}

	if (logp == -std::numeric_limits<Real>::infinity()) {
		return true;
	}

	for (size_t i = 0; i < BivariateGMMs.size(); i++) {
		size_t ix1 = BivariateGMM_Indices[i].first;
		size_t ix2 = BivariateGMM_Indices[i].second;

		Real x = values[ix1];
		Real y = values[ix2];

		ETransform t1 = BivariateGMM_Transforms[i].first;
		if (t1 == Transform_Log) {
			if (x <= 0.0) {
				logp = -std::numeric_limits<Real>::infinity();
				break;
			}
			x = log(x);
		} else if (t1 == Transform_Logit) {
			if (x <= 0.0 || x >= 1.0) {
				logp = -std::numeric_limits<Real>::infinity();
				break;
			}
			x = log(x / (1.0 - x));
		}
		ETransform t2 = BivariateGMM_Transforms[i].second;
		if (t2 == Transform_Log) {
			if (y <= 0.0) {
				logp = -std::numeric_limits<Real>::infinity();
				break;
			}
			y = log(y);
		} else if (t2 == Transform_Logit) {
			if (y <= 0.0 || y >= 1.0) {
				logp = -std::numeric_limits<Real>::infinity();
				break;
			}
			y = log(y / (1.0 - y));
		}

		Real ljointp, lp1, lp2;
		if (!BivariateGMMs[i]->EvaluateLogPDF(ljointp, x, y)) {
			return false;
		}
	
		// Transformed PDF = PDF of transform times the absolute value of the derivative of the transformation
		// For a multivariate PDF it's the determinant of the jacobian, which in this case is just the product of the two transformations
		if (t1 == Transform_Log) {
			ljointp -= x;
		} else if (t1 == Transform_Logit) {
			ljointp -= log(values[ix1] * (1.0 - values[ix1]));
		}
		if (t2 == Transform_Log) {
			ljointp -= y;
		} else if (t2 == Transform_Logit) {
			ljointp -= log(values[ix2] * (1.0 - values[ix2]));
		}

		// Divide by PDF of the marginals
		if (!EvaluateMarginalLogPDF(ix1, values[ix1], lp1)) {
			return false;
		}
		if (!EvaluateMarginalLogPDF(ix2, values[ix2], lp2)) {
			return false;
		}

		logp += ljointp - lp1 - lp2;
	}
	
	return true;
}

bool PriorBivariateMarginalGMM::Sample(Real* values, RNG* rng) const
{
	return PriorMarginalGMM::Sample(values, rng);
}

bool PriorBivariateMarginalGMM::EvaluateMarginalVariance(size_t i, Real& var) const
{
	return PriorMarginalGMM::EvaluateMarginalVariance(i, var);
}

}
