#include "Utils.h"
#include "PriorGMM.h"
#include "RNG.h"
#include "VariableSet.h"

#define USE_TRANFORMS 1

namespace bcm {

PriorGMM::PriorGMM(VariableSet* varset)
	: Prior(varset)
	, num_components(0)
{
}

PriorGMM::~PriorGMM()
{
}

bool PriorGMM::LoadFromXML(const boost::property_tree::ptree& xml_node)
{
	try {
		num_components = xml_node.count("component");

		weights = VectorReal::Constant(num_components, std::numeric_limits<Real>::quiet_NaN());
		means.resize(num_components, VectorReal::Constant(varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN()));
		covariances.resize(num_components, MatrixReal::Constant(varset->GetNumVariables(), varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN()));

		size_t i = 0;
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "component") {
				weights(i) = var.second.get<Real>("<xmlattr>.weight");

				std::string mean_str = var.second.get<std::string>("<xmlattr>.mean");
				std::vector<std::string> mean_vec;
				bcm::tokenize(mean_str, mean_vec, ";");
				if (mean_vec.size() != varset->GetNumVariables()) {
					LOGERROR("Dimension of means vector does not match the variable set");
					return false;
				}
				for (size_t j = 0; j < varset->GetNumVariables(); j++) {
					means[i](j) = boost::lexical_cast<Real>(mean_vec[j]);
				}
				
				std::string cov_str = var.second.get<std::string>("<xmlattr>.covariance");
				std::vector<std::string> cov_vec;
				bcm::tokenize(cov_str, cov_vec, ";");
				if (cov_vec.size() != varset->GetNumVariables()) {
					LOGERROR("Dimension of covariance matrix does not match the variable set");
					return false;
				}
				for (size_t j = 0; j < varset->GetNumVariables(); j++) {
					std::vector<std::string> cov_vec2;
					bcm::tokenize(cov_vec[j], cov_vec2, ",");
					if (cov_vec2.size() != varset->GetNumVariables()) {
						LOGERROR("Dimension of covariance matrix does not match the variable set");
						return false;
					}
					
					for (size_t k = 0; k < varset->GetNumVariables(); k++) {
						covariances[i](j,k) = boost::lexical_cast<Real>(cov_vec2[k]);
					}
				}

				i++;
			}
		}

		ASSERT(i == num_components);
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing prior file: %s", e.what());
		return false;
	}

	if (!LoadTransformationsFromXML(xml_node)) {
		return false;
	}

	llt.resize(num_components);
	log_normalizing_constants = VectorReal::Constant(num_components, std::numeric_limits<Real>::quiet_NaN());
	log_weights = VectorReal::Constant(num_components, std::numeric_limits<Real>::quiet_NaN());

	for (size_t i = 0; i < num_components; i++) {
		llt[i].compute(covariances[i]);
		log_normalizing_constants(i) = -0.5 * (varset->GetNumVariables() * log(2.0 * M_PI) + log(covariances[i].determinant()));
		log_weights(i) = log(weights(i));
	}

	return true;
}

bool PriorGMM::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	size_t N = varset->GetNumVariables();
	VectorReal v(N);
	
#if USE_TRANFORMS
	// Apply transformations
	if (!ApplyTransformations(values, v)) {
		logp = -std::numeric_limits<Real>::infinity();
		return true;
	}
#else
	for (size_t i = 0; i < N; i++) {
		v(i) = values[i];
	}
#endif
	
	logp = -std::numeric_limits<Real>::infinity();
	VectorReal x(N);
	for (size_t i = 0; i < num_components; i++) {
		// Evaluate multivariate normal density
		x = v - means[i];
		llt[i].matrixL().solveInPlace(x);
		Real p = log_normalizing_constants(i) - 0.5 * x.dot(x);

		// Increment total according to this component's weight
		logp = bcm::logsum(logp, p + log_weights(i));
	}

#if USE_TRANFORMS
	// Correct for transformations
	if (!CorrectProbabilityForTransform(logp, values, v)) {
		return false;
	}
#else
	for (size_t i = 0; i < N; i++) {
		if (transforms[i] == Transform_Log) {
			if (values[i] <= 0) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		} else if (transforms[i] == Transform_Logit) {
			if (values[i] <= 0 || values[i] >= 1.0) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		} else if (transforms[i] == Transform_LogitScale) {
			const Real a = transform_parameters[i](0);
			const Real b = transform_parameters[i](1);
			if (values[i] <= a || values[i] >= b) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		}
	}
#endif

	return true;
}

bool PriorGMM::Sample(Real* values, RNG* rng) const
{
	bool accept = false;

	while (!accept) {
		// Pick one of the components according to their weights
		Real p = rng->GetReal();
		size_t component_ix = num_components - 1;
		Real t = 0;
		for (size_t i = 0; i < num_components - 1; i++) {
			t += weights(i);
			if (p < t) {
				component_ix = i;
				break;
			}
		}

		// Sample from that component
		size_t N = varset->GetNumVariables();
		Eigen::Map<VectorReal> v(values, N);
		for (size_t i = 0; i < N; i++) {
			v(i) = rng->GetNormal(0.0, 1.0);
		}
		v = llt[component_ix].matrixL() * v;
		v += means[component_ix];
		
		accept = true;
#if !USE_TRANFORMS
		for (size_t i = 0; i < N; i++) {
			if (transforms[i] == Transform_Log) {
				if (values[i] <= 0) {
					accept = false;
					break;
				}
			} else if (transforms[i] == Transform_Logit) {
				if (values[i] <= 0 || values[i] >= 1.0) {
					accept = false;
					break;
				}
			} else if (transforms[i] == Transform_LogitScale) {
				const Real a = transform_parameters[i](0);
				const Real b = transform_parameters[i](1);
				if (values[i] <= a || values[i] >= b) {
					accept = false;
					break;
				}
			}
		}
#endif
	}

#if USE_TRANFORMS
	// Apply inverse transformations
	if (!ApplyInverseTransformations(values)) {
		return false;
	}
#endif

	return true;
}

bool PriorGMM::EvaluateMarginalVariance(size_t i, Real& var) const
{
	// TODO - this does not take variable transformations into account
	var = 0.0;
	for (size_t j = 0; j < num_components - 1; j++) {
		var += weights(j) * weights(j) * covariances[j](i,i);
	}
	return true;
}

}
