#include "Utils.h"
#include "PriorKDE.h"
#include "RNG.h"
#include "VariableSet.h"

#define USE_TRANFORMS 0

namespace bcm {

PriorKDE::PriorKDE(VariableSet* varset)
	: Prior(varset)
{
}

PriorKDE::~PriorKDE()
{
}

bool PriorKDE::LoadFromXML(const boost::property_tree::ptree& xml_node)
{
	try {
		num_points = xml_node.count("point");

		means = MatrixReal::Constant(num_points, varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN());
		bandwidths = VectorReal::Constant(varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN());
		inv_bandwidths = VectorReal::Constant(varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN());

		size_t i = 0;
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "point") {
				std::string mean_str = var.second.get<std::string>("<xmlattr>.mean");
				std::vector<std::string> mean_vec;
				bcm::tokenize(mean_str, mean_vec, ";");
				if (mean_vec.size() != varset->GetNumVariables()) {
					LOGERROR("Dimension of means vector does not match the variable set");
					return false;
				}
				for (size_t j = 0; j < varset->GetNumVariables(); j++) {
					means(i,j) = boost::lexical_cast<Real>(mean_vec[j]);
				}
				i++;
			}
		}
		ASSERT(i == num_points);

		log_normalizing_constant = 0.0;

		const boost::property_tree::ptree& node = xml_node.get_child("bandwidth");
		std::string bw_str = node.get<std::string>("<xmlattr>.bandwidth");
		std::vector<std::string> bw_vec;
		bcm::tokenize(bw_str, bw_vec, ";");
		if (bw_vec.size() != varset->GetNumVariables()) {
			LOGERROR("Dimension of bandwidth vector does not match the variable set");
			return false;
		}
		for (size_t j = 0; j < varset->GetNumVariables(); j++) {
			bandwidths(j) = boost::lexical_cast<Real>(bw_vec[j]);
			inv_bandwidths(j) = -1.0 / (2.0 * bandwidths(j) * bandwidths(j));
			log_normalizing_constant -= log(bandwidths(j) * sqrt(2.0 * M_PI));
		}
		log_normalizing_constant -= log((Real)num_points);

	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing prior file: %s", e.what());
		return false;
	}

	if (!LoadTransformationsFromXML(xml_node)) {
		return false;
	}

	return true;
}

bool PriorKDE::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	size_t N = varset->GetNumVariables();
	VectorReal v(N);
	
#if USE_TRANFORMS
	// Apply transformations
	if (!ApplyTransformations(values, v)) {
		logp = -std::numeric_limits<Real>::infinity();
		return true;
	}
#else
	for (size_t i = 0; i < N; i++) {
		v(i) = values[i];
	}
#endif

#if 0
	logp = -std::numeric_limits<Real>::infinity();
	for (size_t i = 0; i < num_points; i++) {
		Real thislogp = 0.0;
		for (size_t j = 0; j < N; j++) {
			Real d = v(j) - means(i,j);
			Real p = d * d * inv_bandwidths(j);
			thislogp += p;
		}

		logp = bcm::logsum(logp, thislogp);
	}
	logp += log_normalizing_constant;
#else
	logp = 0.0;
	const Real nc = pow(2 * M_PI, (Real)N / 2.0);
	for (size_t i = 0; i < num_points; i++) {
		Real dp = 1.0;
		Real det = 1.0;
		for (size_t j = 0; j < N; j++) {
			det *= bandwidths(j);
			const Real d = v(j) - means(i,j);
			dp *= exp(d * d * inv_bandwidths(j));
		}
		logp += dp * nc * bcm::rsqrt(det);
	}
	logp = log(logp) - log((Real)num_points);
#endif
	
#if USE_TRANFORMS
	// Correct for transformations
	if (!CorrectProbabilityForTransform(logp, values, v)) {
		return false;
	}
#else
	for (size_t i = 0; i < N; i++) {
		if (transforms[i] == Transform_Log) {
			if (values[i] <= 0) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		} else if (transforms[i] == Transform_Logit) {
			if (values[i] <= 0 || values[i] >= 1.0) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		} else if (transforms[i] == Transform_LogitScale) {
			const Real a = transform_parameters[i](0);
			const Real b = transform_parameters[i](1);
			if (values[i] <= a || values[i] >= b) {
				logp = -std::numeric_limits<Real>::infinity();
			}
		}
	}
#endif

	return true;
}

bool PriorKDE::Sample(Real* values, RNG* rng) const
{
	bool accept = false;

	while (!accept) {
		// Select a point
		size_t ix = rng->GetUniformInt((unsigned long)num_points);

		// Generate from the kernel
		for (size_t j = 0; j < varset->GetNumVariables(); j++) {
			values[j] = rng->GetNormal(means(ix,j), bandwidths(j));
		}

		accept = true;
#if !USE_TRANFORMS
		for (size_t i = 0; i < varset->GetNumVariables(); i++) {
			if (transforms[i] == Transform_Log) {
				if (values[i] <= 0) {
					accept = false;
					break;
				}
			} else if (transforms[i] == Transform_Logit) {
				if (values[i] <= 0 || values[i] >= 1.0) {
					accept = false;
					break;
				}
			} else if (transforms[i] == Transform_LogitScale) {
				const Real a = transform_parameters[i](0);
				const Real b = transform_parameters[i](1);
				if (values[i] <= a || values[i] >= b) {
					accept = false;
					break;
				}
			}
		}
#endif
	}
	
#if USE_TRANFORMS
	// Apply inverse transformations
	if (!ApplyInverseTransformations(values)) {
		return false;
	}
#endif

	return true;
}

bool PriorKDE::EvaluateMarginalVariance(size_t i, Real& var) const
{
	// TODO - this does not take variable transformations into account
	var = num_points * bandwidths(i) * bandwidths(i);
	return true;
}

}
