#include "Utils.h"
#include "PriorVineCopula.h"
#include "ProbabilityDistributions.h"
#include "RNG.h"
#include "VariableSet.h"
#include "VectorUtils.h"

#define SINGLE_VALUE_EVALUATION 1

#if PLATFORM_WINDOWS
// For now this is only implemented on windows -- too painful to support multiple compilers

#ifndef NAN
#define NAN (std::numeric_limits<Real>::quiet_NaN())
#endif

#ifndef INFINITY
#define INFINITY (std::numeric_limits<Real>::infinity())
#endif

#if PLATFORM_WINDOWS
#if _MSC_VER <= 1600
#define isfinite(x) _finite(x)
#define isnan(x) _isnan(x)
#include <boost/math/special_functions/expm1.hpp>
#include <boost/math/special_functions/trunc.hpp>
#include <boost/math/special_functions/round.hpp>
#include <boost/math/special_functions/gamma.hpp>
inline Real expm1(Real x) { return boost::math::expm1(x); }
inline Real log1p(Real x) { return boost::math::log1p(x); }
inline Real round(Real x) { return boost::math::round(x); }
inline Real nearbyint(Real x) { return boost::math::round(x); }
inline Real trunc(Real x) { return boost::math::trunc(x); }
inline Real lgamma(Real x) { return boost::math::lgamma(x); }
#endif
#else
using std::isfinite;
using std::isnan;
#define TRUE true
#define FALSE false
#endif

#include "../../dependencies/VineCopula-2.1.0/src/evCopula.c"
#include "../../dependencies/VineCopula-2.1.0/src/hfunc.c"
#include "../../dependencies/VineCopula-2.1.0/src/likelihood.c"

#undef max

namespace bcm {

PriorVineCopula::PriorVineCopula(VariableSet* varset)
	: Prior(varset)
{
}

PriorVineCopula::~PriorVineCopula()
{
}

bool PriorVineCopula::LoadFromXML(const boost::property_tree::ptree& xml_node)
{
	size_t N = varset->GetNumVariables();
	marginals.resize(N);

	try {
		bool result = true;

		// Load marginals
		size_t i = 0;
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "marginal") {
				std::string name = var.second.get<std::string>("<xmlattr>.name");
				size_t ix = varset->GetVariableIndex(name);
				Marginal& m = marginals[ix];
				
				std::string type = var.second.get<std::string>("<xmlattr>.type");
				if (type == "ecdf") {
					m.type = MT_ECDF_KDE;
					m.param1 = VectorReal::Zero(1);
					result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.bw"), m.param1);
					result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.x"), m.param2);

					// Lexical cast properly converts "inf"; which var.second.get<Real> doesn't do.
					std::string minstr = var.second.get<std::string>("<xmlattr>.min");
					m.min = boost::lexical_cast<Real>(minstr);

					std::string maxstr = var.second.get<std::string>("<xmlattr>.max");
					m.max = boost::lexical_cast<Real>(maxstr);

					LowerBounds[ix] = m.min;
					UpperBounds[ix] = m.max;
				} else {
					result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.p"), m.p);
					if (type == "normal") {
						m.type = MT_Normal;
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.mu"), m.param1);
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.sigma"), m.param2);
						if (m.param1.size() != m.p.size() || m.param2.size() != m.p.size()) {
							LOGERROR("Inconsistent marginal parameters for variable \"%s\"", name.c_str());
							return false;
						}
					} else if (type == "beta") {
						m.type = MT_Beta;
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.a"), m.param1);
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.b"), m.param2);
						if (m.param1.size() != m.p.size() || m.param2.size() != m.p.size()) {
							LOGERROR("Inconsistent marginal parameters for variable \"%s\"", name.c_str());
							return false;
						}
						m.min = var.second.get<Real>("<xmlattr>.min");
						m.max = var.second.get<Real>("<xmlattr>.max");
						LowerBounds[ix] = m.min;
						UpperBounds[ix] = m.max;
					} else if (type == "gamma") {
						m.type = MT_Gamma;
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.shape"), m.param1);
						result &= ParseVectorFromString(var.second.get<std::string>("<xmlattr>.scale"), m.param2);
						if (m.param1.size() != m.p.size() || m.param2.size() != m.p.size()) {
							LOGERROR("Inconsistent marginal parameters for variable \"%s\"", name.c_str());
							return false;
						}

						LowerBounds[ix] = 0.0;
					} else {
						LOGERROR("Unknown marginal type \"%s\" for variable \"%s\"", type.c_str(), name.c_str());
						return false;
					}
				}

				i++;
			}
		}
		if (i != N) {
			LOGERROR("Not all marginals have been specified");
			return false;
		}

		// Load vine copula
		const boost::property_tree::ptree& node = xml_node.get_child("RVineCopula");
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.matrix"), M, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.family"), family, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.maxmat"), MaxMat, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.codirect"), CoDirect, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.coindirect"), CoIndirect, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.par"), par, N);
		result &= ParseMatrixFromString(node.get<std::string>("<xmlattr>.par2"), par2, N);
		if (!result) {
			return false;
		}

		std::string names_str = node.get<std::string>("<xmlattr>.names");
		std::vector<std::string> names_vec;
		bcm::tokenize(names_str, names_vec, ";");
		if (names_vec.size() != N) {
			LOGERROR("Insufficient names");
			return false;
		}
		for (size_t i = 0; i < names_vec.size(); i++) {
			size_t varix = varset->GetVariableIndex(names_vec[i]);
			if (varix == std::numeric_limits<size_t>::max()) {
				LOGERROR("Unknown variable name in copula");
				return false;
			}
			marginals[varix].copula_ix = i;
		}

	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing prior file: %s", e.what());
		return false;
	}

	// For the random sample generation
	rpar = Eigen::MatrixXd(N, N);
	rpar2 = Eigen::MatrixXd(N, N);
	rMaxMat = Eigen::MatrixXi(N, N);
	rM = Eigen::MatrixXi(N, N);
	rCoIndirect = Eigen::MatrixXi(N, N);
	rfamily = Eigen::MatrixXi(N, N);
	for (size_t i = 0; i < N; i++) {
		for (size_t j = 0; j < N; j++) {
			rpar(N - i - 1, N - j - 1) = par(i, j);
			rpar2(N - i - 1, N - j - 1) = par2(i, j);
			rMaxMat(N - i - 1, N - j - 1) = MaxMat(i, j);
			rM(N - i - 1, N - j - 1) = M(i, j);
			rCoIndirect(N - i - 1, N - j - 1) = CoIndirect(i, j);
			rfamily(N - i - 1, N - j - 1) = family(i, j);
		}
	}

	return true;
}

bool PriorVineCopula::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	size_t N = varset->GetNumVariables();
	VectorReal v = VectorReal::Constant(N, 0);

	// Transform to U(0,1)
	for (size_t i = 0; i < N; i++) {
		const Marginal& m = marginals[i];
		size_t ix = m.copula_ix;

		switch (m.type) {
		case MT_Normal:
			for (int j = 0; j < m.p.size(); j++) {
				v(ix) += m.p(j) * bcm::CdfNormal(values[i], m.param1(j), m.param2(j));
			}
			break;

		case MT_Beta:
			for (int j = 0; j < m.p.size(); j++) {
				Real val = (values[i] - m.min) / (m.max - m.min);
				v(ix) += m.p(j) * bcm::CdfBeta(val, m.param1(j), m.param2(j));
			}
			break;

		case MT_Gamma:
			for (int j = 0; j < m.p.size(); j++) {
				v(ix) += m.p(j) * bcm::CdfGamma(values[i], m.param1(j), m.param2(j));
			}
			break;
			
		case MT_ECDF_KDE:
			{
				//if (values[i] < m.min || values[i] > m.max) {
				if (values[i] < m.param2(0) || values[i] > m.param2(m.param2.size()-1)) {
					logp = -std::numeric_limits<Real>::infinity();
					return true;
				}

				// TODO - optimize to O(log(n))
				Real x = 0;
				for (int j = 0; j < m.param2.size(); j++) {
					if (m.param2(j) <= values[i]) {
						x += 1.0;
					} else {
						break;
					}
				}
				if (x == 1000.0) {
					int a;
					a = 6;
				}
				v(ix) = x / (Real)m.param2.size();
			}
			break;

		default:
			LOGERROR("Unknown marginal type");
			return false;
		}
	}

	// Evaluate vine copula
	MatrixReal vdirect(N, N);
	MatrixReal vindirect(N, N);
	for (size_t i = 0; i < N; i++) {
		vdirect(N - 1, i) = v(N - 1 - i);
	}
	logp = 0.0;
	for (int i = (int)N - 2; i > -1; i--) {
		for (int k = (int)N - 1; k > i; k--) {
			int m = MaxMat(k, i);
			int kk = 1;

			Real ll;
			if (m == M(k, i)) {
				LL_mod2(const_cast<int*>(&family(k,i)), &kk, &vdirect(k,N-m), &vdirect(k,i), const_cast<double*>(&par(k,i)), const_cast<double*>(&par2(k,i)), &ll);
				if (CoDirect(k-1, i) == 1) {
					Hfunc1(const_cast<int*>(&family(k,i)), &kk, &vdirect(k,i), &vdirect(k, N - m), const_cast<double*>(&par(k,i)), const_cast<double*>(&par2(k,i)), &vdirect(k - 1,i));
				}
				if (CoIndirect(k-1, i) == 1) {
					Hfunc2(const_cast<int*>(&family(k, i)), &kk, &vdirect(k, N - m), &vdirect(k, i), const_cast<double*>(&par(k, i)), const_cast<double*>(&par2(k, i)), &vindirect(k - 1, i));
				}
			}
			else {
				LL_mod2(const_cast<int*>(&family(k, i)), &kk, &vindirect(k, N - m), &vdirect(k, i), const_cast<double*>(&par(k, i)), const_cast<double*>(&par2(k, i)), &ll);
				if (CoDirect(k-1, i) == 1) {
					Hfunc1(const_cast<int*>(&family(k, i)), &kk, &vdirect(k, i), &vindirect(k, N - m), const_cast<double*>(&par(k, i)), const_cast<double*>(&par2(k, i)), &vdirect(k - 1, i));
				}
				if (CoIndirect(k-1, i) == 1) {
					Hfunc2(const_cast<int*>(&family(k, i)), &kk, &vindirect(k, N - m), &vdirect(k, i), const_cast<double*>(&par(k, i)), const_cast<double*>(&par2(k, i)), &vindirect(k - 1, i));
				}
			}
			logp += ll;
		}
	}

	// Correct for variable transformation
	for (size_t i = 0; i < N; i++) {
		const Marginal& m = marginals[i];
		size_t ix = m.copula_ix;

		Real dp = 0.0;
		switch (m.type) {
		case MT_Normal:
			for (int j = 0; j < m.p.size(); j++) {
				dp += m.p(j) * bcm::PdfNormal(values[i], m.param1(j), m.param2(j));
			}
			break;

		case MT_Beta:
			for (int j = 0; j < m.p.size(); j++) {
				Real val = (values[i] - m.min) / (m.max - m.min);
				dp += m.p(j) * bcm::PdfBeta(val, m.param1(j), m.param2(j));
			}
			dp /= (m.max - m.min);
			break;

		case MT_Gamma:
			for (int j = 0; j < m.p.size(); j++) {
				dp += m.p(j) * bcm::PdfGamma(values[i], m.param1(j), m.param2(j));
			}
			break;

		case MT_ECDF_KDE:
			{
				const Real sigma = m.param1(0);
				const Real minus_one_over_two_sigma_sq = -1.0 / (2.0 * square(sigma));
				const Real nc = bcm::rsqrt(2.0 * M_PI * sigma * sigma) / m.param2.size();
				for (int j = 0; j < m.param2.size(); j++) {
					Real val = values[i] - m.param2(j);
					dp += nc * exp(val * val * minus_one_over_two_sigma_sq);
				}
			}
			break;

		default:
			LOGERROR("Unknown marginal type");
			return false;
		}

		logp += log(dp);
	}

	return true;
}

bool PriorVineCopula::Sample(Real* values, RNG* rng) const
{
	const size_t N = varset->GetNumVariables();
	VectorReal v = VectorReal::Constant(N, 0);
	int one = 1;

	// Generate sample from vine copula
	MatrixReal vdirect(N, N);
	MatrixReal vindirect(N, N);
	for (size_t i = 0; i < N; i++) {
		vdirect(i,i) = rng->GetReal();
	}
	vindirect(0,0) = vdirect(0,0);
	for (int i = 1; i < N; i++) {
		for (int k = i-1; k > -1; k--) {
			int m = rMaxMat(k,i);
			if (rM(k,i) == m) {
				Hinv1(const_cast<int*>(&rfamily(k,i)), &one, &vdirect(k + 1,i), &vdirect(k,m - 1), const_cast<double*>(&rpar(k,i)), const_cast<double*>(&rpar2(k,i)), &vdirect(k,i));
			} else {
				Hinv1(const_cast<int*>(&rfamily(k,i)), &one, &vdirect(k + 1,i), &vindirect(k,m - 1), const_cast<double*>(&rpar(k, i)), const_cast<double*>(&rpar2(k, i)), &vdirect(k,i));
			}
			if (i + 1 < N) {
				if (rCoIndirect(k + 1,i) == 1) {
					if (rM(k,i) == m) {
						Hfunc2(const_cast<int*>(&rfamily(k,i)), &one, &vdirect(k,m - 1), &vdirect(k,i), const_cast<double*>(&rpar(k, i)), const_cast<double*>(&rpar2(k, i)), &vindirect(k + 1,i));
					} else {
						Hfunc2(const_cast<int*>(&rfamily(k,i)), &one, &vindirect(k,m - 1), &vdirect(k,i), const_cast<double*>(&rpar(k, i)), const_cast<double*>(&rpar2(k, i)), &vindirect(k + 1,i));
					}
				}
			}
		}
	}
	for (size_t i = 0; i < N; i++) {
		v(i) = vdirect(0,i);
	}

	// Transform from U[0,1] to parameter space
	for (size_t i = 0; i < N; i++) {
		const Marginal& m = marginals[i];
		size_t ix = m.copula_ix;

		if(m.type == MT_ECDF_KDE) {
			// Linear interpolation of the ECDF (R's quantile(type=4))
			int i1 = (int)std::floor(v(ix) * m.param2.size());
			int i2 = (int)std::ceil(v(ix) * m.param2.size());
			if (i1 < 0) {
				values[i] = m.param2(0);
			} else if (i2 >= m.param2.size()) {
				values[i] = m.param2(m.param2.size()-1);
			} else {
				Real p = v(ix) * m.param2.size() - i1;
				values[i] = p * m.param2(i2) + (1.0 - p) * m.param2(i1);
			}
		} else {
			// I'm not sure how to calculate the quantile of mixture distributions
			// Instead, I'll pick one of the components according to their weights, and use the quantile function for that
			// I believe this should give the same result as far as random sampling from the prior is concerned.
			int component_ix = (int)m.p.size() - 1;
			if (m.p.size() > 1) {
				Real p = rng->GetReal();
				Real t = 0;
				for (int i = 0; i < m.p.size() - 1; i++) {
					t += m.p(i);
					if (p < t) {
						component_ix = i;
						break;
					}
				}
			}

			switch (m.type) {
			case MT_Normal:
				values[i] = bcm::QuantileNormal(v(ix), m.param1(component_ix), m.param2(component_ix));
				break;

			case MT_Beta:
				values[i] = m.min + bcm::QuantileBeta(v(ix), m.param1(component_ix), m.param2(component_ix)) * (m.max - m.min);
				break;

			case MT_Gamma:
				values[i] = bcm::QuantileGamma(v(ix), m.param1(component_ix), m.param2(component_ix));
				break;

			default:
				LOGERROR("Unknown marginal type");
				return false;
			}
		}
	}

	return true;
}

bool PriorVineCopula::EvaluateMarginalVariance(size_t i, Real& var) const
{
	const Marginal& m = marginals[i];
	size_t ix = m.copula_ix;

	Real mean = 0.0;
	var = 0.0;
	switch (m.type) {
	case MT_Normal:
		for (int j = 0; j < m.p.size(); j++) {
			mean += m.p(j) * m.param1(j);
		}
		for (int j = 0; j < m.p.size(); j++) {
			Real v = m.param2(j) * m.param2(j);
			var += m.p(j) * (square(m.param1(j) - mean) + v);
		}
		break;

	case MT_Beta:
		// TODO - still need to take care of min/max
		for (int j = 0; j < m.p.size(); j++) {
			mean += m.p(j) * m.param1(j) / (m.param2(j) + m.param1(j));
		}
		for (int j = 0; j < m.p.size(); j++) {
			Real a = m.param1(j);
			Real b = m.param2(j);
			Real v = a * b / (square(a + b) * (a + b + 1));
			var += m.p(j) * (square(m.param1(j) - mean) + v);
		}
		break;

	case MT_Gamma:
		for (int j = 0; j < m.p.size(); j++) {
			mean += m.p(j) * m.param1(j) * m.param2(j);
		}
		for (int j = 0; j < m.p.size(); j++) {
			Real v = m.param1(j) * m.param2(j) * m.param2(j);
			var += m.p(j) * (square(m.param1(j) - mean) + v);
		}
		break;

	case MT_ECDF_KDE:
		{
			Real mu = m.param2.mean();
			Real sigma = m.param1(0);
			VectorReal d = m.param2.array() - mu;
			var = (d.cwiseProduct(d)).mean() + square(sigma);
		}
		break;

	default:
		LOGERROR("Unknown marginal type");
		return false;
	}
	return true;
}

}

#else

namespace bcm {

PriorVineCopula::PriorVineCopula(VariableSet* varset)
	: Prior(varset)
{
}

PriorVineCopula::~PriorVineCopula()
{
}

bool PriorVineCopula::LoadFromXML(const boost::property_tree::ptree& xml_node)
{
	return false;
}

bool PriorVineCopula::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	return false;
}

bool PriorVineCopula::Sample(Real* values, RNG* rng) const
{
	return false;
}

bool PriorVineCopula::EvaluateMarginalVariance(size_t i, Real& var) const
{
	return false;
}

}

#endif