#include "Utils.h"
#include "Likelihood.h"
#include "Prior.h"
#include "ProbabilityDistributions.h"
#include "RNG_SFMT.h"
#include "SamplerMH.h"
#include "VariableSet.h"

namespace bcm {

SamplerMH::SamplerMH(size_t num_threads)
	: Sampler(num_threads)
	, burnin_period(1000)
	, sampling_period(1000)
	, use_every_nth(100)
	, proposal_scaling(1.0)
	, rng(NULL)
	, lprior(-std::numeric_limits<Real>::infinity())
	, llh(-std::numeric_limits<Real>::infinity())
	, lposterior(-std::numeric_limits<Real>::infinity())
	, mode_lposterior(-std::numeric_limits<Real>::infinity())
	, mode_llh(-std::numeric_limits<Real>::infinity())
	, mode_lprior(-std::numeric_limits<Real>::infinity())
{
}

SamplerMH::~SamplerMH()
{
	safe_delete(rng);
}

bool SamplerMH::ParseImpl(const boost::program_options::variables_map& vm)
{
	try {
		burnin_period = vm["sampler.burnin_period"].as<size_t>();
		sampling_period = vm["sampler.sampling_period"].as<size_t>();
		use_every_nth = vm["sampler.use_every_nth"].as<size_t>();
		proposal_scaling = vm["sampler.proposal_scaling"].as<Real>();
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}

	return true;
}

bool SamplerMH::InitializeImpl()
{
	rng = new RNG_SFMT(rngseed);

	proposal_sigma.resize(num_variables);
	size_t index = 0;
	Real default_scale = (2.4 * 2.4) / num_variables;
	for (size_t i = 0; i < num_variables; i++) {
		Real prior_var;
		if (!prior->EvaluateMarginalVariance(i, prior_var)) {
			// TODO - somehow come up with something reasonable?
			prior_var = 1.0;
		}
		proposal_sigma[index++] = proposal_scaling * sqrt(default_scale * prior_var);
	}

	current_var_values.resize(num_variables);
	current_output_values.resize(num_output_values);
	mode_values.resize(num_variables);
	mode_lposterior = -std::numeric_limits<Real>::infinity();
	mode_llh = -std::numeric_limits<Real>::infinity();
	mode_lprior = -std::numeric_limits<Real>::infinity();

	return true;
}

Sampler::ERunResult SamplerMH::RunImpl()
{
	LOG("Finding starting point...");
	while (1) {
		if (!prior->Sample(current_var_values.data(), rng)) {
			LOGERROR("Sampling from prior failed");
			return RUN_ERROR;
		}

		if (!EvaluatePriorLikelihood(0, current_var_values.data(), lprior, llh)) {
			return RUN_ERROR;
		}
		lposterior = lprior + llh;

		if (lposterior != -std::numeric_limits<Real>::infinity()) {
			if (!GetOutputValues(0, current_output_values.data())) {
				return RUN_ERROR;
			}
			break;
		}
	}

	LOG("Starting sampling loop...");

	// Main sampling loop
	size_t totalSamples = (burnin_period + sampling_period) * use_every_nth;
	size_t num_accepted = 0;
	size_t si = 0;

	VectorReal new_values;
	Real new_lprior;
	Real new_llh;
	Real new_lposterior;
	new_values.resize(num_variables);

	Real chib_enum = 0.0;
	Real chib_denom = 0.0;

	while (si < totalSamples) {
		size_t sampleIx = si / use_every_nth;
		bool burnin = sampleIx < burnin_period;

		if (si % (totalSamples / 100) == 0) {
			double progress = 100.0 * si / (double)totalSamples;
			LOG("Progress %.0f%%", progress);
		}

		for (size_t i = 0; i < num_variables; i++) {
			Real sigma = proposal_sigma[i];
			Real x = rng->GetNormal(0.0, sigma);
			Real nx = current_var_values[i] + x;

			// Reflect for variables with bounded support, this is still a symmetrical proposal
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}

			new_values[i] = nx;
		}

		if (!EvaluatePriorLikelihood(0, new_values.data(), new_lprior, new_llh)) {
			return RUN_ERROR;
		}
		new_lposterior = new_lprior + new_llh;

		Real transition_probability = new_lposterior - lposterior;
		transition_probability = exp(transition_probability);
		transition_probability = std::min(transition_probability, (Real)1.0);

		Real alpha = rng->GetReal();
		if (alpha < transition_probability) {
			num_accepted++;

			current_var_values = new_values;
			if (!GetOutputValues(0, current_output_values.data())) {
				return RUN_ERROR;
			}
			lprior = new_lprior;
			llh = new_llh;
			lposterior = new_lposterior;
		} else {
		}

		// In the burnin period, store the highest posterior values, which
		// will be used to calculate the evidence using Chib's method.
		if (sampleIx < burnin_period) {
			if (lposterior > mode_lposterior) {
				mode_lposterior = lposterior;
				mode_llh = llh;
				mode_lprior = lprior;
				mode_values = current_var_values;
			}
		}

		if ((si+1) % use_every_nth == 0 && sampleIx >= burnin_period) {
			if (posterior_sample_handler) {
				posterior_sample_handler(current_var_values.data(), lposterior, llh, lprior, 1.0 / (Real)sampling_period, posterior_sample_handler_user);
			}
			if (output_values_handler && num_output_values > 0) {
				output_values_handler(current_output_values.data(), output_values_handler_user);
			}

			// Chib's method to calculate the evidence
			Real lq = 0.0;
			for (size_t i = 0; i < num_variables; i++) {
				Real sigma = proposal_sigma[i];
				lq += LogPdfNormal(current_var_values[i], mode_values[i], sigma);
			}
			Real ltransition_probability = mode_lposterior - lposterior;
			ltransition_probability = std::min((Real)0.0, ltransition_probability);
			chib_enum = logsum(chib_enum, ltransition_probability + lq);
		}

		si++;
	}
	
	LOG("");
	LOG("Acceptance rate: %g", (Real)num_accepted / (Real)totalSamples);
	LOG("");

	// Denominator for Chib's method
	LOG("Evaluating denominator for Chib's method...");
	for (size_t j = 0; j < sampling_period; j++) {
		for (size_t i = 0; i < num_variables; i++) {
			Real sigma = proposal_sigma[i];
			Real x = rng->GetNormal(0.0, sigma);
			Real nx = mode_values[i] + x;

			// Reflect for variables with bounded support, this is still a symmetrical proposal
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}

			new_values[i] = nx;
		}
	
		if (!EvaluatePriorLikelihood(0, new_values.data(), new_lprior, new_llh)) {
			return RUN_ERROR;
		}
		new_lposterior = new_lprior + new_llh;

		Real alpha = new_lposterior - mode_lposterior;
		alpha = std::min((Real)0.0, alpha);
		chib_denom = logsum(chib_denom, alpha);
	}
	
	evidence = mode_lposterior - (chib_enum - chib_denom);

	return RUN_FINISHED;
}

}
