#include "Utils.h"
#include "Likelihood.h"
#include "Prior.h"
#include "ProgressIndicator.h"
#include "RNG_SFMT.h"
#include "SamplerNested.h"
#include "NestedSamplingStrategyEllipsoid.h"
#include "NestedSamplingStrategyMCMC.h"
#include "NestedSamplingStrategyMultiNest.h"
#include "NestedSamplingStrategyPrior.h"

namespace bcm {

static const size_t NUM_EVIDENCE_SAMPLES = 50;
static const size_t NUM_SIMULATIONS_PER_THREAD = 2;

struct ComparePosteriorSample
{
	bool operator()(const Sample& s1, const Sample& s2) const
	{
		return s1.weight > s2.weight;
	}
};

SamplerNested::SamplerNested(size_t num_threads)
	: Sampler(num_threads)
	, population_size(500)
	, num_posterior_samples(1000)
	, sample_weights(true)
	, termination_criterion(SC_NtimesHSteps)
	, termination_num_steps(4000)
	, termination_remaining_contribution_threshold(0.01)
	, ntimesh_factor(2.0)
	, sampling_strategy_str("multinest")
	, initial_sampling_strategy_acceptance_threshold(0.1)
	, sampling_strategy_mcmc_numsteps(100)
	, rng(NULL)
	, sampling_strategy(NULL)
	, initial_sampling_strategy(NULL)
	, sampling_from_strategy(NULL)
	, information(0.0)
{
}

SamplerNested::~SamplerNested()
{
	safe_delete(rng);
	safe_delete(sampling_strategy);
	safe_delete(initial_sampling_strategy);
}

bool SamplerNested::ParseImpl(const boost::program_options::variables_map& vm)
{
	std::string termination_criterion_str;

	try {
		population_size = vm["sampler.population_size"].as<size_t>();
		num_posterior_samples = vm["sampler.nested_num_posterior_samples"].as<size_t>();
		sample_weights = vm["sampler.sample_weights"].as<bool>();
		termination_criterion_str = vm["sampler.termination_criterion"].as<std::string>();
		termination_num_steps = vm["sampler.termination_num_steps"].as<size_t>();
		termination_remaining_contribution_threshold = vm["sampler.termination_remaining_evidence"].as<Real>();
		ntimesh_factor = vm["sampler.termination_ntimesh_factor"].as<Real>();
		sampling_strategy_str = vm["sampler.nested_sampling_strategy"].as<std::string>();
		sampling_strategy_mcmc_numsteps = vm["sampler.nested_sampling_strategy_mcmc_numsteps"].as<size_t>();
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}
	
	std::transform(termination_criterion_str.begin(), termination_criterion_str.end(), termination_criterion_str.begin(), ::tolower);
	std::transform(sampling_strategy_str.begin(), sampling_strategy_str.end(), sampling_strategy_str.begin(), ::tolower);

	// Make sure all of the options have reasonable values.
	if (population_size < 2) {
		LOGERROR("population_size must be at least 2");
		return false;
	}

	if (termination_criterion_str == "fixed") {
		termination_criterion = SC_FixedStepNumber;
		termination_num_steps = std::max(termination_num_steps, num_posterior_samples);
	} else if (termination_criterion_str == "ntimesh") {
		termination_criterion = SC_NtimesHSteps;
	} else if (termination_criterion_str == "remaining_evidence") {
		termination_criterion = SC_RemainingContribution;
	} else {
		LOGERROR("Unknown termination_criterion_str \"%s\"", termination_criterion_str.c_str());
		return false;
	}

	if (termination_criterion == SC_FixedStepNumber && termination_num_steps == 0) {
		LOGERROR("termination_num_steps must be greater than 0");
		return false;
	}

	if (termination_criterion == SC_RemainingContribution && termination_remaining_contribution_threshold <= 0) {
		LOGERROR("termination_remaining_contribution_threshold must be greater than 0");
		return false;
	}

	if (termination_criterion == SC_NtimesHSteps && ntimesh_factor <= 0) {
		LOGERROR("termination_remaining_contribution_threshold must be greater than 0");
		return false;
	}
	
	if (sampling_strategy_str != "prior" && sampling_strategy_str != "ellipsoid" && sampling_strategy_str != "multinest" && sampling_strategy_str != "mcmc") {
		LOGERROR("Unknown sampling strategy \"%s\"", sampling_strategy_str.c_str());
		return false;
	}

	if (sampling_strategy_str == "mcmc" && sampling_strategy_mcmc_numsteps == 0) {
		LOGERROR("sampling_strategy_mcmc_numsteps must be greater than 0");
		return false;
	}

	return true;
}

bool SamplerNested::InitializeImpl()
{
	rng = new RNG_SFMT(rngseed);

	sampling_population.Reserve(population_size + num_threads * NUM_SIMULATIONS_PER_THREAD, num_variables, num_output_values);
	posterior_reservoir.reserve(num_posterior_samples);
	
	sampled_evidence.resize(NUM_EVIDENCE_SAMPLES);

	if (sampling_strategy_str == "prior") {
		sampling_strategy = new NestedSamplingStrategyPrior(this);
	} else if (sampling_strategy_str == "ellipsoid") {
		sampling_strategy = new NestedSamplingStrategyEllipsoid(this);
	} else if (sampling_strategy_str == "multinest") {
		sampling_strategy = new NestedSamplingStrategyMultiNest(this);
	} else if (sampling_strategy_str == "mcmc") {
		sampling_strategy = new NestedSamplingStrategyMCMC(this, sampling_strategy_mcmc_numsteps);
	} else {
		LOGERROR("Unknown sampling strategy \"%s\"", sampling_strategy_str.c_str());
		return false;
	}

	initial_sampling_strategy = new NestedSamplingStrategyPrior(this);
	sampling_from_strategy = initial_sampling_strategy;
	current_acceptance_rate_ema = 1.0;

	// Async parameters
	async.resize(num_threads);
	for (size_t i = 0; i < num_threads; i++) {
		async[i].rng = new RNG_SFMT(rng->GetInt());
	}

	async_tasks.resize(num_threads * NUM_SIMULATIONS_PER_THREAD);
	for (size_t i = 0; i < num_threads * NUM_SIMULATIONS_PER_THREAD; i++) {
		async_tasks[i].taskix = 0;
		async_tasks[i].new_sample = NULL;
		async_tasks[i].failure = false;
		async_tasks[i].min_llh = -std::numeric_limits<Real>::infinity();
		async_tasks[i].acceptance_rate = 1.0;
	}

	return true;
}

void SamplerNested::UninitializeImpl()
{
	safe_delete(sampling_strategy);
	safe_delete(initial_sampling_strategy);

	posterior_reservoir.clear();
	sampled_evidence.clear();

	safe_delete(rng);
}

Sampler::ERunResult SamplerNested::RunImpl()
{
	bool result = SampleInitialPopulation();
	if (!result) {
		return RUN_ERROR;
	}
	
	// Give the sampling strategy a chance to update to the initial active set if it wants.
	if (!initial_sampling_strategy->Initialize(1.0, rng)) {
		return RUN_ERROR;
	}
#if 0
	FILE* file = fopen("samples.tsv", "wb");
	for (TSampleSet::iterator si = samples.begin(); si != samples.end(); ++si) {
		const Sample* s = *si;
		fprintf(file, "%g,%g,%g,%g\n", s->values(0), s->values(1), s->llh, 1.0 / population_size);
	}
	fclose(file);
#endif
	
	// Initialize the trapezoidal integration
	if (sample_weights) {
		evidence = std::numeric_limits<Real>::quiet_NaN();
		information = std::numeric_limits<Real>::quiet_NaN();

		for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
			EvidenceSample& es = sampled_evidence[esi];
			es.evidence = -(std::numeric_limits<Real>::max() * std::numeric_limits<Real>::epsilon());
			es.information = 0.0;
			es.lXim1 = 0.0;
			es.lXi   = 0.0;
			
			// Sample a t by taking the max of population_size samples from uniform(0,1)
			// Inverse transform sampling with cdf(t) = t^(population_size - 1)
			Real sampled_t = pow(rng->GetReal(), (Real)1.0 / (population_size - (Real)1.0));
			
			es.lXip1 = es.lXim1 + log(sampled_t);
		}
		
		information = CalculateAverageInformation();
		evidence = CalculateAverageEvidence();
	} else {
		evidence = (-std::numeric_limits<Real>::max() * std::numeric_limits<Real>::epsilon());
		information = 0;
	}

	Real lweight_factor = log(0.5) + log(exp(1.0 / population_size) - exp(-1.0 / population_size));
	
	// Run the sampler until the stopping criterion is reached
	size_t step = 0;
	size_t emitted_samples = 0;
	size_t new_ix = 0;
	VectorReal new_value = VectorReal(num_variables);
	while (1) {
		if (termination_criterion == SC_FixedStepNumber) {
			UpdateProgress((Real)step / (std::max)(termination_num_steps, num_posterior_samples));
		} else if (termination_criterion == SC_NtimesHSteps) {
			UpdateProgress((Real)step / (ntimesh_factor * population_size * information));
		} else if (termination_criterion == SC_RemainingContribution) {
			UpdateProgress((Real)step / (2 * population_size * information));
		} else {
			LOGERROR("Invalid termination criterion");
			return RUN_ERROR;
		}

		if (step % 100 == 0) {
			LOG("Step %d: information=%g", step, information);
		}
		
		// Find the samples with the lowest likelihood
		std::list<Sample*>::iterator lowest_lh_sample = sampling_population.GetLowestLLHSample();
		Sample* remove_sample = *lowest_lh_sample;
		Real min_llh = remove_sample->llh;

		// Calculate contribution to evidence and add it using trapezoidal integration
		// Note that this requires X_(i+1), so we calculate/sample the next weight already now.
		Real lposterior_weight = 0.0;
		if (sample_weights) {
			for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
				EvidenceSample& es = sampled_evidence[esi];

				// Sample a t by taking the max of population_size samples from uniform(0,1)
				// Inverse transform sampling with cdf(t) = t^(population_size - 1)
				Real sampled_t = pow(rng->GetReal(), (Real)1.0 / (population_size - (Real)1.0));
				
				es.lXim1 = es.lXi;
				es.lXi   = es.lXip1;
				es.lXip1 = es.lXi + log(sampled_t);

				// Calculate evidence with these sampled X_i's
				Real lweight = log(0.5) + logaminusb(es.lXim1, es.lXip1);
				Real new_evidence = logsum(es.evidence, remove_sample->llh + lweight);
				if (remove_sample->llh == -std::numeric_limits<Real>::infinity()) {
					es.information = exp(es.evidence - new_evidence) * (es.information + es.evidence) - new_evidence;
				} else {
					es.information = exp(lweight + remove_sample->llh - new_evidence) * remove_sample->llh + exp(es.evidence - new_evidence) * (es.information + es.evidence) - new_evidence;
				}
				es.evidence = new_evidence;

				lposterior_weight += lweight;
			}
			lposterior_weight /= static_cast<Real>(NUM_EVIDENCE_SAMPLES);
			lposterior_weight += remove_sample->llh;
		} else {
			// Assuming X_i = exp(-i / N)
			Real lweight = -(Real)(step+1) / (Real)population_size + lweight_factor;

			Real new_evidence = logsum(evidence, remove_sample->llh + lweight);
			information = exp(lweight + remove_sample->llh - new_evidence) * remove_sample->llh + exp(evidence - new_evidence) * (information + evidence) - new_evidence;
			evidence = new_evidence;

			lposterior_weight = lweight + remove_sample->llh;
		}

		// Add this sample to the posterior according to its weight
		EmitPosteriorSample(remove_sample, lposterior_weight);
		emitted_samples++;
		
		// Remove this sample from the population
		sampling_population.RemoveSample(lowest_lh_sample);

		// Update the average evidence and information if they were sampled
		if (sample_weights) {
			information = CalculateAverageInformation();
			evidence = CalculateAverageEvidence();
		}
		
		// Allow the sampling strategy to update
		// The expectation of prior volume should be accurate enough for this.
		Real prior_volume_estimate = exp(-(Real)(step+1) / (Real)population_size);
		bool reset = false;
		if (!sampling_from_strategy->UpdateRemoveSample(remove_sample, prior_volume_estimate, reset, rng)) {
			return RUN_ERROR;
		}
		
		// Find new samples with a likelihood strictly higher than the likelihood we just removed.
		Sample* accepted_sample = NULL;
		size_t discarded_samples = 0;
		while (1) {
			// Check if we can use any of the stores samples
			if (reset) {
				while (!stored_samples.empty()) {
					Sample* new_sample = stored_samples.front();
					sampling_population.FreeSample(new_sample);
					stored_samples.pop();
				}
				reset = false;
			} else {
				while (!stored_samples.empty()) {
					Sample* new_sample = stored_samples.front();
					stored_samples.pop();
					if (new_sample->llh > min_llh) {
						sampling_population.AddSample(new_sample);
						accepted_sample = new_sample;
						break;
					} else {
						sampling_population.FreeSample(new_sample);
						discarded_samples++;
					}
				}
			}

			if (accepted_sample) {
				break;
			}

			// Generate a bunch of new samples
			async_tasks.resize(num_threads * NUM_SIMULATIONS_PER_THREAD);
			for (size_t asynci = 0; asynci < num_threads * NUM_SIMULATIONS_PER_THREAD; asynci++) {
				// Get new sample
				Sample* new_sample = sampling_population.AllocateSample();
				async_tasks[asynci].new_sample = new_sample;
				async_tasks[asynci].min_llh = min_llh;

				// Get new sample values asynchronously
				TTask task = boost::bind(&SamplerNested::GetNewSampleAsync, this, _1, _2);
				async_tasks[asynci].taskix = task_manager->AddTask(task, reinterpret_cast<void*>(asynci));
			}
			for (size_t asynci = 0; asynci < num_threads * NUM_SIMULATIONS_PER_THREAD; asynci++) {
				// Wait for asynchronous operation
				task_manager->WaitTask(async_tasks[asynci].taskix);

				if (async_tasks[asynci].failure) {
					return RUN_ERROR;
				}
			
				// Store sample
				stored_samples.push(async_tasks[asynci].new_sample);
			}
		}

		const Real ema_alpha = 2.0 / (20 + 1);
		const Real acceptance_rate = 1.0 / (Real)(discarded_samples + 1);
		current_acceptance_rate_ema += (acceptance_rate - current_acceptance_rate_ema) * ema_alpha;

		// Check whether we need to switch sampling strategies
		if (sampling_from_strategy == initial_sampling_strategy && current_acceptance_rate_ema < initial_sampling_strategy_acceptance_threshold) {
			LOG("Switching away from prior sampling to main sampling strategy");
			sampling_from_strategy = sampling_strategy;
			if (!sampling_strategy->Initialize(prior_volume_estimate, rng)) {
				return RUN_ERROR;
			}
		} else {
			// Allow the sampling strategy to update
			if (!sampling_from_strategy->UpdateAddSample(accepted_sample, rng)) {
				return RUN_ERROR;
			}
		}

		step++;

		// Check the termination criterion
		if (termination_criterion == SC_FixedStepNumber) {
			if (step >= termination_num_steps && emitted_samples >= num_posterior_samples) {
				break;
			}
		} else if (termination_criterion == SC_NtimesHSteps) {
			if (step > ntimesh_factor * population_size * information && step >= population_size && emitted_samples >= num_posterior_samples) {
				break;
			}
		} else if (termination_criterion == SC_RemainingContribution) {
			if (emitted_samples >= num_posterior_samples) {
				Real current_evidence;
				if (sample_weights) {
					current_evidence = CalculateAverageEvidence();
				} else {
					current_evidence = evidence;
				}

				Real lweight = -(Real)(step+1) / (Real)population_size + lweight_factor;
				Real maxllh = (*sampling_population.GetHighestLLHSample())->llh;
				Real maxz = logsum(current_evidence, maxllh + lweight);
				if (maxz - current_evidence < termination_remaining_contribution_threshold) {
					break;
				}
			}
		} else {
			LOGERROR("Invalid termination criterion");
			return RUN_ERROR;
		}
	}
	
	// Wrap up the evidence calculation & error estimate
	AddFinalContribution(step, lweight_factor);
	if (sample_weights) {
		evidence = CalculateAverageEvidence();
	}
	CalculateEvidenceErrorEstimate();

	// Add the remaining live samples to the posterior samples
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		const Sample* sample = *si;

		Real lposterior_weight = 0.0;
		for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
			EvidenceSample& es = sampled_evidence[esi];

			// Sample a t by taking the max of population_size samples from uniform(0,1)
			// Inverse transform sampling with cdf(t) = t^(population_size - 1)
			Real sampled_t = pow(rng->GetReal(), (Real)1.0 / (population_size - (Real)1.0));

			// Calculate evidence with these sampled X_i's
			Real lweight = es.lXi + log(sampled_t);
			lposterior_weight += lweight;
		}
		lposterior_weight /= static_cast<Real>(NUM_EVIDENCE_SAMPLES);
		lposterior_weight += sample->llh;
		lposterior_weight -= log((Real)population_size);

		//EmitPosteriorSample(sample, -(Real)(step+1) / (Real)population_size + lweight_factor + sample->llh);
		EmitPosteriorSample(sample, lposterior_weight);
	}

	// Output the posterior samples
	// First normalize them; and sort them just because we can.
	std::sort_heap(posterior_reservoir.begin(), posterior_reservoir.end(), ComparePosteriorSample());
	ASSERT(posterior_reservoir.size() == num_posterior_samples);
	Real total_lweight = -std::numeric_limits<Real>::infinity();
	for (size_t i = 0; i < num_posterior_samples; i++) {
		const Sample& s = posterior_reservoir[i];
		total_lweight = logsum(total_lweight, s.weight);
	}
	for (size_t i = 0; i < num_posterior_samples; i++) {
		Sample& s = posterior_reservoir[i];
		s.weight = exp(s.weight - total_lweight);
		if (posterior_sample_handler) {
			posterior_sample_handler(s.variable_values.data(), s.lposterior, s.llh, s.lprior, s.weight, posterior_sample_handler_user);
		}
		if (output_values_handler && num_output_values > 0) {
			output_values_handler(s.output_values.data(), output_values_handler_user);
		}
	}

	UpdateProgress(1.0);
	LOG("Finished!");

	return RUN_FINISHED;
}

bool SamplerNested::SampleInitialPopulation()
{
	// Initial sampling from prior
	LOG("Sampling initial population from prior...");
	for (size_t j = 0; j < population_size; j++) {
		Sample* sample = sampling_population.AllocateSample();
		
		if (!prior->Sample(sample->variable_values.data(), rng)) {
			LOGERROR("Sampling from prior failed");
			return false;
		}

		if (!EvaluatePriorLikelihood(0, sample->variable_values.data(), sample->lprior, sample->llh)) {
			return false;
		}
		GetOutputValues(0, sample->output_values.data());
		sample->lposterior = sample->lprior + sample->llh;

		sampling_population.AddSample(sample);
	}

	return true;
}

void SamplerNested::AddFinalContribution(size_t steps_taken, Real lweight_factor)
{
	if (sample_weights) {
		for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
			EvidenceSample& es = sampled_evidence[esi];

			Real lweight = es.lXi - log((Real)population_size);
			for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
				const Sample* sample = *si;

				Real new_evidence = logsum(es.evidence, sample->llh + lweight);
				es.information = exp(lweight + sample->llh - new_evidence) * sample->llh + exp(es.evidence - new_evidence) * (information + es.evidence) - new_evidence;
				es.evidence = new_evidence;
			}
		}
	} else {
		Real lweight = -(Real)(steps_taken+1) / (Real)population_size + lweight_factor - log((Real)population_size);
		for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
			const Sample* sample = *si;

			Real new_evidence = logsum(evidence, sample->llh + lweight);
			information = exp(lweight + sample->llh - new_evidence) * sample->llh + exp(evidence - new_evidence) * (information + evidence) - new_evidence;
			evidence = new_evidence;
		}
	}
}

void SamplerNested::CalculateEvidenceErrorEstimate()
{
	if (sample_weights) {
		evidence_sd = 0.0;
		for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
			EvidenceSample& es = sampled_evidence[esi];
			Real d = es.evidence - evidence;
			evidence_sd += (d * d) - evidence_sd / (esi + 1);
		}
		evidence_sd *= (NUM_EVIDENCE_SAMPLES / (NUM_EVIDENCE_SAMPLES - 1.0));
		evidence_sd = sqrt(evidence_sd);
	} else {
		evidence_sd = sqrt(information / population_size);
	}
}

Real SamplerNested::CalculateAverageEvidence()
{
	Real result = 0.0;
	for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
		EvidenceSample& es = sampled_evidence[esi];
		result += es.evidence / NUM_EVIDENCE_SAMPLES;
	}
	return result;
}

Real SamplerNested::CalculateAverageInformation()
{
	Real result = 0.0;
	for (size_t esi = 0; esi < NUM_EVIDENCE_SAMPLES; esi++) {
		EvidenceSample& es = sampled_evidence[esi];
		result += es.information / NUM_EVIDENCE_SAMPLES;
	}
	return result;
}

void SamplerNested::UpdateProgress(Real p)
{
	if (progress) {
		char buf[16];
		snprintf(buf, sizeof(buf), "%6g", information);
		progress->UpdateInfo("H", buf);
		snprintf(buf, sizeof(buf), "%6g", evidence);
		progress->UpdateInfo("evidence", buf);
		snprintf(buf, sizeof(buf), "%4g", current_acceptance_rate_ema);
		progress->UpdateInfo("acceptance rate", buf);
		
		if (p != p || p < -0.0) {
			progress->ReportProgress(0.0);
		} else {
			progress->ReportProgress(p);
		}
	}
}

void SamplerNested::GetNewSampleAsync(void* user, size_t threadix)
{
	size_t ti = reinterpret_cast<size_t>(user);
	if (!sampling_from_strategy->GetNewSample(async_tasks[ti].min_llh, async_tasks[ti].new_sample, async[threadix].rng, threadix)) {
		async_tasks[ti].failure = true;
	}
}

void SamplerNested::EmitPosteriorSample(const Sample* sample, Real lposterior_weight)
{
	// Construct posterior sample
	Sample ps = *sample;
	ps.weight = lposterior_weight;

	// Check if the reservoir has been filled already
	if (posterior_reservoir.size() == num_posterior_samples) {
		// It has, replace the lowest weight sample if the new sample has higher weight.
		const Sample& least = posterior_reservoir.front();
		if (least.weight > lposterior_weight) {
			// Do nothing.
		} else {
			// Remove the sample with the least weight
			std::pop_heap(posterior_reservoir.begin(), posterior_reservoir.end(), ComparePosteriorSample());
			posterior_reservoir.pop_back();

			// Insert the new sample into the heap
			posterior_reservoir.push_back(ps);
			std::push_heap(posterior_reservoir.begin(), posterior_reservoir.end(), ComparePosteriorSample());
		}
	} else {
		// Reservoir isn't full yet, just insert the new sample.
		posterior_reservoir.push_back(ps);
		std::push_heap(posterior_reservoir.begin(), posterior_reservoir.end(), ComparePosteriorSample());
	}

#if 0
	// Attempt at reservoir sampling - it's not entirely trivial due to the large difference in weights, which causes overflow in the key calculation.

	// Reservoir sampling (Efraimidis & Spirakis, 2005)

	// Test if this sample should be inserted into the posterior set
	Real new_key = pow(rng->GetReal(), exp(-lposterior_weight));
	bool insert = false;
	if (posterior_reservoir.size() == num_posterior_samples) {
		// Reservoir is full, replace sample if u^(1/w_i) > min
		std::set<PosteriorSample, ComparePosteriorSample>::iterator min_key_sample = posterior_reservoir.begin();
		if (new_key > min_key_sample->key) {
			posterior_reservoir.erase(min_key_sample);
			insert = true;
		} else {
			insert = false;
		}
	} else {
		// First items are always kept to fill the reservoir
		insert = true;
	}

	if (insert) {
		PosteriorSample ps;
		ps.values = sample->values;
		ps.key = new_key;
		posterior_reservoir.insert(ps);
	}
#endif
}

}
