#pragma once

#include "RNG.h"
#include "Sampler.h"
#include "Serialization.h"

namespace bcm {

class SMCSamplingStrategy;

class SamplerSMC : public Sampler
{
public:
	SamplerSMC(size_t num_threads);
	virtual ~SamplerSMC();

	virtual size_t GetNumSamples() { return population_size / use_every_nth; }
	
	struct Particle {
		template<class Archive> void serialize(Archive& ar, const unsigned int version);

		VectorReal var_values;		///< Current parameter values
		VectorReal output_values;	
		Real lprior;				///< log prior
		Real llh;					///< log likelihood
		Real lpowerposterior;		///< log power posterior (= log prior + temperature * log likelihood)
		Real lweight;				///< log un-normalized weight (incremental weight)
		Real normweight;			///< normalized weight
		Real lnormweight;			///< log normalized weight
		Real accumulated_lweight;	///< Used in calculating the evidence term.
	};
	typedef std::vector<Particle> TParticleVector;

    friend class boost::serialization::access;
    template<class Archive> void serialize(Archive& ar, const unsigned int version);

private:
	virtual bool ParseImpl(const boost::program_options::variables_map& vm);
	virtual bool InitializeImpl();
	virtual void UninitializeImpl();
	virtual ERunResult RunImpl();
	
	bool SampleInitialPopulation();
	bool SampleNewPopulation();
	void Resample();
	void AsyncDoInitialSample(void* user, size_t threadix);
	void AsyncDoSample(void* user, size_t threadix);
	bool CalculateNextTemperature(uint64 step);
	bool NormalizeWeights();
	Real CalculateESS();
	void CalculateMeanLogLikelihood(Real ess, Real& mean, Real& var);
	Real CalculateCorrelation(TParticleVector* p1, TParticleVector* p2);
	void CalculateEvidenceRatio(TParticleVector* population, Real& ratio, Real& ratio_sd);
	void UpdateProgress(Real p, Real t, Real ess, Real r);

	enum ETemperatureSchedule {
		TS_FixedLinear,			///< Run the sequential sampler for a fixed number of steps, going linearly from start to end temperature
		TS_FixedLogLinear,		///< Run the sequential sampler for a fixed number of steps, going linearly in log space from start to end temperature
		TS_AdaptiveFromESS,		///< Run the sequential sampler with an adaptive step size, such that the approximate ESS of each step is at least schedule_ess_ratio * current_ess (ratio between 0 and 1).
	};
	enum EResamplingStrategy {
		RS_Multinomial,
		RS_Residual,
	};

	// Settings
	size_t	population_size;					///< Number of particles
	size_t	use_every_nth;

	Real	start_temperature;					///< Start temperature, between 0 and 1
	Real	end_temperature;					///< End temperature, typically 1
	ETemperatureSchedule temperature_schedule;	///< Specifies how the sampler moves from start to end temperature
	size_t	schedule_num_populations;			///< Number of iterations, only meaningful if temperature_schedule is one of the two fixed options
	Real	schedule_ess_ratio;					///< Controls the adaptive temperature schedule, only meaningful if temperature_schedule is AdaptiveFromESS
	Real	resampling_ess_threshold;			///< Threshold at which resampling of the particles should occur, specified as a fraction between 0 and 1.
	EResamplingStrategy resampling_strategy;	///< Resampling strategy
	std::string sampling_strategy_str;
	size_t	mcmc_num_steps;
	size_t	mcmc_recalculate;
	size_t	mcmc_local_num_nearby;
	Real	mcmc_max_correlation;
	Real	final_ess_threshold;
	bool	run_until_decorrelation;

	std::vector<Real> output_temperatures;		///< Dump posterior at each of these temperatures

	// Run-time variables
	RNG* rng;
	SMCSamplingStrategy* sampling_strategy;

	TParticleVector populations[2];
	TParticleVector* current_population;
	TParticleVector* previous_population;
	TParticleVector population_at_resample;
	size_t current_iteration;
	size_t population_ix;
	size_t steps_between_population;
	VectorReal correlation_work_vector[3];

	Real current_temperature;
	Real current_ess;
	Real previous_temperature;
	Real evidence_since_resample;

	struct AsyncParams {
		RNG* rng;
		Real logtot;
		bool failure;
	};
	std::vector<AsyncParams> async;
	std::vector<uint64> async_tasks;

	friend class SMCSamplingStrategy;
};

template<class Archive> void SamplerSMC::Particle::serialize(Archive& ar, const unsigned int version)
{
	ar & var_values;
	ar & output_values;
	ar & lprior;
	ar & llh;
	ar & lpowerposterior;
	ar & lweight;
	ar & normweight;
	ar & lnormweight;
	ar & accumulated_lweight;
}

template<class Archive> void SamplerSMC::serialize(Archive& ar, const unsigned int version)
{
    ar & boost::serialization::base_object<Sampler>(*this);

	ar & populations[0];
	ar & populations[1];
	ar & population_at_resample;
	ar & current_iteration;
	ar & population_ix;
	ar & steps_between_population;

	ar & current_temperature;
	ar & current_ess;
	ar & previous_temperature;
	ar & evidence_since_resample;
}

}
