#include "Utils.h"
#include "Correlation.h"
#include "SamplerSMC.h"

namespace bcm {

bool SamplerSMC::CalculateNextTemperature(uint64 step)
{
	if (temperature_schedule == TS_FixedLinear) {
		if (step == schedule_num_populations - 1) {
			// This is just to make sure there's no floating point errors.
			current_temperature = end_temperature;
		} else {
			Real alpha = (step / (Real)(schedule_num_populations - 1));
			current_temperature = start_temperature + alpha * (end_temperature - start_temperature);
		}
		return true;
	} else if (temperature_schedule == TS_FixedLogLinear) {
		if (step == schedule_num_populations - 1) {
			// This is just to make sure there's no floating point errors.
			current_temperature = end_temperature;
		} else {
			Real alpha = (step / (Real)(schedule_num_populations - 1));
			Real tstart = start_temperature;
			if (tstart == 0) {
				tstart = std::numeric_limits<Real>::denorm_min();
			}
			current_temperature = exp(log(tstart) + alpha * (log(end_temperature) - log(tstart)));
		}
		return true;
	} else if (temperature_schedule == TS_AdaptiveFromESS) {
		// Here we assume the next target density is approximately equal to the previous one,
		// so that ESS(next) is independent of the actual samples we will generate in the next iteration.
		Real proposed_temperature;
		if (step == 1) {
			proposed_temperature = end_temperature;
		} else {
			// Let's be slightly optimistic, and hope we can make a step 4 times as big as the previous step.
			proposed_temperature = current_temperature + 4 * (current_temperature - previous_temperature);
			proposed_temperature = std::min(proposed_temperature, end_temperature);
		}
		while (1) {
			Real logtot = -std::numeric_limits<Real>::infinity();
			for (size_t i = 0; i < population_size; i++) {
				Particle& new_particle = (*current_population)[i];
				const Particle& previous_particle = (*previous_population)[i];
				if (previous_particle.normweight <= 0) {
					new_particle.lweight = -std::numeric_limits<Real>::infinity();
				} else {
					new_particle.lweight = (previous_particle.lprior + proposed_temperature * previous_particle.llh) - previous_particle.lpowerposterior;
					logtot = logsum(logtot, previous_particle.lnormweight + new_particle.lweight);
				}
			}
			
			bool adjust_further = true;
			if (logtot > -std::numeric_limits<Real>::infinity()) {
				for (size_t i = 0; i < population_size; i++) {
					const Particle& previous_particle = (*previous_population)[i];
					Particle& current_particle = (*current_population)[i];
					current_particle.lnormweight = previous_particle.lnormweight + current_particle.lweight - logtot;
					current_particle.normweight = exp(current_particle.lnormweight);
				}

				Real proposed_ess = CalculateESS();
				if (proposed_ess >= schedule_ess_ratio * current_ess) {
					adjust_further = false;
				}
			}
			
			if (adjust_further) {
				proposed_temperature = 0.5 * (proposed_temperature + current_temperature);
			} else {
				break;
			}
		}
		previous_temperature = current_temperature;
		current_temperature = proposed_temperature;
		return true;
	} else {
		LOGERROR("Unknown temperature schedule %d", (int)temperature_schedule);
		return false;
	}
}

bool SamplerSMC::NormalizeWeights()
{
	Real logtot = -std::numeric_limits<Real>::infinity();
	for (size_t i = 0; i < population_size; i++) {
		const Particle& previous_particle = (*previous_population)[i];
		const Particle& current_particle = (*current_population)[i];
		if (previous_particle.lnormweight == -std::numeric_limits<Real>::infinity() || current_particle.lweight == -std::numeric_limits<Real>::infinity()) {
		} else {
			logtot = logsum(logtot, previous_particle.lnormweight + current_particle.lweight);
		}
	}
	
	if (logtot > -std::numeric_limits<Real>::infinity()) {
		for (size_t i = 0; i < population_size; i++) {
			const Particle& previous_particle = (*previous_population)[i];
			Particle& current_particle = (*current_population)[i];
			if (previous_particle.lnormweight == -std::numeric_limits<Real>::infinity() || current_particle.lweight == -std::numeric_limits<Real>::infinity()) {
				current_particle.lnormweight = -std::numeric_limits<Real>::infinity();
				current_particle.normweight = 0;
			} else {
				current_particle.lnormweight = previous_particle.lnormweight + current_particle.lweight - logtot;
				current_particle.normweight = exp(current_particle.lnormweight);
			}
		}
		return true;
	} else {
		LOGERROR("Zero particles left! Consider using more temperature steps or a different temperature schedule.");
		return false;
	}
}

Real SamplerSMC::CalculateESS()
{
	Real sum = 0.0;
	for (size_t i = 0; i < population_size; i++) {
		Real w = (*current_population)[i].normweight;
		sum += w*w;
	}
	return 1.0 / sum;
}

void SamplerSMC::CalculateMeanLogLikelihood(Real ess, Real& mean, Real& var)
{
	mean = 0.0;
	Real acc_weight = 0.0;
	for (size_t i = 0; i < population_size; i++) {
		const Real x = (*current_population)[i].llh;
		const Real w = (*current_population)[i].normweight * population_size;
		acc_weight += w;
		if (w > 0.0) {
			mean += (x - mean) * w / acc_weight;
		}
	}
	
	var = 0.0;
	acc_weight = 0.0;
	for (size_t i = 0; i < population_size; ++i) {
		const Real x = (*current_population)[i].llh;
		const Real d = x - mean;
		const Real w = (*current_population)[i].normweight * population_size;
		acc_weight += w;
		if (w > 0.0) {
			var += (d*d - var) * w / acc_weight;
		}
	}
	const Real n = static_cast<Real>(population_size);
	var = var * (n / (ess - 1.0));
}

Real SamplerSMC::CalculateCorrelation(TParticleVector* p1, TParticleVector* p2)
{
#if 1
	VectorReal correlations = VectorReal::Zero(num_variables);
	for (size_t j = 0; j < num_variables; j++) {
		for (size_t i = 0; i < population_size; i++) {
			correlation_work_vector[0](i) = (*p1)[i].var_values(j);
			correlation_work_vector[1](i) = (*p2)[i].var_values(j);
			correlation_work_vector[2](i) = (*p1)[i].normweight * population_size;
		}

		//correlations(j) = pearson_correlation(correlation_work_vector[0], correlation_work_vector[1]);
		//correlations(j) = pearson_correlation_weighted(correlation_work_vector[0], correlation_work_vector[1], correlation_work_vector[2]);
		correlations(j) = spearman_correlation_weighted(correlation_work_vector[0], correlation_work_vector[1], correlation_work_vector[2]);
	}

	return correlations.maxCoeff();
#else
	for (size_t i = 0; i < population_size; i++) {
		correlation_work_vector[0](i) = (*p1)[i].var_values(0);
		correlation_work_vector[1](i) = (*p2)[i].var_values(0);
		correlation_work_vector[2](i) = (*p2)[i].normweight * population_size;
	}
	return spearman_correlation_weighted(correlation_work_vector[0], correlation_work_vector[1], correlation_work_vector[2]);
#endif
}

void SamplerSMC::CalculateEvidenceRatio(TParticleVector* population, Real& ratio, Real& ratio_sd)
{
	ratio = -std::numeric_limits<Real>::infinity();
	for (size_t i = 0; i < population_size; i++) {
		const Particle& particle = (*population)[i];
		ratio = logsum(ratio, particle.accumulated_lweight);
	}
			
	ratio_sd = 0;
	for (size_t i = 0; i < population_size; i++) {
		const Particle& particle = (*population)[i];
		if (particle.accumulated_lweight > -std::numeric_limits<Real>::infinity()) {
			Real d = particle.accumulated_lweight - ratio;
			ratio_sd += d*d;
		}
	}
	ratio_sd *= 1.0 / ((Real)population_size * (Real)population_size);
}

}
