#include "Utils.h"
#include "LikelihoodTimeAverage.h"
#include "SBMLModel.h"

static const size_t NUM_TRAJECTORY_TIMEPOINTS = 20;

LikelihoodTimeAverage::LikelihoodTimeAverage()
	: NumReplicates(0)
	, StartTime(std::numeric_limits<Real>::quiet_NaN())
	, EndTime(std::numeric_limits<Real>::quiet_NaN())
{
}

LikelihoodTimeAverage::~LikelihoodTimeAverage()
{
}

bool LikelihoodTimeAverage::Load(const boost::property_tree::ptree& xml_node, const bcm::NetCDFDataFile* datafile)
{
	if (!LikelihoodPart::Load(xml_node, datafile)) {
		return false;
	}

	if (LikelihoodFunction == LF_Delta) {
		LOGERROR("Cannot have delta likelihood function on a timepoint (cannot force trajectory to certain values in the ODE solver)");
		return false;
	}
	
	try {
		StartTime  = xml_node.get<Real>("<xmlattr>.start_time");
		EndTime = xml_node.get<Real>("<xmlattr>.end_time");
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	// Load the data into memory
	size_t dimcount;
	if (!datafile->GetDimensionCount(DataGroup, DataNameWithoutIndex, &dimcount)) {
		return false;
	}
	if (dimcount == 1) {
		NumReplicates = 1;
		
		Data.resize(1);
		if (!datafile->GetValue(DataGroup, DataNameWithoutIndex, DataIndices[0], &Data[0])) {
			return false;
		}
	} else if (dimcount == 2) {
		if (!datafile->GetDimensionSize(DataGroup, DataNameWithoutIndex + "_replicates", &NumReplicates)) {
			return false;
		}

		Data.resize(NumReplicates);
		for (size_t ri = 0; ri < NumReplicates; ri++) {
			if (!LookupDataValue1D(datafile, ri, &Data[ri])) {
				return false;
			}
		}
	} else {
		LOGERROR("Unsupported number of dimensions for data variable \"%s\"", DataNameWithoutIndex.c_str());
		return false;
	}
	
	for (size_t ri = 0; ri < NumReplicates; ri++) {
		Data[ri] = bcm::log10(Data[ri]);
	}

	return true;
}

bool LikelihoodTimeAverage::SetVariableSet(const bcm::VariableSet* variables)
{
	if (!LikelihoodPart::SetVariableSet(variables)) {
		return false;
	}

	return true;
}

bool LikelihoodTimeAverage::GetRequiredTimepoints(std::vector<Real>& timepoints)
{
	timepoints.resize(NUM_TRAJECTORY_TIMEPOINTS);
	timepoints[0] = StartTime;
	for (size_t i = 1; i < NUM_TRAJECTORY_TIMEPOINTS - 1; i++) {
		timepoints[i] = StartTime + i * (EndTime - StartTime) / (Real)NUM_TRAJECTORY_TIMEPOINTS;
	}
	timepoints[NUM_TRAJECTORY_TIMEPOINTS-1] = EndTime;

	return true;
}

bool LikelihoodTimeAverage::EvaluateLogProbability(size_t threadix, const Real* values, SBMLModel* model, std::vector<Real> simulated_time, Real& logp)
{
	VectorReal simulated_trajectory;
	if (!model->GetOutput(SpeciesName, simulated_trajectory, threadix)) {
		return false;
	}

	if (simulated_time.size() < 2) {
		LOGERROR("Insufficient time points to integrate");
		return false;
	}

	// This assumes that StartTime and EndTime are included in the simulated_time.
	Real integral = 0.0;
	for (size_t sti = 1; sti < simulated_time.size(); sti++) {
		if (simulated_time[sti] <= StartTime) {
			continue;
		}

		Real dt = simulated_time[sti] - simulated_time[sti-1];
		Real x = simulated_trajectory(sti) + simulated_trajectory(sti-1);
		integral += 0.5 * dt * x;

		if (simulated_time[sti] == EndTime) {
			break;
		}
	}

	Real weighted_average = integral / (EndTime - StartTime);

	logp = 0.0;
	for (size_t ri = 0; ri < NumReplicates; ri++) {
		Real observed = Data[ri];
		if (observed != observed) {
			// Ignore NaN values
			continue;
		} else {
			Real modeled;
			if (weighted_average <= 0.0) {
				modeled = -std::numeric_limits<Real>::infinity();
			} else {
				modeled = bcm::log10(weighted_average);
			}
			logp += EvaluateLogPdf(values, modeled, observed);
		}
	}

	return true;
}
