#include "Utils.h"
#include "LikelihoodDataFile.h"
#include "SBMLLikelihood.h"
#include "SBMLModel.h"

#include <boost/property_tree/xml_parser.hpp>

SBMLLikelihood::SBMLLikelihood(size_t numthreads)
	: numthreads(numthreads)
	, Model(NULL)
	, OutputHandler(NULL)
{
}

SBMLLikelihood::~SBMLLikelihood()
{
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		delete *dfi;
	}
}

bool SBMLLikelihood::LoadFromXML(const std::string& filename)
{
	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(filename, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading likelihood file: %s", e.what());
		return false;
	}
	
	try {
		boost::property_tree::ptree likelihood_node = pt.get_child("likelihood");

		// Load the model
		boost::property_tree::ptree model = likelihood_node.get_child("model");
		std::string modelfn = model.get<std::string>("<xmlattr>.file");

		Model = new SBMLModel(numthreads);
		if (!Model->LoadSBML(modelfn)) {
			return false;
		}
		
		// Load all data files
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& data, likelihood_node.get_child("")) {
			if (data.first == "datafile") {
				LikelihoodDataFile* datafile = new LikelihoodDataFile(numthreads);
				if (!datafile->Load(data.second)) {
					return false;
				}
				DataFiles.push_back(datafile);
			}
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	return true;
}

bool SBMLLikelihood::SetVariableSet(const bcm::VariableSet* varset)
{
	if (!Model->SetVariableSet(varset)) {
		return false;
	}

	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		if (!(*dfi)->SetVariableSet(varset)) {
			return false;
		}
	}

	return true;
}

bool SBMLLikelihood::Initialize()
{
	if (!Model) {
		LOGERROR("No model specified");
		return false;
	}
	
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		if (!(*dfi)->Initialize()) {
			return false;
		}
	}

	if (!Model->Initialize()) {
		return false;
	}
	
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		(*dfi)->SetNumModelSpecies(Model->GetNumSimulatedSpecies());
	}

	return true;
}

bool SBMLLikelihood::SetupAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_count)
{
	bool result = true;
	
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		if (!(*dfi)->SetupAdditionalOutput(file, Model, sample_count)) {
			return false;
		}
	}

	return result;
}

void SBMLLikelihood::HandleAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_ix, const Real* values)
{
	size_t ix = 0;
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		size_t datafile_numvalues = 0;
		(*dfi)->HandleAdditionalOutput(file, Model, sample_ix + ix, values, datafile_numvalues);
		ix += datafile_numvalues;
	}
}

bool SBMLLikelihood::EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
{
	logp = 0.0;
	
	for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
		Real dflogp = -std::numeric_limits<Real>::quiet_NaN();
		if (!(*dfi)->EvaluateLogProbability(threadix, values, Model, dflogp)) {
			return false;
		}
		logp += dflogp;

		if (logp == -std::numeric_limits<Real>::infinity()) {
			// No need to continue
			break;
		}
	}

	return true;
}

bool SBMLLikelihood::GetOutput(size_t threadix, Real* output_values, size_t& numvalues)
{
	if (output_values) {
		size_t ix = 0;
		for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
			size_t datafile_numvalues = 0;
			if (!(*dfi)->GetOutput(threadix, output_values + ix, datafile_numvalues)) {
				return false;
			}
			ix += datafile_numvalues;
		}
	} else {
		for (std::vector<LikelihoodDataFile*>::iterator dfi = DataFiles.begin(); dfi != DataFiles.end(); ++dfi) {
			size_t datafile_numvalues = 0;
			if (!(*dfi)->GetOutput(threadix, output_values, datafile_numvalues)) {
				return false;
			}
			numvalues += datafile_numvalues;
		}
	}

	return true;
}

void SBMLLikelihood::PostInference(const std::string& output_dir)
{
	if (Model) {
		//Model->DumpODEStatistics(output_dir + std::string("ode_solve"));
	}
}
