#include "Utils.h"

#include "SBMLLikelihood.h"
#include "SBMLModel.h"
#include "NetCDFDataFile.h"
#include "OutputHandlerNetCDF.h"
#include "Prior.h"
#include "ProgressIndicatorConsole.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "VariableSet.h"
#include "RNG_SFMT.h"

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <fstream>
#include <boost/property_tree/xml_parser.hpp>

namespace po = boost::program_options;

int run(const po::variables_map& vm)
{
	SBMLLikelihood* likelihood = NULL;
	bcm::VariableSet* varset = NULL;
	bcm::Prior* prior = NULL;
	bcm::RNG* rng = NULL;
	bcm::Sampler* sampler = NULL;
	bcm::OutputHandlerNetCDF output;
	bool result = true;
	
	try {
		// Determine the number of threads we should use
		size_t numthreads = vm["numthreads"].as<size_t>();
		if (numthreads <= 0) {
			numthreads = boost::thread::hardware_concurrency();
			printf("numthreads <= 0 specified, found %u hardware concurrency\n", (unsigned int)numthreads);
			if (numthreads <= 0) {
				numthreads = 1;
			}
		}
		
		varset = new bcm::VariableSet;
		result = varset->LoadFromXML(vm["prior"].as<std::string>());
		if (!result) {
			return -2;
		}

		prior = bcm::Prior::Create(vm["prior"].as<std::string>(), varset);
		if (!prior) {
			return -3;
		}

		likelihood = new SBMLLikelihood(numthreads);
		result = likelihood->LoadFromXML(vm["likelihood"].as<std::string>());
		if (!result) {
			return -5;
		}
		result = likelihood->SetVariableSet(varset);
		if (!result) {
			return -6;
		}
		result = likelihood->Initialize();
		if (!result) {
			return -8;
		}

		std::string output_path = vm["output.folder"].as<std::string>();
		
		// Make sure output folder exists
		bcm::fix_path(output_path);
		if (!boost::filesystem::is_directory(output_path)) {
			std::string create_path = output_path.substr(0, output_path.size()-1);
			if (!boost::filesystem::create_directories(create_path)) {
				LOGERROR("Could not create output directory.");
				return false;
			}
		}

		// Open up a log file
		std::string log_filename = output_path + "log.txt";
		bcm::logger->SetLogToFile(bcm::Logger::Info, log_filename.c_str());
		bcm::logger->LogTime();
		
		// Create the sampler
		sampler = bcm::SamplerFactory::Create(vm, numthreads);
		if (!sampler) {
			return -2;
		}
		sampler->SetLatentVariableSet(varset);
		sampler->SetReentrantPrior(prior);
		sampler->SetReentrantLikelihood(likelihood);

		output.SetFile(output_path + "output.nc");
		std::vector<std::string> varnames(varset->GetNumVariables());
		for (size_t vi = 0; vi < varset->GetNumVariables(); vi++) {
			varnames[vi] = varset->GetVariableName(vi);
		}
		output.SetAdditionalOutput(likelihood);
		output.Initialize(sampler->GetNumSamples(), varnames);
		sampler->SetPosteriorSampleHandler(output.GetSampleFunctionPointer(), NULL);
		sampler->SetOutputValuesHandler(output.GetAdditionalOutputFunctionPointer(), NULL);
		sampler->SetOutputPath(output_path);
		
#if 0
		bcm::NetCDFDataFile output;
		if (!output.Open("output.nc", false)) {
			LOGERROR("Cannot open output file");
			return -10;
		}
		
		const size_t sample_count = 1000;
		std::vector<unsigned int> sampleix(sample_count);
		for (size_t i = 0; i < sample_count; i++) {
			sampleix[i] = (unsigned int)i;
		}
		
		std::vector<std::string> species;
		model->GetSimulatedSpeciesFullNames(species);
		
		output.CreateGroup("sampled_output");
		output.CreateDimension("sampled_output", "samples", sampleix);
		output.CreateDimension("sampled_output", "species", species);
		output.CreateDimension("sampled_output", "time", timepoints);
		output.CreateVariable("sampled_output", "trajectories", "samples", "species", "time");

		rng = new bcm::RNG_SFMT;

		for (size_t sample = 0; sample < sample_count; sample++) {
			if ((sample % 10) == 0) {
				printf("%u\n", sample);
			}

			bool success = false;
			std::vector<Real> variable_sample;
			variable_sample.resize(varset->GetNumVariables(), std::numeric_limits<Real>::quiet_NaN());
			while (!success) {
				prior->Sample(&variable_sample[0], rng);
				success = model->Simulate(&variable_sample[0], 0);
			}

			for (size_t si = 0; si < species.size(); si++) {
				VectorReal trajectory;
				model->GetOutput(species[si], 0, trajectory);
				for (size_t tpi = 0; tpi < timepoints.size(); tpi++) {
					output.PutValue("sampled_output", "trajectories", sample, si, tpi, trajectory(tpi));
				}
			}
		}

		output.Close();
#endif
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	// Everything has been set, initialize the sampler
	if (!sampler->Initialize()) {
		return -10;
	}

	printf("Starting sampler...\n");
	bcm::ProgressIndicatorConsole pic;
	sampler->SetProgressIndicator(&pic);

	// Run the sampler
	Real evidence = 0.0, evidence_sd = 0.0;
	result = sampler->Run(&evidence, &evidence_sd);

	printf("Sampling finished.\n");
	printf("Evidence: %g +/- %g\n", evidence, evidence_sd);

	sampler->Uninitialize();
	output.Close();

	bcm::safe_delete(sampler);
	bcm::safe_delete(likelihood);
	bcm::safe_delete(prior);
	bcm::safe_delete(varset);
	bcm::safe_delete(rng);
	
	if (result) {
		return 0;
	} else {
		return -20;
	}
}

int gen_variable_file(const po::variables_map& vm)
{
	SBMLModel* model = NULL;
	bool result = true;
	
	try {
		model = new SBMLModel(1);
		result = model->LoadSBML(vm["model"].as<std::string>());
		if (!result) {
			return -1;
		}

		std::vector<std::string> parameters;
		model->GetParameters(parameters);
		std::sort(parameters.begin(), parameters.end());
		std::vector<std::string> species_names;
		model->GetSimulatedSpeciesFullNames(species_names);
		std::sort(species_names.begin(), species_names.end());

		try {
			boost::property_tree::ptree pt;
			boost::property_tree::ptree& root = pt.put("variableset", "");
			for (size_t i = 0; i < parameters.size(); i++) {
				boost::property_tree::ptree& param = root.add("variable", "");
				param.put("<xmlattr>.name", parameters[i]);
				param.put("<xmlattr>.distribution", "uniform");
				param.put("<xmlattr>.lower", "-1.0");
				param.put("<xmlattr>.upper", "1.0");
			}
			
			for (size_t i = 0; i < species_names.size(); i++) {
				boost::property_tree::ptree& param = root.add("variable", "");
				param.put("<xmlattr>.name", std::string("initial_value_") + species_names[i]);
				param.put("<xmlattr>.distribution", "uniform");
				param.put("<xmlattr>.lower", "-1.0");
				param.put("<xmlattr>.upper", "1.0");
				param.put("<xmlattr>.tolerance", "1e-6");
			}

#if BOOST_VERSION / 100 % 1000 >= 58
			boost::property_tree::xml_parser::write_xml("prior.xml", pt, std::locale(), 
				boost::property_tree::xml_writer_make_settings<std::string>(' ', 1));
#else
			boost::property_tree::xml_parser::write_xml("prior.xml", pt, std::locale(), 
				boost::property_tree::xml_writer_make_settings<char>(' ', 1));
#endif
		} catch (boost::property_tree::ptree_error &e) {
			LOGERROR("Error creating property tree: %s", e.what());
			return false;
		}
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	return 0;
}

int gen_code_file(const po::variables_map& vm)
{
	SBMLModel* model = NULL;
	bcm::VariableSet* varset = NULL;
	bool result = true;
	
	try {
		model = new SBMLModel(1);
		result = model->LoadSBML(vm["model"].as<std::string>());
		if (!result) {
			return -1;
		}

		varset = new bcm::VariableSet;
		result = varset->LoadFromXML(vm["prior"].as<std::string>());
		if (!result) {
			return -2;
		}

		if (!model->SetVariableSet(varset)) {
			return -3;
		}
		if (!model->Initialize()) {
			return -4;
		}

		std::string code = model->GenerateCode();

		std::ofstream f;
		f.open("code.c", std::fstream::out);
		if (f.is_open()) {
			f << code;
			f.close();
		} else {
			LOGERROR("Unable to open output file");
			return -5;
		}

	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	bcm::safe_delete(model);
	bcm::safe_delete(varset);

	return 0;
}

int main(int argc, char* argv[])
{
	printf("BCM sbmlodeinf version %u.%u.%u\n", BCM_VERSION_MAJOR, BCM_VERSION_MINOR, BCM_VERSION_FIX);

	bcm::logger->SetLogToConsole(bcm::Logger::Error);
	
	int result = -1;
	try {
		po::options_description generic_options("Generic options", 120);
		generic_options.add_options()
			("help",															"produce help message")
			("numthreads,j", po::value<size_t>()->default_value(0),				"number of threads to use; set to 0 for using all available hardware concurrency")
		;

		po::options_description config_options("Configuration", 120);
		config_options.add_options()
			("model,m",										po::value<std::string>()->default_value("model.xml"),			"SBML file")
			("prior",										po::value<std::string>()->default_value("prior.xml"),			"File describing the priors for the parameters/initial conditions")
			("likelihood",									po::value<std::string>()->default_value("likelihood.xml"),		"File describing the likelihood functions")
			("output.folder",								po::value<std::string>()->default_value("output"),				"Folder in which output files will be generated, subfolder of the model directory.")
			("gen_variable_file",																							"Generate standard variableset file for the given sbml model")
			("gen_code_file",																								"Generate code for calculation of derivative/jacobian")
		;
		bcm::SamplerFactory::AddOptionsDescription(config_options);

		po::options_description cmdline_only_options("Command-line only options", 120);
		cmdline_only_options.add_options()
			("config_file,c", po::value<std::string>()->default_value("config.txt"), "Configuration file")
		;
		
		// Add all options together
		po::options_description cmdline_options("Allowed options", 120);
		cmdline_options.add(generic_options).add(cmdline_only_options).add(config_options);

		po::positional_options_description pod;
		pod.add("model", 1);
		
		// Parse command line
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(pod).run(), vm);

		// Handle commands
		if (vm.count("help")) {
			std::cout << cmdline_options << "\n";
			result = 0;
		} else if (vm.count("gen_variable_file")) {
			result = gen_variable_file(vm);
		} else if (vm.count("gen_code_file")) {
			result = gen_code_file(vm);
		} else {
			// Parse config file, which may have been specified on the command line, otherwise load the default.
			std::string config_file = vm["config_file"].as<std::string>();
			if (!config_file.empty()) {
				po::store(po::parse_config_file<char>(config_file.c_str(), config_options), vm);
			}

			result = run(vm);
		}
	} catch (po::error &e) {
		LOGERROR("Error processing program options: %s", e.what());
	}

	return result;
}
