#include "Utils.h"

#include "NetCDFDataFile.h"
#include "Prediction.h"
#include "SBMLModel.h"

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <fstream>
#include <boost/property_tree/xml_parser.hpp>

namespace po = boost::program_options;

int run(const po::variables_map& vm)
{
	SBMLModel* model = NULL;
	bcm::VariableSet* varset = NULL;
	bcm::NetCDFDataFile* samples = NULL;
	bcm::NetCDFDataFile* output = NULL;
	Prediction* prediction = NULL;
	bool result = true;
	
	try {
		// Determine the number of threads we should use
		size_t numthreads = vm["numthreads"].as<size_t>();
		if (numthreads <= 0) {
			numthreads = boost::thread::hardware_concurrency();
			printf("numthreads <= 0 specified, found %u hardware concurrency\n", (unsigned int)numthreads);
			if (numthreads <= 0) {
				numthreads = 1;
			}
		}
		
		// Load the variable names from the prior file
		varset = new bcm::VariableSet;
		result = varset->LoadFromXML(vm["prior"].as<std::string>());
		if (!result) {
			return -2;
		}

		// Load samples
		samples = new bcm::NetCDFDataFile();
		result = samples->Open(vm["samples"].as<std::string>(), netCDF::NcFile::read);
		if (!result) {
			return -3;
		}

		// Load prediction config file
		prediction = new Prediction();
		result = prediction->Load(vm["prediction"].as<std::string>());
		if (!result) {
			return -4;
		}

		// Load model
		model = new SBMLModel(numthreads);
		result = model->LoadSBML(prediction->GetModelFilename());
		if (!result) {
			return -5;
		}
		result = model->SetVariableSet(varset);
		if (!result) {
			return -6;
		}
		result = model->Initialize();
		if (!result) {
			return -7;
		}

		// Initialize output file
		output = new bcm::NetCDFDataFile();
		result = output->Open(vm["output"].as<std::string>(), netCDF::NcFile::replace);
		if (!result) {
			return -8;
		}
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	size_t sample_count;
	size_t variable_count;
	result = samples->GetDimensionSize("samples", "sample_ix", &sample_count);
	result &= samples->GetDimensionSize("samples", "variables", &variable_count);
	if (!result) {
		LOGERROR("Samples file does not appear to be a valid samples file");
		return -10;
	}
	if (variable_count != varset->GetNumVariables()) {
		LOGERROR("Samples file contains %d variables, but variable set expects %d variables", variable_count, varset->GetNumVariables());
		return -11;
	}
	
	std::vector<Real> simulated_timepoints;
	simulated_timepoints = prediction->GetSimulatedTimepoints();
	
	// Set up output file
	std::vector<unsigned int> sampleix(sample_count);
	for (size_t i = 0; i < sample_count; i++) {
		sampleix[i] = (unsigned int)i+1;
	}
	std::vector<std::string> variables(variable_count);
	for (size_t i = 0; i < variable_count; i++) {
		samples->GetValue("samples", "variables", i, variables[i]);
	}
	std::vector<std::string> species;
	model->GetSimulatedSpeciesFullNames(species);

	result &= output->CreateGroup("samples");
	result &= output->CreateDimension("samples", "sample_ix", sampleix);
	result &= output->CreateDimension("samples", "variables", variables);
	result &= output->CreateDimension("samples", "species", species);
	result &= output->CreateDimension("samples", "time", simulated_timepoints);
	result &= output->CreateVariable("samples", "variable_values", "sample_ix", "variables");
	result &= output->CreateVariable("samples", "weight", "sample_ix");
	result &= output->CreateVariable("samples", "trajectories", "sample_ix", "species", "time");

	printf("Starting simulation...\n");
	printf("Progres:   0%%");

	bcm::Timer timer;
	timer.Start();
	
	VectorReal sample(variable_count);
	for (size_t i = 0; i < sample_count; i++) {
		printf("\b\b\b\b%3u%%", (unsigned int)((100*i) / sample_count));

		// Load sample
		samples->GetValuesDim2("samples", "variable_values", i, 0, variable_count, sample);
		output->PutValuesDim2("samples", "variable_values", i, 0, sample);

		// Modify based on the prediction
		if (!prediction->SetConditionsAtModel(model, 0)) {
			return -15;
		}

		// Simulate model
		result = model->Simulate(sample.data(), simulated_timepoints, 0);
		if (result) {
			for (size_t si = 0; si < species.size(); ++si) {
				VectorReal trajectory;
				model->GetOutput(si, trajectory, 0);
				output->PutValuesDim3("samples", "trajectories", i, si, 0, trajectory);
			}
		} else {
			for (size_t si = 0; si < species.size(); ++si) {
				VectorReal trajectory = VectorReal::Constant(simulated_timepoints.size(), std::numeric_limits<Real>::quiet_NaN());
				output->PutValuesDim3("samples", "trajectories", i, si, 0, trajectory);
			}
		}
		
		output->PutValue("samples", "sample_ix", i, (unsigned int)i);

		Real weight;
		samples->GetValue("samples", "weight", i, &weight);
		output->PutValue("samples", "weight", i, weight);
	}

	printf("\b\b\b\b100%%\n");
	printf("Simulation finished.\n");
	
	Real simulation_time = timer.GetElapsedSeconds();
	printf("Simulation time: %g\n", simulation_time);
	printf("Evaluations per second: %g\n", sample_count / simulation_time);
	
	bcm::safe_delete(prediction);
	bcm::safe_delete(samples);
	bcm::safe_delete(output);
	bcm::safe_delete(model);
	bcm::safe_delete(varset);

	if (result) {
		return 0;
	} else {
		return -20;
	}
}

int main(int argc, char* argv[])
{
	printf("BCM sbmlodepred version %u.%u.%u\n", BCM_VERSION_MAJOR, BCM_VERSION_MINOR, BCM_VERSION_FIX);

	bcm::logger->SetLogToConsole(bcm::Logger::Error);
	
	int result = -1;
	try {
		po::options_description generic_options("Generic options", 120);
		generic_options.add_options()
			("help",															"produce help message")
			("numthreads,j", po::value<size_t>()->default_value(0),				"number of threads to use; set to 0 for using all available hardware concurrency")
		;

		po::options_description config_options("Configuration", 120);
		config_options.add_options()
			("prior",										po::value<std::string>()->default_value("prior.xml"),			"File describing the priors for the parameters/initial conditions")
			("samples,s",									po::value<std::string>()->default_value("samples.nc"),			"Samples to use for prediction")
			("prediction,p",								po::value<std::string>()->default_value("prediction.xml"),		"File describing what to predict")
			("output,o",									po::value<std::string>()->default_value("output.nc"),			"Output file")
		;

		po::options_description cmdline_only_options("Command-line only options", 120);
		cmdline_only_options.add_options()
			("config_file,c", po::value<std::string>()->default_value(""), "Configuration file")
		;
		
		// Add all options together
		po::options_description cmdline_options("Allowed options", 120);
		cmdline_options.add(generic_options).add(cmdline_only_options).add(config_options);

		po::positional_options_description pod;
		pod.add("model", 1);
		
		// Parse command line
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(pod).run(), vm);

		// Handle commands
		if (vm.count("help")) {
			std::cout << cmdline_options << "\n";
			result = 0;
		} else {
			// Parse config file, which may have been specified on the command line.
			std::string config_file = vm["config_file"].as<std::string>();
			if (!config_file.empty()) {
				po::store(po::parse_config_file<char>(config_file.c_str(), config_options), vm);
			}

			result = run(vm);
		}
	} catch (po::error &e) {
		LOGERROR("Error processing program options: %s", e.what());
	}

	return result;
}
