#include "Utils.h"
#include "SBMLPDLikelihood.h"
#include "NetCDFDataFile.h"
#include "ProbabilityDistributions.h"
#include "SignalingNetwork.h"
#include "SBMLPDExperiment.h"

#include <boost/property_tree/xml_parser.hpp>

SBMLPDLikelihood::SBMLPDLikelihood(size_t numthreads)
	: numthreads(numthreads)
	, network(NULL)
{
}

SBMLPDLikelihood::~SBMLPDLikelihood()
{
	bcm::safe_delete(network);
	for (size_t i = 0; i < Experiments.size(); i++) {
		delete Experiments[i];
	}
}

bool SBMLPDLikelihood::SetVariableSet(const bcm::VariableSet* varset)
{
	VarSet = varset;
	return true;
}

bool SBMLPDLikelihood::Initialize(const std::string& likelihood_file)
{
	if (!VarSet) {
		LOGERROR("Variable Set has not been specified");
		return false;
	}

	bool result = true;

	// Load likelihood file
	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(likelihood_file, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading likelihood file: %s", e.what());
		return false;
	}
	
	try {
		boost::property_tree::ptree likelihood_node = pt.get_child("sbml_pd_likelihood");

		// Load the model
		boost::property_tree::ptree model = likelihood_node.get_child("model");
		std::string modelfn = model.get<std::string>("<xmlattr>.file");

		network = new SignalingNetwork(numthreads);
		if (!network->Initialize(modelfn, VarSet)) {
			return false;
		}
		
		// Load all experiments
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& data, likelihood_node.get_child("")) {
			if (data.first == "experiment") {
				SBMLPDExperiment* experiment = SBMLPDExperiment::Create(data.second, network, VarSet, Experiments, numthreads);
				if (!experiment) {
					return false;
				}
				Experiments.push_back(experiment);
			}
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	return result;
}

bool SBMLPDLikelihood::SetupAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_count)
{
	bool result = true;

	std::vector<unsigned int> sampleix(sample_count);
	for (size_t i = 0; i < sample_count; i++) {
		sampleix[i] = (unsigned int)i+1;
	}

	std::vector<std::string> signaling_molecule_names(network->GetMoleculeCount());
	for (size_t i = 0; i < signaling_molecule_names.size(); i++) {
		signaling_molecule_names[i] = network->GetMoleculeName(i);
	}

	result &= file->CreateGroup("output_values");
	result &= file->CreateDimension("output_values", "sample_ix", sampleix);
	result &= file->CreateDimension("output_values", "signaling_molecules", signaling_molecule_names);

	for (size_t i = 0; i < Experiments.size(); i++) {
		result &= Experiments[i]->SetupAdditionalOutput(file, sample_count);
	}

	return result;
}

void SBMLPDLikelihood::HandleAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_ix, const Real* values)
{
	size_t value_ix = 0;
	
	for (size_t i = 0; i < Experiments.size(); i++) {
		size_t experiment_numvalues = 0;
		Experiments[i]->HandleAdditionalOutput(file, sample_ix, values + value_ix, experiment_numvalues);
		value_ix += experiment_numvalues;
	}
}

bool SBMLPDLikelihood::EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
{
	logp = 0.0;
	
	for (size_t i = 0; i < Experiments.size(); i++) {
		Real experiment_logp = 0.0;
		bool result = Experiments[i]->EvaluateLogProbability(threadix, values, experiment_logp, Experiments);
		if (result) {
			logp += experiment_logp;
			if (experiment_logp == -std::numeric_limits<Real>::infinity()) {
				break;
			}
		} else {
			return false;
		}
	}

	return true;
}

bool SBMLPDLikelihood::GetOutput(size_t threadix, Real* output_values, size_t& numvalues)
{
	bool result = true;
	if (output_values) {
		size_t ix = 0;
		for (size_t i = 0; i < Experiments.size(); i++) {
			size_t experiment_numvalues = 0;
			result &= Experiments[i]->GetOutput(threadix, output_values + ix, experiment_numvalues);
			ix += experiment_numvalues;
		}
	} else {
		numvalues = 0;
		for (size_t i = 0; i < Experiments.size(); i++) {
			result &= Experiments[i]->GetOutput(threadix, output_values, numvalues);
		}
	}
	return result;
}
