#include "Utils.h"
#include "SignalingNetwork.h"

#include <sbml/SBMLTypes.h>
#include <boost/graph/strong_components.hpp>
#include <boost/graph/transitive_closure.hpp>
#include <boost/algorithm/string.hpp>

#define SINGLE_DOSE_DRUG 0

SignalingNetwork::SignalingNetwork(size_t numthreads)
	: numthreads(numthreads)
	, VarSet(NULL)
	, num_strongly_connected_components(0)
{
}

SignalingNetwork::~SignalingNetwork()
{
}

bool SignalingNetwork::Initialize(const std::string& sbml_file, const bcm::VariableSet* varset)
{
	VarSet = varset;

	SBMLDocument* document = NULL;
	const Model* model = NULL;

	// Load SBML file
	document = readSBML(sbml_file.c_str());
	if (!document) {
		LOGERROR("Errors reading SBML file %s: NULL return value");
		return false;
	}
	if (document->getNumErrors() > 0) {
		LOGERROR("Errors reading SBML file %s: %d", sbml_file.c_str(), document->getNumErrors());
		for (unsigned int i = 0; i < document->getNumErrors(); i++) {
			LOGERROR(" %d: %s", i, document->getError(i)->getMessage().c_str());
		}
		delete document;
		return false;
	}

	model = document->getModel();

	for (unsigned int i = 0; i < model->getNumSpecies(); i++) {
		const Species* sp = model->getSpecies(i);

		SignalingMolecule s;
		s.id = sp->getId();
		s.name = sp->getName();

		const XMLNode* annotation = sp->getAnnotation();
		for (unsigned int ai = 0; ai < annotation->getNumChildren(); ai++) {
			const XMLNode& child = annotation->getChild(ai);
			if (child.getPrefix() == "celldesigner" && child.getName() == "extension") {
				std::string species_class = child.getChild("speciesIdentity").getChild("class").getChild(0).getCharacters();
				if (species_class.empty()) {
					LOGERROR("Can't find species class for species %s [%s] - SBML file should be a CellDesigner file", sp->getId().c_str(), sp->getName().c_str());
					delete document;
					return false;
				}

				if (species_class == "PROTEIN") {
					s.type = SignalingMolecule::Protein;
				} else if (species_class == "SIMPLE_MOLECULE") {
					s.type = SignalingMolecule::SmallMolecule;
				} else if (species_class == "GENE") {
					s.type = SignalingMolecule::Mutation;
				} else if (species_class == "DRUG") {
					s.type = SignalingMolecule::Drug;
				} else if (species_class == "PHENOTYPE") {
					s.type = SignalingMolecule::Phenotype;
				} else if (species_class == "UNKNOWN") {
					s.type = SignalingMolecule::Unknown;
				} else {
					LOGERROR("Unrecognized species type %s for species %s [%s]", species_class.c_str(), sp->getId().c_str(), sp->getName().c_str());
					delete document;
					return false;
				}
			}
		}

		if (s.type == SignalingMolecule::Drug) {
			const XMLNode* notes = sp->getNotes();
			const XMLNode& child = notes->getChild(0);
			const XMLNode& child1 = child.getChild(1);
			const XMLNode& child2 = child1.getChild(0);

			std::string inhibition_type_note = child2.getCharacters();
			boost::trim(inhibition_type_note);
			
			if (inhibition_type_note == "inhibit activity") {
				s.drug_inhibition_type = SignalingMolecule::InhibitActivity;
			} else if (inhibition_type_note == "inhibit activation") {
				s.drug_inhibition_type = SignalingMolecule::InhibitActivation;
			} else if (inhibition_type_note == "") {
				LOGERROR("The drug \"%s\" does not have a note to specify whether the drug inhibits the activity or the activation of its targets. Please add a \"note\" to the species in the SBML file specifying either \"inhibit activity\" or \"inhibit activation\".", s.name.c_str());
				delete document;
				return false;
			} else {
				LOGERROR("The drug \"%s\" has notes, but the notes do not appear to specify whether the drug inhibits the activity or the activation of its targets. Please add a \"note\" to the species in the SBML file specifying either \"inhibit activity\" or \"inhibit activation\".", s.name.c_str());
				delete document;
				return false;
			}
		} else {
			s.drug_inhibition_type = SignalingMolecule::NotADrug;
		}

		signaling_molecules.push_back(s);
	}

	for (unsigned int i = 0; i < model->getNumReactions(); i++) {
		const Reaction* re = model->getReaction(i);

		const XMLNode* annotation = re->getAnnotation();
		bool activating = true;
		for (unsigned int i = 0; i < annotation->getNumChildren(); i++) {
			const XMLNode& child = annotation->getChild(i);
			if (child.getPrefix() == "celldesigner" && child.getName() == "extension") {
				std::string reaction_type_str = child.getChild("reactionType").getChild(0).getCharacters();
				if (reaction_type_str.empty()) {
					LOGERROR("Can't find reaction type for reaction %s - SBML file should be a CellDesigner file", re->getId().c_str());
					delete document;
					return false;
				}
			
				if (reaction_type_str == "POSITIVE_INFLUENCE") {
					activating = true;
				} else if (reaction_type_str == "NEGATIVE_INFLUENCE") {
					activating = false;
				} else {
					LOGERROR("Unrecognized reaction type %s for reaction %s - SBML file should use reduced notation", reaction_type_str.c_str(), re->getId().c_str());
					delete document;
					return false;
				}
			}
		}

		if (re->getNumReactants() != 1) {
			LOGERROR("Can only handle reactions with 1 reactant");
			delete document;
			return false;
		}
		if (re->getNumProducts() != 1) {
			LOGERROR("Can only handle reactions with 1 product");
			delete document;
			return false;
		}

		size_t reactant_ix = GetSignalingMoleculeIxById(re->getReactant(0)->getSpecies());
		if (reactant_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Unknown reactant \"%s\"", re->getReactant(0)->getSpecies().c_str());
			delete document;
			return false;
		}
		size_t product_ix = GetSignalingMoleculeIxById(re->getProduct(0)->getSpecies());
		if (product_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Unknown product \"%s\"", re->getProduct(0)->getSpecies().c_str());
			delete document;
			return false;
		}

		signaling_molecules[product_ix].parents.push_back(reactant_ix);
		signaling_molecules[product_ix].parents_activating.push_back(activating);

		// TODO - check for duplicate reactions?
	}

	// Set up variable mapping
	bool result = true;
	size_t drug_inhibition_counter = 0;
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		SignalingMolecule& sm = signaling_molecules[i];
		
		if (sm.type != SignalingMolecule::Mutation && sm.type != SignalingMolecule::Drug && sm.type != SignalingMolecule::Unknown) {
			result &= GetVariableIx(varset, std::string("base_") + sm.name, sm.base_parameter_ix);
		}

		//if (sm.type == SignalingMolecule::Drug && sm.name == drug) {
		//	drug_signaling_ix = i;
		//}

		sm.parents_strength_parameter_ix.resize(sm.parents.size(), std::numeric_limits<size_t>::max());
		sm.parents_ic50_parameter_ix.resize(sm.parents.size(), std::numeric_limits<size_t>::max());
		sm.parents_maxinhib_parameter_ix.resize(sm.parents.size(), std::numeric_limits<size_t>::max());
		sm.parents_logsteepness_parameter_ix.resize(sm.parents.size(), std::numeric_limits<size_t>::max());
		sm.drug_inhibition_ix.resize(sm.parents.size(), std::numeric_limits<size_t>::max());
		
		for (size_t j = 0; j < sm.parents.size(); j++) {
			size_t parent_ix = sm.parents[j];
			const SignalingMolecule& parent = signaling_molecules[parent_ix];
			if (parent.type == SignalingMolecule::Drug) {
#if !SINGLE_DOSE_DRUG
				result &= GetVariableIx(varset, std::string("ic50_") + parent.name + std::string("_") + sm.name, sm.parents_ic50_parameter_ix[j]);
				result &= GetVariableIx(varset, std::string("logsteepness_") + parent.name + std::string("_") + sm.name, sm.parents_logsteepness_parameter_ix[j]);
#endif
				result &= GetVariableIx(varset, std::string("maxinhib_") + parent.name + std::string("_") + sm.name, sm.parents_maxinhib_parameter_ix[j]);
				sm.drug_inhibition_ix[j] = drug_inhibition_counter++;
			} else {
				result &= GetVariableIx(varset, std::string("strength_") + parent.name + std::string("_") + sm.name, sm.parents_strength_parameter_ix[j]);
			}
		}
	}
	if (!result) {
		delete document;
		return false;
	}

	// Calculate number of inputs factors
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		SignalingMolecule& sm = signaling_molecules[i];

		size_t activating_input_count = 0;
		size_t inhibiting_input_count = 0; // excl drugs
		
		for (size_t j = 0; j < sm.parents.size(); j++) {
			size_t parent_ix = sm.parents[j];
			const SignalingMolecule& parent = signaling_molecules[parent_ix];
			if (parent.type == SignalingMolecule::Drug) {
				if (sm.parents_activating[j]) {
					activating_input_count++;
				}
			} else {
				if (sm.parents_activating[j]) {
					activating_input_count++;
				} else {
					inhibiting_input_count++;
				}
			}
		}

#if 0
		if (activating_input_count > inhibiting_input_count) {
			size_t excess = activating_input_count - inhibiting_input_count;
			sm.number_of_inputs_factor = 1.0 / (Real)excess;
		} else if (activating_input_count == inhibiting_input_count) {
			sm.number_of_inputs_factor = 1.0;
		} else {
			size_t excess = inhibiting_input_count - activating_input_count;
			sm.number_of_inputs_factor = (Real)excess;
		}
#else
		sm.number_of_inputs_factor = 1.0;
#endif
	}

	// Build a graph structure of the network
	ConstructGraph();

	// Find strongly connected components with Tarjan's algorithm
	component_assignment.resize(signaling_molecules.size());
	num_strongly_connected_components = boost::strong_components(graph, boost::make_iterator_property_map(component_assignment.begin(), boost::get(boost::vertex_index, graph)));
	
	// The strongly connected components are in reverse topological order; reverse the indices to get a topological ordering
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		component_assignment[i] = num_strongly_connected_components - component_assignment[i] - 1;
	}

#if 0
	// Remove mutations - these do not need to be calculated
	std::vector<size_t> remove_ix;
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		const SignalingMolecule& sm = signaling_molecules[i];
		//if (sm.type != SignalingMolecule::Mutation && sm.type != SignalingMolecule::Drug && sm.type != SignalingMolecule::Unknown) {
		if (sm.type == SignalingMolecule::Mutation) {
			size_t component_ix = component_assignment[i];
			remove_ix.push_back(component_ix);
			component_assignment[i] = std::numeric_limits<size_t>::max();
		}
	}
#endif

	// Find the components that contain more than 1 node - these will be solved as systems of equations
	component_size.resize(num_strongly_connected_components);
	component_to_signaling_ix.resize(num_strongly_connected_components);
	std::set<size_t> solve_components;
	for (size_t i = 0; i < num_strongly_connected_components; i++) {
		component_size[i] = std::count(component_assignment.begin(), component_assignment.end(), i);
		if (component_size[i] > 1) {
			solve_components.insert(i);
			component_to_signaling_ix[i] = std::numeric_limits<size_t>::max();
		} else {
			component_to_signaling_ix[i] = std::distance(component_assignment.begin(), std::find(component_assignment.begin(), component_assignment.end(), i));
		}
	}

	// Allocate structure for system solving
	parallel_data.resize(numthreads);
	for (size_t ti = 0; ti < numthreads; ti++) {
		parallel_data[ti].systems.resize(solve_components.size());

		size_t system_ix = 0;
		for (size_t i = 0; i < num_strongly_connected_components; i++) {
			if (component_size[i] > 1) {
				size_t n = component_size[i];

				NonlinearSystem& system = parallel_data[ti].systems[system_ix];
				system.jac = MatrixReal::Zero(n, n);
				system.vx = VectorReal::Constant(n, 0.5);
				system.fx = VectorReal::Constant(n, 0.0);
				system.deltax = VectorReal::Constant(n, 0.0);
				system.signaling_ix.resize(n);
				size_t j = 0;
				for (size_t k = 0; k < signaling_molecules.size(); k++) {
					if (component_assignment[k] == i) {
						system.signaling_ix[j] = k;
						j++;
					}
				}
				ASSERT(j == n);

				if (n > 4) {
					//parallel_data[ti].systems[system_ix].lu = 
				} else {
					system.inverse = MatrixReal::Zero(n, n);
				}


				system_ix++;
			}
		}

		parallel_data[ti].drug_inhibition.resize(drug_inhibition_counter, std::numeric_limits<Real>::quiet_NaN());
	}

	// Performance optimizations
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		const SignalingMolecule& sm = signaling_molecules[i];
		if (sm.type == SignalingMolecule::Drug) {
			drug_molecule_indices.push_back(i);
		}
	}

	delete document;
	return true;
}

bool SignalingNetwork::Calculate(size_t threadix, VectorReal& activities, VectorReal& expression, const Real* values)
{
	// Precalculate drug inhibition effects
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		const SignalingMolecule& sm = signaling_molecules[i];
		for (size_t j = 0; j < sm.parents.size(); j++) {
			size_t parent_ix = sm.parents[j];
			const SignalingMolecule& parent = signaling_molecules[parent_ix];

			if (parent.type == SignalingMolecule::Drug) {
				Real signal;
				if (activities[parent_ix] == 0.0) {
					if (sm.parents_activating[j]) {
						signal = 0.0;
					} else {
						signal = 1.0;
					}
				} else {
#if SINGLE_DOSE_DRUG
					Real maxinhib = GetValue(values, sm.parents_maxinhib_parameter_ix[j]);

					if (sm.parents_activating[j]) {
						// Activating drug
						// maxinhib really means max activation here
						signal = maxinhib;
					} else {
						// Inhibiting drug
						signal = maxinhib;
					}
#else
					Real ic50 = GetValue(values, sm.parents_ic50_parameter_ix[j]);
					Real maxinhib = GetValue(values, sm.parents_maxinhib_parameter_ix[j]);
					Real steepness = bcm::fastpow10(GetValue(values, sm.parents_logsteepness_parameter_ix[j]));
					Real log_concentration = log10(activities[parent_ix]);

					if (sm.parents_activating[j]) {
						// Activating drug
						// maxinhib really means max activation here
						signal = maxinhib - maxinhib / (bcm::fastpow10(steepness * (log_concentration - ic50)) + 1);
					} else {
						// Inhibiting drug
						signal = maxinhib + (1 - maxinhib) / (bcm::fastpow10(steepness * (log_concentration - ic50)) + 1);
					}
#endif
				}
				parallel_data[threadix].drug_inhibition[sm.drug_inhibition_ix[j]] = signal;
			}
		}
	}

	// Calculate network activities
	size_t system_ix = 0;
	for (size_t i = 0; i < num_strongly_connected_components; i++) {
		size_t ix = component_to_signaling_ix[i];
		if (ix == std::numeric_limits<size_t>::max()) {
			// Component contains multiple signaling molecules, we have to solve a system
			NonlinearSystem& system = parallel_data[threadix].systems[system_ix];
			if (!SolveSystem(system, activities, expression, values, threadix)) {
				return false;
			}
			system_ix++;
		} else {
			// Component contains a single signaling molecule, calculate activity directly
			const SignalingMolecule& sm = signaling_molecules[ix];

			if (activities(ix) == activities(ix)) {
				// Activities already set. This is only supported for nodes without parents.
				ASSERT(sm.parents.empty());
#if 0 // Attempt linear proliferation node.
			} else if (sm.type == SignalingMolecule::Phenotype) {
				Real input, inhibition;
				CalculateActivationInput(sm, activities, values, threadix, input, inhibition);
				Real base = GetValue(values, sm.base_parameter_ix);
				Real proliferation = (std::max)(0.0, base + input);
				activities(ix) = proliferation * inhibition;
#endif
			} else {
				Real input, inhibition;
				CalculateActivationInput(sm, activities, values, threadix, input, inhibition);
#if 0
				activities(ix) = bcm::logistic_activation_fixed_ks(GetValue(values, sm.base_parameter_ix), input) * inhibition * expression(ix);
#else
				input += GetValue(values, sm.base_parameter_ix);
				activities(ix) = bcm::logistic_activation_fixed_bks(input) * inhibition * expression(ix);
#endif
			}
		}
	}

	return true;
}

void SignalingNetwork::SetDrugConcentrations(VectorReal& activities, Real concentration)
{
	for (std::vector<size_t>::const_iterator dmi = drug_molecule_indices.cbegin(); dmi != drug_molecule_indices.cend(); ++dmi) {
		activities(*dmi) = concentration;
	}
}

size_t SignalingNetwork::GetMoleculeCount() const
{
	return signaling_molecules.size();
}

std::string SignalingNetwork::GetMoleculeName(size_t ix) const
{
	if (ix < signaling_molecules.size()) {
		return signaling_molecules[ix].name;
	} else {
		LOGERROR("Invalid molecule index");
		return std::string();
	}
}

size_t SignalingNetwork::GetSignalingMoleculeIxById(const std::string& id)
{
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		if (signaling_molecules[i].id == id) {
			return i;
		}
	}
	return std::numeric_limits<size_t>::max();
}

size_t SignalingNetwork::GetSignalingMoleculeIxByName(const std::string& name)
{
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		if (signaling_molecules[i].name == name) {
			return i;
		}
	}
	return std::numeric_limits<size_t>::max();
}

void SignalingNetwork::ConstructGraph()
{
	for (size_t i = 0; i < signaling_molecules.size(); i++) {
		const SignalingMolecule& sm = signaling_molecules[i];
		for (size_t j = 0; j < sm.parents.size(); j++) {
			boost::add_edge(sm.parents[j], i, graph);
		}
	}
}

void SignalingNetwork::CalculateActivationInput(const SignalingMolecule& sm, const VectorReal& activities, const Real* values, size_t threadix, Real& sum, Real& inhibition) const
{
	// Add signal from all parents which are not inhibiting drugs
	sum = 0.0;
	inhibition = 1.0;
	for (size_t j = 0; j < sm.parents.size(); j++) {
		size_t parent_ix = sm.parents[j];
		const SignalingMolecule& parent = signaling_molecules[parent_ix];

		if (parent.type == SignalingMolecule::Drug) {
			if (sm.parents_activating[j]) {
				Real signal = parallel_data[threadix].drug_inhibition[sm.drug_inhibition_ix[j]];
				sum += signal;
			} else {
				if (parent.drug_inhibition_type == SignalingMolecule::InhibitActivation || sm.name == "proliferation") {
					Real signal = parallel_data[threadix].drug_inhibition[sm.drug_inhibition_ix[j]];
					inhibition *= signal;
				} else if (parent.drug_inhibition_type == SignalingMolecule::InhibitActivity) {
					// Drugs inhibiting their targets activity only inhibit the downstream signal of their targets, thus they do not affect their immediate target directly
				}
			}
		} else {
			Real signal;
			if (sm.parents_activating[j]) {
				signal = GetValue(values, sm.parents_strength_parameter_ix[j]) * activities(parent_ix) * sm.number_of_inputs_factor;
			} else {
				signal = -GetValue(values, sm.parents_strength_parameter_ix[j]) * activities(parent_ix);
			}

			// Is the parent inhibited by a drug?
			for (size_t k = 0; k < parent.parents.size(); k++) {
				size_t parent_parent_ix = parent.parents[k];
				const SignalingMolecule& parent_parent = signaling_molecules[parent_parent_ix];
				if (parent_parent.type == SignalingMolecule::Drug &&
					parent_parent.drug_inhibition_type == SignalingMolecule::InhibitActivity &&
					activities(parent_parent_ix) > 0 &&
					!parent.parents_activating[k])
				{
					signal *= parallel_data[threadix].drug_inhibition[parent.drug_inhibition_ix[k]];
				}
			}
			sum += signal;
		}
	}

	//sum *= sm.number_of_inputs_factor;
}


inline Real square(Real a) { return a*a; }

inline Real logistic_deriv_part(Real sum)
{
	if (sum < -20 || sum > 20) {
		return 0.0;
	} else {
		Real g = exp(sum);
		return g / square(g + 1.0);
	}
}

inline Real logistic_activation_fixed_bks_deriv(Real sum, Real sum_d)
{
	return 9.19024 * sum_d * logistic_deriv_part(-9.19024 * (sum - 0.5));
}

inline Real logistic_activation_fixed_ks_deriv(Real b, Real sum, Real sum_d)
{
	return 9.19024 * sum_d * (1.0 - b) * logistic_deriv_part(-9.19024 * (sum - 0.5));
}

bool SignalingNetwork::SolveSystem(NonlinearSystem& system, VectorReal& activities, VectorReal& expression, const Real* values, size_t threadix) const
{
	static const size_t MAX_NEWTON_ITERATIONS = 10;
	static const Real TOLERANCE = 1e-5;
	
	bool converged = false;

	// The input activities vector is used as starting point; however, check that it contains reasonable values
	for (size_t i = 0; i < system.signaling_ix.size(); i++) {
		size_t ix = system.signaling_ix[i];
		const SignalingMolecule& sm = signaling_molecules[ix];
		if (activities(ix) != activities(ix) || activities(ix) < 0.0 || activities(ix) > 1.0) {
			activities(ix) = 0.5;
		}
		system.vx(i) = activities(ix);
	}
	
	for (size_t ni = 0; ni < MAX_NEWTON_ITERATIONS; ni++) {
		// Calculate F and jacobian
		for (size_t i = 0; i < system.signaling_ix.size(); i++) {
			const SignalingMolecule& sm = signaling_molecules[system.signaling_ix[i]];

			Real input, inhibition;
			CalculateActivationInput(sm, activities, values, threadix, input, inhibition);

			// F
#if 0
			system.fx(i) = bcm::logistic_activation_fixed_ks(GetValue(values, sm.base_parameter_ix), input) * inhibition * expression(system.signaling_ix[i]) - system.vx(i);
#else
			input += GetValue(values, sm.base_parameter_ix);
			system.fx(i) = bcm::logistic_activation_fixed_bks(input) * inhibition * expression(system.signaling_ix[i]) - system.vx(i);
#endif

			// Jac
			for (size_t j = 0; j < system.signaling_ix.size(); j++) {
				if (i == j) {
					system.jac(i,j) = -1.0;
				} else {
					std::vector<size_t>::const_iterator pi = std::find(sm.parents.begin(), sm.parents.end(), system.signaling_ix[j]);
					if (pi == sm.parents.end()) {
						system.jac(i,j) = 0.0;
					} else {
						const SignalingMolecule& smj = signaling_molecules[*pi];

						Real signal;
						size_t pix = std::distance(sm.parents.begin(), pi);
						ASSERT(smj.type != SignalingMolecule::Drug || smj.parents_activating[*pi]);
						if (sm.parents_activating[pix]) {
							signal = GetValue(values, sm.parents_strength_parameter_ix[pix]) * activities(*pi) * sm.number_of_inputs_factor;
						} else {
							signal = -GetValue(values, sm.parents_strength_parameter_ix[pix]) * activities(*pi);
						}

						// Is this parent inhibited by a drug?
						for (size_t k = 0; k < smj.parents.size(); k++) {
							size_t parent_parent_ix = smj.parents[k];
							const SignalingMolecule& parent_parent = signaling_molecules[parent_parent_ix];
							if (parent_parent.type == SignalingMolecule::Drug &&
								parent_parent.drug_inhibition_type == SignalingMolecule::InhibitActivity &&
								activities(parent_parent_ix) > 0 &&
								!smj.parents_activating[k])
							{
								signal *= parallel_data[threadix].drug_inhibition[smj.drug_inhibition_ix[k]];
							}
						}

#if 0
						system.jac(i,j) = logistic_activation_fixed_ks_deriv(GetValue(values, sm.base_parameter_ix), input, signal) * inhibition * expression(system.signaling_ix[i]);
#else
						system.jac(i,j) = logistic_activation_fixed_bks_deriv(input, signal) * inhibition * expression(system.signaling_ix[i]);
#endif
					}
				}
			}
		}

		// Solve linear system
		// For size 2-4, use matrix inverse; otherwise LU-decomposition
		switch(system.signaling_ix.size()) {
		case 2:
			Eigen::internal::compute_inverse<MatrixReal, MatrixReal, 2>::run(system.jac, system.inverse);
			system.deltax.noalias() = system.inverse * system.fx;
			break;

		case 3:
			Eigen::internal::compute_inverse<MatrixReal, MatrixReal, 3>::run(system.jac, system.inverse);
			system.deltax.noalias() = system.inverse * system.fx;
			break;

		case 4:
			Eigen::internal::compute_inverse_size4<Eigen::Architecture::SSE, double, MatrixReal, MatrixReal>::run(system.jac, system.inverse);
			system.deltax.noalias() = system.inverse * system.fx;
			break;

		default:
			system.lu.compute(system.jac);
			system.deltax.noalias() = system.lu.solve(system.fx);
			break;
		}

		// Calculate new x
		system.vx -= system.deltax;
		
		// Update activities vector (this is used in F and Jac calculation)
		for (size_t i = 0; i < system.signaling_ix.size(); i++) {
			size_t ix = system.signaling_ix[i];
			const SignalingMolecule& sm = signaling_molecules[ix];
			activities(ix) = system.vx(i);
		}

		// Check whether the change for all x was less than the tolerance
		int conv_count = 0;
		for (int i = 0; i < system.signaling_ix.size(); i++) {
			if (fabs(system.deltax(i)) < TOLERANCE) {
				++conv_count;
			}
		}
		if (conv_count == system.signaling_ix.size()) {
			converged = true;
			break;
		}
	}

	return converged;
}

bool SignalingNetwork::GetVariableIx(const bcm::VariableSet* varset, const std::string& varstr, size_t& ix)
{
	ix = varset->GetVariableIndex(varstr);
	if (ix == std::numeric_limits<size_t>::max()) {
		return false;
	} else {
		return true;
	}
}
