#pragma once

#include "VariableSet.h"
#include <boost/graph/adjacency_list.hpp>

class SignalingNetwork
{
public:
	SignalingNetwork(size_t numthreads);
	~SignalingNetwork();

	bool Initialize(const std::string& sbml_file, const bcm::VariableSet* varset);
	bool Calculate(size_t threadix, VectorReal& activities, VectorReal& expression, const Real* values);
	void SetDrugConcentrations(VectorReal& activities, Real concentration);

	size_t GetMoleculeCount() const;
	std::string GetMoleculeName(size_t ix) const;
	size_t GetSignalingMoleculeIxById(const std::string& id);
	size_t GetSignalingMoleculeIxByName(const std::string& name);

private:
	struct SignalingMolecule {
		enum EType {
			Protein,
			SmallMolecule,
			Mutation,
			Drug,
			Phenotype,
			Unknown,
		};
		enum EDrugInhibitionType {
			InhibitActivity,
			InhibitActivation,
			NotADrug,
		};

		std::string id;
		std::string name;
		EType type;
		EDrugInhibitionType drug_inhibition_type;
		size_t base_parameter_ix;
		Real number_of_inputs_factor;

		std::vector<size_t> parents;
		std::vector<size_t> parents_strength_parameter_ix;
		std::vector<size_t> parents_ic50_parameter_ix;
		std::vector<size_t> parents_maxinhib_parameter_ix;
		std::vector<size_t> parents_logsteepness_parameter_ix;
		std::vector<size_t> drug_inhibition_ix;
		std::vector<bool> parents_activating;
	};

	struct NonlinearSystem {
		Eigen::PartialPivLU< MatrixReal > lu;
		VectorReal vx, fx, deltax;
		MatrixReal jac;
		MatrixReal inverse;
		std::vector<size_t> signaling_ix;
	};
	
	struct ParallelData {
		std::vector<NonlinearSystem> systems;
		std::vector<Real> drug_inhibition;
		std::vector<Real> expression;
	};

	void ConstructGraph();
	void CalculateActivationInput(const SignalingMolecule& sm, const VectorReal& activities, const Real* values, size_t threadix, Real& sum, Real& inhibition) const;
	bool SolveSystem(NonlinearSystem& system, VectorReal& activities, VectorReal& expression, const Real* values, size_t threadix) const;
	bool GetVariableIx(const bcm::VariableSet* varset, const std::string& varstr, size_t& ix);
	
	inline Real GetValue(const Real* values, size_t varix) const {
		ASSERT(varix < VarSet->GetNumVariables());
		return values[varix];
	}

	size_t numthreads;
	std::vector<ParallelData> parallel_data;
	
	const bcm::VariableSet* VarSet;

	std::vector<SignalingMolecule> signaling_molecules;

	boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS> graph;
	size_t num_strongly_connected_components;
	std::vector<size_t> component_assignment;
	std::vector<size_t> component_size;
	std::vector<size_t> component_to_signaling_ix;

	// Optimizations
	std::vector<size_t> drug_molecule_indices;
};
