#include "Utils.h"

#include "SBMLPDLikelihood.h"
#include "OutputHandlerNetCDF.h"
#include "Prior.h"
#include "ProgressIndicatorConsole.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "UnivariateMarginal.h"
#include "VariableSet.h"
#include "RNG_SFMT.h"

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <fstream>
#include <boost/property_tree/xml_parser.hpp>
#include <sbml/SBMLTypes.h>

static const int version_major = 0;
static const int version_minor = 1;
static const int version_fix = 0;

namespace po = boost::program_options;

int run(const po::variables_map& vm)
{
	bcm::VariableSet* varset = NULL;
	bcm::Prior* prior = NULL;
	bcm::RNG* rng = NULL;
	bcm::Sampler* sampler = NULL;
	SBMLPDLikelihood* likelihood = NULL;
	bcm::OutputHandlerNetCDF output;
	bool result = true;
	
	try {
		// Determine the number of threads we should use
		size_t numthreads = vm["numthreads"].as<size_t>();
		if (numthreads <= 0) {
			numthreads = boost::thread::hardware_concurrency();
			printf("numthreads <= 0 specified, found %u hardware concurrency\n", (unsigned int)numthreads);
			if (numthreads <= 0) {
				numthreads = 1;
			}
		}
		
		varset = new bcm::VariableSet;
		result = varset->LoadFromXML(vm["prior"].as<std::string>());
		if (!result) {
			return -2;
		}

		prior = bcm::Prior::Create(vm["prior"].as<std::string>(), varset);
		if (!prior) {
			return -3;
		}

		likelihood = new SBMLPDLikelihood(numthreads);
		result = likelihood->SetVariableSet(varset);
		if (!result) {
			return -6;
		}
		result = likelihood->Initialize(vm["likelihood"].as<std::string>());
		if (!result) {
			return -8;
		}

		std::string output_path = vm["output.folder"].as<std::string>();
		
		// Make sure output folder exists
		bcm::fix_path(output_path);
		if (!boost::filesystem::is_directory(output_path)) {
			std::string create_path = output_path.substr(0, output_path.size()-1);
			if (!boost::filesystem::create_directories(create_path)) {
				LOGERROR("Could not create output directory.");
				return false;
			}
		}

		// Open up a log file
		std::string log_filename = output_path + "log.txt";
		bcm::logger->SetLogToFile(bcm::Logger::Info, log_filename.c_str());
		bcm::logger->LogTime();
		
		// Create the sampler
		sampler = bcm::SamplerFactory::Create(vm, numthreads);
		if (!sampler) {
			return -2;
		}
		sampler->SetLatentVariableSet(varset);
		sampler->SetReentrantPrior(prior);
		sampler->SetReentrantLikelihood(likelihood);

		output.SetFile(output_path + "output.nc");
		std::vector<std::string> varnames(varset->GetNumVariables());
		for (size_t vi = 0; vi < varset->GetNumVariables(); vi++) {
			varnames[vi] = varset->GetVariableName(vi);
		}
		output.SetAdditionalOutput(likelihood);
		output.Initialize(sampler->GetNumSamples(), varnames);
		sampler->SetPosteriorSampleHandler(output.GetSampleFunctionPointer(), NULL);
		sampler->SetOutputValuesHandler(output.GetAdditionalOutputFunctionPointer(), NULL);
		sampler->SetOutputPath(output_path);
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	// Everything has been set, initialize the sampler
	if (!sampler->Initialize()) {
		return -10;
	}

	printf("Starting sampler...\n");
	bcm::ProgressIndicatorConsole pic;
	sampler->SetProgressIndicator(&pic);

	// Run the sampler
	Real evidence = 0.0, evidence_sd = 0.0;
	result = sampler->Run(&evidence, &evidence_sd);

	printf("Sampling finished.\n");
	printf("Evidence: %g +/- %g\n", evidence, evidence_sd);

	sampler->Uninitialize();
	output.Close();
	bcm::safe_delete(sampler);
	bcm::safe_delete(likelihood);
	bcm::safe_delete(prior);
	bcm::safe_delete(varset);
	bcm::safe_delete(rng);
	
	if (result) {
		return 0;
	} else {
		return -20;
	}
}

int gen_variable_file(const po::variables_map& vm)
{
	SBMLDocument* Document = NULL;
	Model* model = NULL;
	bool result = true;
	
	try {
		// Load SBML file
		std::string fn = vm["model"].as<std::string>();
		Document = readSBML(fn.c_str());
		if (Document->getNumErrors() > 0) {
			LOGERROR("Errors reading SBML file %s: %d", fn.c_str(), Document->getNumErrors());
			for (unsigned int i = 0; i < Document->getNumErrors(); i++) {
				LOGERROR(" %d: %s", i, Document->getError(i)->getMessage().c_str());
			}
			bcm::safe_delete(Document);
			return false;
		}

		model = Document->getModel();

		// Generate variables
		boost::property_tree::ptree pt;
		boost::property_tree::ptree& root = pt.put("variableset", "");

		std::vector<std::string> species_class;
		species_class.resize(model->getNumSpecies());

		for (unsigned int i = 0; i < model->getNumSpecies(); i++) {
			const Species* sp = model->getSpecies(i);
			
			const XMLNode* annotation = sp->getAnnotation();
			for (unsigned int ai = 0; ai < annotation->getNumChildren(); ai++) {
				const XMLNode& child = annotation->getChild(ai);
				if (child.getPrefix() == "celldesigner" && child.getName() == "extension") {
					species_class[i] = child.getChild("speciesIdentity").getChild("class").getChild(0).getCharacters();
					if (species_class[i].empty()) {
						LOGERROR("Can't find species class for species %s [%s] - SBML file should be a CellDesigner file", sp->getId().c_str(), sp->getName().c_str());
						return false;
					}
				}
			}

			if (species_class[i] != "GENE" && species_class[i] != "DRUG" && species_class[i] != "UNKNOWN") {
				boost::property_tree::ptree& param = root.add("variable", "");
				param.put("<xmlattr>.name", std::string("base_") + sp->getName());
				param.put("<xmlattr>.distribution", "uniform");
				param.put("<xmlattr>.lower", "0.0");
				param.put("<xmlattr>.upper", "1.0");
			}
		}

		for (unsigned int i = 0; i < model->getNumReactions(); i++) {
			const Reaction* re = model->getReaction(i);

			if (re->getNumReactants() != 1) {
				LOGERROR("Can only handle reactions with 1 reactant");
				return false;
			}
			if (re->getNumProducts() != 1) {
				LOGERROR("Can only handle reactions with 1 product");
				return false;
			}

			unsigned int reactant_ix = std::numeric_limits<unsigned int>::max(), product_ix = std::numeric_limits<unsigned int>::max();
			for (unsigned int i = 0; i < model->getNumSpecies(); i++) {
				const Species* sp = model->getSpecies(i);
				if (sp->getId() == re->getReactant(0)->getSpecies()) {
					reactant_ix = i;
				}
				if (sp->getId() == re->getProduct(0)->getSpecies()) {
					product_ix = i;
				}
			}
			if (reactant_ix == std::numeric_limits<unsigned int>::max()) {
				LOGERROR("Could not find reactant for reaction %s", re->getId().c_str());
				return false;
			}
			if (product_ix == std::numeric_limits<unsigned int>::max()) {
				LOGERROR("Could not find product for reaction %s", re->getId().c_str());
				return false;
			}

			std::string reactant_name = model->getSpecies(reactant_ix)->getName();
			std::string product_name = model->getSpecies(product_ix)->getName();
			
			if (species_class[reactant_ix] == "DRUG") {
				boost::property_tree::ptree& ic50 = root.add("variable", "");
				ic50.put("<xmlattr>.name", std::string("ic50_") + reactant_name + std::string("_") + product_name);
				ic50.put("<xmlattr>.distribution", "normal");
				ic50.put("<xmlattr>.mu", "-7.0");
				ic50.put("<xmlattr>.sigma", "2.0");
				
				boost::property_tree::ptree& maxinhib = root.add("variable", "");
				maxinhib.put("<xmlattr>.name", std::string("maxinhib_") + reactant_name + std::string("_") + product_name);
				maxinhib.put("<xmlattr>.distribution", "uniform");
				maxinhib.put("<xmlattr>.lower", "0.0");
				maxinhib.put("<xmlattr>.upper", "1.0");
				
				boost::property_tree::ptree& logsteepness = root.add("variable", "");
				logsteepness.put("<xmlattr>.name", std::string("logsteepness_") + reactant_name + std::string("_") + product_name);
				logsteepness.put("<xmlattr>.distribution", "uniform");
				logsteepness.put("<xmlattr>.lower", "-1.0");
				logsteepness.put("<xmlattr>.upper", "1.0");
#if 0
			} else if (species_class[reactant_ix] == "PHENOTYPE") {
				boost::property_tree::ptree& param = root.add("variable", "");
				param.put("<xmlattr>.name", std::string("strength_") + reactant_name + std::string("_") + product_name);
				param.put("<xmlattr>.distribution", "exponential");
				param.put("<xmlattr>.lambda", "25.0");
#endif
			} else {
				boost::property_tree::ptree& param = root.add("variable", "");
				param.put("<xmlattr>.name", std::string("strength_") + reactant_name + std::string("_") + product_name);
				param.put("<xmlattr>.distribution", "uniform");

				if (species_class[reactant_ix] == "GENE") {
					param.put("<xmlattr>.lower", "0.0");
					param.put("<xmlattr>.upper", "1.0");
				} else {
					param.put("<xmlattr>.lower", "-1.0");
					param.put("<xmlattr>.upper", "1.0");
					param.put("<xmlattr>.logspace", "true");
				}
			}
		}
#if BOOST_VERSION / 100 % 1000 >= 58
			boost::property_tree::xml_parser::write_xml("prior.xml", pt, std::locale(), 
				boost::property_tree::xml_writer_make_settings<std::string>(' ', 1));
#else
			boost::property_tree::xml_parser::write_xml("prior.xml", pt, std::locale(), 
				boost::property_tree::xml_writer_make_settings<char>(' ', 1));
#endif
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	return 0;
}

int predict(const po::variables_map& vm)
{
	bcm::VariableSet* varset = NULL;
	bcm::RNG* rng = NULL;
	SBMLPDLikelihood* likelihood = NULL;
	bcm::NetCDFDataFile* samples = NULL;
	bcm::NetCDFDataFile* output = NULL;
	bool result = true;
	
	try {
#if 0
		// Determine the number of threads we should use
		size_t numthreads = vm["numthreads"].as<size_t>();
		if (numthreads <= 0) {
			numthreads = boost::thread::hardware_concurrency();
			printf("numthreads <= 0 specified, found %u hardware concurrency\n", (unsigned int)numthreads);
			if (numthreads <= 0) {
				numthreads = 1;
			}
		}
#else
		size_t numthreads = 1;
#endif
		
		varset = new bcm::VariableSet;
		result = varset->LoadFromXML(vm["prior"].as<std::string>());
		if (!result) {
			return -2;
		}

		likelihood = new SBMLPDLikelihood(numthreads);
		result = likelihood->SetVariableSet(varset);
		if (!result) {
			return -6;
		}
		result = likelihood->Initialize(vm["likelihood"].as<std::string>());
		if (!result) {
			return -8;
		}
		
		// Initialize input file
		samples = new bcm::NetCDFDataFile();
		result = samples->Open(vm["predict.input"].as<std::string>(), netCDF::NcFile::read);
		if (!result) {
			return -8;
		}

		std::string output_path = vm["output.folder"].as<std::string>();
		
		// Make sure output folder exists
		bcm::fix_path(output_path);
		if (!boost::filesystem::is_directory(output_path)) {
			std::string create_path = output_path.substr(0, output_path.size()-1);
			if (!boost::filesystem::create_directories(create_path)) {
				LOGERROR("Could not create output directory.");
				return false;
			}
		}

		// Initialize output file
		output = new bcm::NetCDFDataFile();
		result = output->Open(output_path + "prediction.nc", netCDF::NcFile::replace);
		if (!result) {
			return -8;
		}
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	size_t sample_count;
	size_t variable_count;
	result = samples->GetDimensionSize("samples", "sample_ix", &sample_count);
	result &= samples->GetDimensionSize("samples", "variables", &variable_count);
	if (!result) {
		LOGERROR("Input samples file does not appear to be a valid samples file");
		return -10;
	}
	if (variable_count != varset->GetNumVariables()) {
		LOGERROR("Input samples file contains %d variables, but variable set expects %d variables", variable_count, varset->GetNumVariables());
		return -11;
	}
	
	size_t output_values_count = 0;
	result &= likelihood->GetOutput(0, NULL, output_values_count);
	
	// Set up output file
	std::vector<unsigned int> sampleix(sample_count);
	for (size_t i = 0; i < sample_count; i++) {
		sampleix[i] = (unsigned int)i+1;
	}
	std::vector<std::string> variables(variable_count);
	for (size_t i = 0; i < variable_count; i++) {
		samples->GetValue("samples", "variables", i, variables[i]);
	}

	// Allocate variables
	result &= output->CreateGroup("samples");
	result &= output->CreateDimension("samples", "sample_ix", sampleix);
	result &= output->CreateDimension("samples", "variables", variables);
	result &= output->CreateVariable("samples", "variable_values", "sample_ix", "variables");
	result &= output->CreateVariable("samples", "log_prior", "sample_ix");
	result &= output->CreateVariable("samples", "log_likelihood", "sample_ix");
	result &= output->CreateVariable("samples", "log_posterior", "sample_ix");
	result &= likelihood->SetupAdditionalOutput(output, sample_count);

	printf("Starting simulation...\n");
	printf("Progres:   0%%");

	bcm::Timer timer;
	timer.Start();
	
	VectorReal sample(variable_count);
	VectorReal output_values(output_values_count);

	for (size_t i = 0; i < sample_count; i++) {
		printf("\b\b\b\b%3u%%", (unsigned int)((100*i) / sample_count));

		// Load sample
		samples->GetValuesDim2("samples", "variable_values", i, 0, variable_count, sample);
		output->PutValuesDim2("samples", "variable_values", i, 0, sample);

		Real lprior;
		samples->GetValue("samples", "log_prior", i, &lprior);

		Real llh = 0.0;
		result = likelihood->EvaluateLogProbability(0, sample.data(), llh);
		if (result) {
			likelihood->GetOutput(0, output_values.data(), output_values_count);
			likelihood->HandleAdditionalOutput(output, i, output_values.data());
		} else {
			llh = std::numeric_limits<Real>::quiet_NaN();
		}

		Real lposterior = llh + lprior;
		
		output->PutValue("samples", "sample_ix", i, (unsigned int)i);
		output->PutValue("samples", "log_prior", i, lprior);
		output->PutValue("samples", "log_likelihood", i, llh);
		output->PutValue("samples", "log_posterior", i, lposterior);
	}

	printf("\b\b\b\b100%%\n");
	printf("Simulation finished.\n");
	
	Real simulation_time = timer.GetElapsedSeconds();
	printf("Simulation time: %g\n", simulation_time);
	printf("Evaluations per second: %g\n", sample_count / simulation_time);
	
	samples->Close();
	output->Close();

	bcm::safe_delete(samples);
	bcm::safe_delete(output);
	bcm::safe_delete(likelihood);
	bcm::safe_delete(varset);
	bcm::safe_delete(rng);
	
	if (result) {
		return 0;
	} else {
		return -20;
	}
}

int main(int argc, char* argv[])
{
	printf("SBML Process Description inference tool - version %u.%u.%u\n", version_major, version_minor, version_fix);

	bcm::logger->SetLogToConsole(bcm::Logger::Error);
	
	int result = -1;
	try {
		po::options_description generic_options("Generic options", 120);
		generic_options.add_options()
			("help",															"produce help message")
			("numthreads,j", po::value<size_t>()->default_value(0),				"number of threads to use; set to 0 for using all available hardware concurrency")
		;

		po::options_description config_options("Configuration", 120);
		config_options.add_options()
			("model,m",										po::value<std::string>()->default_value("model.xml"),			"SBML file")
			("prior",										po::value<std::string>()->default_value("prior.xml"),			"File describing the priors for the parameters/initial conditions")
			("likelihood",									po::value<std::string>()->default_value("likelihood.xml"),		"File describing the likelihood functions")
			("output.folder",								po::value<std::string>()->default_value("output"),				"Folder in which output files will be generated, subfolder of the model directory.")
			("gen_variable_file",																							"Generate standard variableset file for the given sbml model.")
			("predict",																										"Make a prediction for the function described in the likelihood file, given a set of output samples.")
			("predict.input",								po::value<std::string>()->default_value("output.nc"),			"This should point to the output.nc file of a previous inference.")
		;
		bcm::SamplerFactory::AddOptionsDescription(config_options);

		po::options_description cmdline_only_options("Command-line only options", 120);
		cmdline_only_options.add_options()
			("config_file,c", po::value<std::string>()->default_value("config.txt"), "Configuration file")
		;
		
		// Add all options together
		po::options_description cmdline_options("Allowed options", 120);
		cmdline_options.add(generic_options).add(cmdline_only_options).add(config_options);

		po::positional_options_description pod;
		pod.add("model", 1);
		
		// Parse command line
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(pod).run(), vm);

		// Handle commands
		if (vm.count("help")) {
			std::cout << cmdline_options << "\n";
			result = 0;
		} else {
			// Parse config file, which may have been specified on the command line, otherwise load the default.
			std::string config_file = vm["config_file"].as<std::string>();
			if (!config_file.empty()) {
				po::store(po::parse_config_file<char>(config_file.c_str(), config_options), vm);
			}

			if (vm.count("gen_variable_file")) {
				result = gen_variable_file(vm);
			} else if (vm.count("predict")) {
				result = predict(vm);
			} else {
				result = run(vm);
			}
		}
	} catch (po::error &e) {
		LOGERROR("Error processing program options: %s", e.what());
	}

	return result;
}
